package dascosa.algebra;

import dascosa.algebra.function.Function;
import dascosa.data.PacketHeader;
import dascosa.data.PartialResult;
import dascosa.util.ArrayHashMap;
import dascosa.site.query.Executor;

import javax.sql.DataSource;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 18.feb.2008
 * Time: 16:05:37
 * To change this template use File | Settings | File Templates.
 */
public final class AggregateNode extends AlgebraNode
{
  private static final double SIZE_FACTOR = 0.1;
  private final Function functions[];
  private final String selectAttributes[];
  private final String groupAttributes[];
  private transient ArrayHashMap<Object[]> partialResults;
  private transient int selectColumns[];

  public AggregateNode(AlgebraNode resultNode, String selectAttributes[], Function functions[], String groupAttributes[])
  {
    super(resultNode);
    this.selectAttributes = selectAttributes;
    this.functions = functions;
    this.groupAttributes = groupAttributes;
  }

  public String toString()
  {
    return "AggregateNode(" + Arrays.toString(selectAttributes) + " f= " +
           Arrays.toString(functions) + " group by " + Arrays.toString(groupAttributes) + ')';
  }

  public boolean doOperation(Executor executor, DataSource dataSource)
  {
    if (operands == null)
      return false;

    if (result == null) {
      List<String> columnNames = new ArrayList<String>();
      selectColumns = new int[0];
      if (selectAttributes != null && selectAttributes.length > 0 && selectAttributes[0] != null) {
        columnNames.addAll(Arrays.asList(selectAttributes));
        selectColumns = new int[selectAttributes.length];
        for (int i = 0; i < selectAttributes.length; i++)
          selectColumns[i] = operands[0].getColumn(selectAttributes[i]);
      }

      for (Function function : functions) {
        columnNames.add(function.getColumnName());
        function.init(operands[0]);
      }

      result = new PartialResult(1, tuplesToSkip);
      PacketHeader resultHeader = new PacketHeader(columnNames, getOperationID(), getNodeID(),
                                                   getResultNodeID(0));
      result.addPacketHeader(executor, resultHeader);

      if (groupAttributes == null || groupAttributes.length == 0 || groupAttributes[0] == null) { // No grouping
        partialResults = new ArrayHashMap<Object[]>(null);
      }
      else {
        int groupColumns[] = new int[groupAttributes.length];
        for (int i = 0; i < groupColumns.length; i++)
          groupColumns[i] = operands[0].getColumn(groupAttributes[i]);
        partialResults = new ArrayHashMap<Object[]>(groupColumns);
      }
    }

    for (Iterator<Object[]> iterator = operands[0].unprocessedIterator(); iterator.hasNext();) {
      Object tuple[] = iterator.next();
      Object partialResultTuple[] = partialResults.get(tuple);
      if (partialResultTuple == null) {
        partialResultTuple = new Object[functions.length];
        partialResults.put(tuple, partialResultTuple);
      }
      for (int i = 0; i < functions.length; i++)
        partialResultTuple[i] = functions[i].processTuple(partialResultTuple[i], tuple);
    }

    if (operands[0].isComplete()) {
      PacketHeader resultHeader = result.getPacketHeader(0);
      for (Object[] groupTuple : partialResults.keySet()) {
        int counter = 0;
        Object resultTuple[] = new Object[resultHeader.getColumnCount()];

        for (int selectColumn : selectColumns)
          resultTuple[counter++] = groupTuple[selectColumn];

        for (int i = 0; i < functions.length; i++)
          resultTuple[counter++] = functions[i].processPartialResult(partialResults.get(groupTuple)[i]);

        result.addTuple(executor, resultHeader, resultTuple);
      }

      result.lastTupleAdded(executor, resultHeader);
    }
    return false;
  }

  public double getCostEstimate()
  {
    double cost = 0.0;
    double size = 0.0;

    for (AlgebraNode n : getOperandNodes()) {
      cost += n.getCostEstimate();
      size += n.getSizeEstimate();
    }
    cost += size;

    return cost;
  }

  public double getSizeEstimate()
  {
    double size = 0.0;

    for (AlgebraNode n : getOperandNodes()) {
      size += n.getSizeEstimate();
    }
    size *= SIZE_FACTOR;

    return size;
  }

  public PartialResult simulate()
  {
    if (operands == null)
      return null;

    for (Function function : functions)
      function.init(operands[0]);
    if (groupAttributes != null)
      for (String groupAttribute : groupAttributes)
        if (operands[0].getColumn(groupAttribute) == -1)
          throw new RuntimeException("Aggregation: Unable to find " + groupAttribute);
    if (selectAttributes != null)
      for (String selectAttribute : selectAttributes)
        if (operands[0].getColumn(selectAttribute) == -1)
          throw new RuntimeException("Aggregation: Unable to find " + selectAttribute);

    List<String> columnNames = new ArrayList<String>();
    if (selectAttributes != null && selectAttributes.length > 0 && selectAttributes[0] != null)
      columnNames.addAll(Arrays.asList(selectAttributes));
    for (Function function : functions)
      columnNames.add(function.getColumnName());

    PartialResult simulation = new PartialResult(1);
    PacketHeader resultPart = new PacketHeader(columnNames, getOperationID(), getNodeID(), getResultNodeID(0));
    simulation.addPacketHeader(null, resultPart);
    operands = null;

    return simulation;
  }
}
