package dascosa.algebra;

import dascosa.algebra.constraints.*;
import dascosa.algebra.function.*;
import dascosa.algebra.project.*;
import dascosa.algebra.scan.LogicalTableAccessNode;
import dascosa.parser.*;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 02.mai.2007
 * Time: 14:59:24
 * To change this template use File | Settings | File Templates.
 */
public final class AlgebraConverter implements SqlScriptTreeConstants
{
  public static AlgebraNode convertAlgebra(Node sqlScript) throws ParseException
  {
    AlgebraNode globalRoot = new DisplayNode();
    AlgebraNode currentRoot = globalRoot;

    Node skylineOfClause = findSubTree(sqlScript, JJTSKYLINEOFCLAUSE);
    if (skylineOfClause != null){
      currentRoot = parseSkylineOfClause(skylineOfClause, currentRoot);
    }

    Node orderByClause = findSubTree(sqlScript, JJTORDERBYCLAUSE);
    if (orderByClause != null)
      currentRoot = parseOrderByClause(orderByClause, currentRoot);

    parseSelectWithoutOrder(findSubTree(sqlScript, JJTSELECTWITHOUTORDER), currentRoot);

    return globalRoot;
  }

  public static int findPrimaryKeyValue(Node sqlScript, String primaryKey) throws ParseException
  {
    Node queryStatement = sqlScript.jjtGetChild(0);
    int columnNr = -1;
    for (int i = 0; i < queryStatement.jjtGetNumChildren(); i++) {
      SimpleNode child = (SimpleNode)queryStatement.jjtGetChild(i);
      if (child.getId() == JJTTABLECOLUMN) {
        columnNr++;
        if (getLeafValue(child).equalsIgnoreCase(primaryKey))
          break;
      }
    }
    if (columnNr == -1)
      throw new ParseException("Primary key not found");

    Node expressionNode = findSubTree(queryStatement, JJTPLSQLEXPRESSIONLIST);
    Node valueNode = expressionNode.jjtGetChild(columnNr);
    return Integer.parseInt(getLeafValue(valueNode));
  }

  private static Node findSubTree(Node sqlNode, int subtreeTopNodeID)
  {
    if (sqlNode instanceof SimpleNode && ((SimpleNode)sqlNode).getId() == subtreeTopNodeID)
      return sqlNode;
    for (int i = 0; i < sqlNode.jjtGetNumChildren(); i++) {
      Node node = findSubTree(sqlNode.jjtGetChild(i), subtreeTopNodeID);
      if (node != null)
        return node;
    }
    return null;
  }

  public static String findTableName(Node sqlScript)
  {
    return getLeafValue(findSubTree(sqlScript, JJTTABLEREFERENCE));
  }

  private static String getLeafValue(Node node)
  {
    while (true) {
      if (node instanceof SqlValue)
        return ((SqlValue)node).getValue();
      node = node.jjtGetChild(0);
    }
  }

  private static AlgebraNode parseAggregation(Node selectList, Node groupByClause, AlgebraNode currentRoot) throws ParseException
  {
    List<Function> functions = new ArrayList<Function>();
    List<String> selectAttributes = new ArrayList<String>();
    for (int i = 0; i < selectList.jjtGetNumChildren(); i++) {
      Node selectItem = selectList.jjtGetChild(i);
      SimpleNode child = (SimpleNode)selectItem.jjtGetChild(0);
      if (child.getId() == JJTTABLECOLUMN)
        selectAttributes.add(parseTableColumn(child));
      else if (child.getId() == JJTFUNCTIONCALL) {
        Function function = parseFunctionCall(child, findSubTree(selectItem, JJTASOBJECTNAME));
        if (function != null)
          functions.add(function);
      }
    }

    List<String> groupAttributes = new ArrayList<String>();
    if (groupByClause != null) {
      Node expressionList = groupByClause.jjtGetChild(0);
      for (int i = 0; i < expressionList.jjtGetNumChildren(); i++) {
        Node primaryExpression = expressionList.jjtGetChild(i);
        groupAttributes.add(parseTableColumn(findSubTree(primaryExpression, JJTTABLECOLUMN)));
      }
      Node sqlExpression = findSubTree(groupByClause, JJTSQLEXPRESSION);
      if (sqlExpression != null)  // Having
        currentRoot = parseSqlExpression(sqlExpression, currentRoot, null);
    }

    if (!functions.isEmpty())
      currentRoot = new AggregateNode(currentRoot,
                                      selectAttributes.toArray(new String[selectAttributes.size()]),
                                      functions.toArray(new Function[functions.size()]),
                                      groupAttributes.toArray(new String[groupAttributes.size()]));

    return currentRoot;
  }

  private static void parseFromClause(Node fromClause, AlgebraNode currentRoot) throws ParseException
  {
    for (int i = 0; i < fromClause.jjtGetNumChildren(); i++) {
      if (i != fromClause.jjtGetNumChildren() - 1) { // Not last site
        currentRoot = new CrossProductNode(currentRoot);
      }
      Node child = fromClause.jjtGetChild(i);
      if (child instanceof SimpleNode && ((SimpleNode)child).getId() == JJTFROMITEM)
        parseFromItem(child, currentRoot);
    }
  }

  private static void parseFromItem(Node fromItem, AlgebraNode currentRoot) throws ParseException
  {
    if (fromItem.jjtGetNumChildren() == 1 && ((SimpleNode)fromItem.jjtGetChild(0)).getId() != JJTFROMITEMEXPRESSION)
      new LogicalTableAccessNode(currentRoot, getLeafValue(fromItem));
    else if (fromItem.jjtGetNumChildren() == 2 && ((SimpleNode)fromItem.jjtGetChild(1)).getId() == JJTASOBJECTNAME) {
      String tableName = getLeafValue(fromItem.jjtGetChild(0));
      String asName = getLeafValue(fromItem.jjtGetChild(1));
      new LogicalTableAccessNode(currentRoot, tableName, asName);
    }
    else if (fromItem.jjtGetNumChildren() == 2 && ((SimpleNode)fromItem.jjtGetChild(1)).getId() == JJTJOINEREXPRESSION) {
      Node joinerExpression = fromItem.jjtGetChild(1);
      String tableName1 = getLeafValue(fromItem.jjtGetChild(0));
      String tableName2 = getLeafValue(findSubTree(joinerExpression, JJTTABLEREFERENCE));
      if (((SimpleNode)joinerExpression.jjtGetChild(0)).getId() == JJTOUTERJOIN) {
        String type = getLeafValue(joinerExpression.jjtGetChild(0));
        AlgebraNode outerJoin;
        if (type.equals("left outer join"))
          outerJoin = new OuterJoinNode(currentRoot, OuterJoinNode.TYPE_LEFT_OUTER);
        else if (type.equals("right outer join"))
          outerJoin = new OuterJoinNode(currentRoot, OuterJoinNode.TYPE_RIGHT_OUTER);
        else
          outerJoin = new OuterJoinNode(currentRoot, OuterJoinNode.TYPE_FULL_OUTER);
        new LogicalTableAccessNode(outerJoin, tableName1);
        new LogicalTableAccessNode(outerJoin, tableName2);
      }
      else {
        AlgebraNode naturalJoin = new NaturalJoinNode(currentRoot);
        new LogicalTableAccessNode(naturalJoin, tableName1);
        new LogicalTableAccessNode(naturalJoin, tableName2);
      }
    }
    else if (((SimpleNode)fromItem.jjtGetChild(0)).getId() == JJTFROMITEMEXPRESSION) {
      Node fromItemExpression = fromItem.jjtGetChild(0);
      if (fromItemExpression.jjtGetNumChildren() > 1 && ((SimpleNode)fromItemExpression.jjtGetChild(1)).getId() == JJTJOINEREXPRESSION) {
        String tableName1 = getLeafValue(fromItemExpression.jjtGetChild(0));
        AlgebraNode currentNode = new NaturalJoinNode(currentRoot);
        new LogicalTableAccessNode(currentNode, tableName1);
        for (int i = 1; i < fromItemExpression.jjtGetNumChildren() - 1; i++) {
          Node joinerExpression = fromItemExpression.jjtGetChild(i);
          String tableName2 = getLeafValue(findSubTree(joinerExpression, JJTTABLEREFERENCE));
          currentNode = new NaturalJoinNode(currentNode);
          new LogicalTableAccessNode(currentNode, tableName2);
        }
        Node lastNode = fromItemExpression.jjtGetChild(fromItemExpression.jjtGetNumChildren() - 1);
        String tableNameN = getLeafValue(findSubTree(lastNode, JJTTABLEREFERENCE));
        new LogicalTableAccessNode(currentNode, tableNameN);
      }
      else { // TODO: Complex from...
        throw new ParseException("Complex from not supported: " + fromItemExpression);
      }
    }
  }

  private static Function parseFunctionCall(Node functionCall, Node asObjectName) throws ParseException
  {
    String functionType = getLeafValue(functionCall.jjtGetChild(0));
    String functionAttribute = null;
    if (functionCall.jjtGetNumChildren() > 1)
      functionAttribute = parseTableColumn(findSubTree(functionCall.jjtGetChild(1), JJTTABLECOLUMN));
    Function function;
    if (asObjectName == null) {
      if (functionType.equalsIgnoreCase("sum"))
        function = new SumFunction(functionAttribute);
      else if (functionType.equalsIgnoreCase("count"))
        function = new CountFunction();
      else if (functionType.equalsIgnoreCase("min"))
        function = new MinFunction(functionAttribute);
      else if (functionType.equalsIgnoreCase("max"))
        function = new MaxFunction(functionAttribute);
      else if (functionType.equalsIgnoreCase("avg"))
        function = new AvgFunction(functionAttribute);
      else
        throw new ParseException("Unsupported function: " + functionType);
    }
    else {
      String columnName = getLeafValue(asObjectName);
      if (functionType.equalsIgnoreCase("sum"))
        function = new SumFunction(functionAttribute, columnName);
      else if (functionType.equalsIgnoreCase("count"))
        function = new CountFunction(columnName);
      else if (functionType.equalsIgnoreCase("min"))
        function = new MinFunction(functionAttribute, columnName);
      else if (functionType.equalsIgnoreCase("max"))
        function = new MaxFunction(functionAttribute, columnName);
      else if (functionType.equalsIgnoreCase("avg"))
        function = new AvgFunction(functionAttribute, columnName);
      else
        throw new ParseException("Unsupported function: " + functionType);
    }
    return function;
  }

  private static AlgebraNode parseOrderByClause(Node orderByClause, AlgebraNode currentRoot)
  {
    String[] attributeNames = new String[orderByClause.jjtGetNumChildren()];
    boolean ascending[] = new boolean[attributeNames.length];
    Arrays.fill(ascending, true); //The default order is acending

    /**
     * The first child of an orderByClause is the OrderSpec which
     * can have one child to identify the attribute (S_Identifier) and
     * another child to identify the order (AscendingOrderSpec or DescendingOrderSpec)
     */
    for (int i = 0; i < orderByClause.jjtGetNumChildren(); i++){
        Node orderSpec = orderByClause.jjtGetChild(i);
        if(orderSpec!=null){
            Node attribute = orderSpec.jjtGetChild(0);
            if(attribute!=null){
                attributeNames[i] = getLeafValue(attribute.jjtGetChild(0));
            }
            if(orderSpec.jjtGetNumChildren()>1){
                Node sortingOrder = orderSpec.jjtGetChild(1);
                if(sortingOrder!=null){
                    if(((SimpleNode)sortingOrder).getId() == JJTDESCENDINGORDERSPEC){
                        ascending[i]=false;
                    }
                }
            }
        }        
    }        

    return new OrderNode(currentRoot, attributeNames, ascending);
  }

  /**
   * Creates Skyline AlgebraNode.
   * @param skylineOfClause
   * @param currentRoot
   * @return
   */
  private static AlgebraNode parseSkylineOfClause(Node skylineOfClause, AlgebraNode currentRoot)
  {
    boolean distinct;
    int nextChild=0;

    //The first child of an skyilneOfClause may be an SQLDistinct
    SimpleNode firstChild = (SimpleNode)skylineOfClause.jjtGetChild(0);
    if(firstChild!=null && firstChild.getId() == JJTSQLDISTINCT){
        nextChild=1;
        distinct = true;
    }else{
        distinct = false;
    }

    //It is created after we know if the distinct is part of the query or not
    String[] attributeNames = new String[skylineOfClause.jjtGetNumChildren()-nextChild];
    SkylineNode.SkylineParams[] skyParams = new SkylineNode.SkylineParams[attributeNames.length];
    Arrays.fill(skyParams, SkylineNode.SkylineParams.MIN); //The default is to minimize


     /* The SkylineSpec have one child to identify the attribute (S_Identifier) and
     * another child to identify the skyline parameter which can be (min, max or diff)
     */
    for (int i = nextChild; i < skylineOfClause.jjtGetNumChildren(); i++){
        Node skylineSpec = skylineOfClause.jjtGetChild(i);
        int index = i - nextChild;
        
        if(skylineSpec!=null){
            Node attribute = skylineSpec.jjtGetChild(0);
            if(attribute!=null){
                attributeNames[index] = getLeafValue(attribute.jjtGetChild(0));
            }
            if(skylineSpec.jjtGetNumChildren()>1){
                SimpleNode skylineParam = (SimpleNode)skylineSpec.jjtGetChild(1);
                switch(skylineParam.getId()){
                    case JJTMAXSPEC:
                        skyParams[index] = SkylineNode.SkylineParams.MAX;
                    break;
                    case JJTMINSPEC:
                        skyParams[index] = SkylineNode.SkylineParams.MIN;
                    break;
                    case JJTDIFFSPEC:
                        skyParams[index] = SkylineNode.SkylineParams.DIFF;
                    break;
                }                
            }
        }
    }

    return new SkylineNode(currentRoot, attributeNames, skyParams, distinct);
  }

  private static AlgebraNode parseSelectList(Node selectList, AlgebraNode currentRoot) throws ParseException
  {
    Collection<ProjectOperation> projectOperations = new ArrayList<ProjectOperation>();
    for (int i = 0; i < selectList.jjtGetNumChildren(); i++) {
      Node child = selectList.jjtGetChild(i);
      if (child instanceof SimpleNode && ((SimpleNode)child).getId() == JJTSELECTITEM) {
        Node tableColumn = findSubTree(child, JJTTABLECOLUMN);
        String attributeName = parseTableColumn(tableColumn);
        Node asObjectName = findSubTree(child, JJTASOBJECTNAME);
        if (asObjectName == null)
          projectOperations.add(new Project(attributeName));
        else
          projectOperations.add(new Rename(attributeName, getLeafValue(asObjectName)));
      }
      else {
        throw new ParseException("Unsupported select item: " + child);
      }
    }
    if (!projectOperations.isEmpty()) {
      ProjectOperation operations[] = projectOperations.toArray(new ProjectOperation[projectOperations.size()]);
      return new ExtendedProjectNode(currentRoot, operations);
    }
    return currentRoot;
  }

  private static void parseSelectWithoutOrder(Node selectWithoutOrder, AlgebraNode currentRoot) throws ParseException
  {
    Node distinct = findSubTree(selectWithoutOrder, JJTSQLDISTINCT);
    if (distinct != null)
      currentRoot = new RemoveDuplicatesNode(currentRoot);
    Node selectList = findSubTree(selectWithoutOrder, JJTSELECTLIST);
    Node functionCall = findSubTree(selectList, JJTFUNCTIONCALL);
    if (functionCall != null) { // Aggregation
      Node groupByClause = findSubTree(selectWithoutOrder, JJTGROUPBYCLAUSE);
      currentRoot = parseAggregation(selectList, groupByClause, currentRoot);
    }
    else { // Normal Select
      currentRoot = parseSelectList(selectList, currentRoot);
    }

    Node whereClause = findSubTree(selectWithoutOrder, JJTWHERECLAUSE);
    if (whereClause != null) {
      Collection<Node> subQueries = new ArrayList<Node>();
      currentRoot = parseSqlExpression(whereClause.jjtGetChild(0), currentRoot, subQueries);
      parseFromClause(findSubTree(selectWithoutOrder, JJTFROMCLAUSE), currentRoot);
      // Have to parse the subqueries after FROM to ensure that they are the second node added to
      // ConstraintNode as the order nodes are added are important
      for (Node subQuery : subQueries)
        parseSelectWithoutOrder(subQuery, currentRoot);
    }
    else
      parseFromClause(findSubTree(selectWithoutOrder, JJTFROMCLAUSE), currentRoot);
  }

  private static AlgebraNode parseSqlExpression(Node sqlExpression, AlgebraNode currentRoot, Collection<Node> subQueries) throws ParseException
  {
    SimpleNode child = (SimpleNode)sqlExpression.jjtGetChild(0);
    ConstraintNode constraintNode;
    if (child.getId() == JJTSQLANDEXPRESSION) {
      Constraint constraints[][] = new Constraint[child.jjtGetNumChildren()][];
      for (int i = 0; i < child.jjtGetNumChildren(); i++) {
        SimpleNode childchild = (SimpleNode)child.jjtGetChild(i);
        if (childchild.getId() == JJTSQLPRIMARYEXPRESSION) {
          SimpleNode orNode = (SimpleNode)findSubTree(childchild, JJTSQLOREXPRESSION);
          if (orNode != null) {
            Constraint orBlock[] = new Constraint[orNode.jjtGetNumChildren()];
            for (int j = 0; j < orNode.jjtGetNumChildren(); j++)
              orBlock[j] = parseUnaryLogicalExpression((SimpleNode)orNode.jjtGetChild(j), subQueries);
            constraints[i] = orBlock;
          }
        }
        else if (childchild.getId() == JJTSQLOREXPRESSION) {
          Constraint orBlock[] = new Constraint[childchild.jjtGetNumChildren()];
          for (int j = 0; j < childchild.jjtGetNumChildren(); j++)
            orBlock[j] = parseUnaryLogicalExpression((SimpleNode)childchild.jjtGetChild(j), subQueries);
          constraints[i] = orBlock;
        }
        else
          constraints[i] = new Constraint[] {parseUnaryLogicalExpression(childchild, subQueries)};
      }
      constraintNode = new ConstraintNode(currentRoot, constraints, true);
    }
    else if (child.getId() == JJTSQLOREXPRESSION) {
      Constraint constraints[][] = new Constraint[child.jjtGetNumChildren()][];
      for (int i = 0; i < child.jjtGetNumChildren(); i++) {
        SimpleNode childchild = (SimpleNode)child.jjtGetChild(i);
        if (childchild.getId() == JJTSQLPRIMARYEXPRESSION) {
          SimpleNode andNode = (SimpleNode)findSubTree(childchild, JJTSQLANDEXPRESSION);
          if (andNode != null) {
            Constraint andBlock[] = new Constraint[andNode.jjtGetNumChildren()];
            for (int j = 0; j < andNode.jjtGetNumChildren(); j++)
              andBlock[j] = parseUnaryLogicalExpression((SimpleNode)andNode.jjtGetChild(j), subQueries);
            constraints[i] = andBlock;
          }
        }
        else if (childchild.getId() == JJTSQLANDEXPRESSION) {
          Constraint andBlock[] = new Constraint[childchild.jjtGetNumChildren()];
          for (int j = 0; j < childchild.jjtGetNumChildren(); j++)
            andBlock[j] = parseUnaryLogicalExpression((SimpleNode)childchild.jjtGetChild(j), subQueries);
          constraints[i] = andBlock;
        }
        else
          constraints[i] = new Constraint[] {parseUnaryLogicalExpression(childchild, subQueries)};
      }
      constraintNode = new ConstraintNode(currentRoot, constraints, false);
    }
    else
      constraintNode = new ConstraintNode(currentRoot, parseUnaryLogicalExpression(child, subQueries));
    return constraintNode;
  }

  private static String parseTableColumn(Node tableColumn)
  {
    StringBuilder name = new StringBuilder(getLeafValue(tableColumn.jjtGetChild(0)));
    for (int i = 1; i < tableColumn.jjtGetNumChildren(); i++)
      name.append('.').append(getLeafValue(tableColumn.jjtGetChild(i)));
    return name.toString();
  }

  private static Constraint parseUnaryLogicalExpression(SimpleNode unaryExpr, Collection<Node> subQueries) throws ParseException
  {
    if (unaryExpr.getId() == JJTSQLRELOPEXPRESSION) {
      SimpleNode operand1 = (SimpleNode)unaryExpr.jjtGetChild(0).jjtGetChild(0);
      String tableColumn1;
      if (operand1.getId() == JJTTABLECOLUMN)
        tableColumn1 = parseTableColumn(operand1);
      else if (operand1.getId() == JJTFUNCTIONCALL)
        tableColumn1 = parseFunctionCall(operand1, null).getColumnName();
      else
        throw new ParseException("Unsuppored operand: " + operand1);
      String relationalOperator = getLeafValue(unaryExpr.jjtGetChild(1));
      SimpleNode operand2 = (SimpleNode)unaryExpr.jjtGetChild(2);

      if (operand2.getId() == JJTSQLPRIMARYEXPRESSION) {
        SimpleNode value = (SimpleNode)operand2.jjtGetChild(0);
        if (value.getId() == JJTS_NUMBER) {
          try {
            long longValue = Long.parseLong(getLeafValue(operand2));
            return new LongConstraint(tableColumn1, relationalOperator, longValue);
          } catch (NumberFormatException e) {
            throw new ParseException("Not a number: " + getLeafValue(operand2));
          }
        }
        else if (value.getId() == JJTTABLECOLUMN) {
          Node tableColumn2 = findSubTree(operand2, JJTTABLECOLUMN); // Strange that strings are table columns...
          if (getLeafValue(tableColumn2).startsWith("'")) {
            String stringValue = getLeafValue(tableColumn2);
            stringValue = stringValue.substring(1, stringValue.length() - 1);
            return new ComparableConstraint(tableColumn1, relationalOperator, stringValue);
          }
          else
            return new AttributeConstraint(tableColumn1, relationalOperator, parseTableColumn(tableColumn2));
        }
        else
          throw new ParseException("Unsuppored operand: " + operand2);
      }
      else if (operand2.getId() == JJTSUBQUERY) {
        subQueries.add(operand2.jjtGetChild(0));
        return new UncorrelatedNumberConstraint(tableColumn1, relationalOperator);
      }
      else
        throw new ParseException("Unsuppored operand: " + operand2);
    }
    else if (unaryExpr.getId() == JJTSQLRELATIONALEXPRESSION) {
      String tableColumn = parseTableColumn(findSubTree(unaryExpr.jjtGetChild(0), JJTTABLECOLUMN));
      SimpleNode relOp = (SimpleNode)unaryExpr.jjtGetChild(1);
      if (relOp.getId() == JJTSQLLIKECLAUSE) {
        boolean not = findSubTree(relOp, JJTSQLNOT) != null;
        String stringValue = getLeafValue(findSubTree(relOp, JJTSQLSIMPLEEXPRESSION));
        stringValue = stringValue.substring(1, stringValue.length() - 1);
        String regex = stringValue.replaceAll("%", ".*");
        regex = regex.replaceAll("_", ".");
        return new RegexConstraint(tableColumn, regex, not, stringValue);
      }
      else if (relOp.getId() == JJTISNULLCLAUSE) {
        boolean not = findSubTree(relOp, JJTSQLNOT) != null;
        if (not)
          return new ComparableConstraint(tableColumn, "<>", null);
        else
          return new ComparableConstraint(tableColumn, "=", null);
      }
      else if (relOp.getId() == JJTSQLINCLAUSE) {
        boolean not = findSubTree(relOp, JJTSQLNOT) != null;
        subQueries.add(findSubTree(relOp, JJTSUBQUERY));
        return new NestedInConstraint(tableColumn, not);
      }
      else
        throw new ParseException("Unsuppored operator: " + relOp);
    }
    else
      throw new ParseException("Unsuppored expression: " + unaryExpr);
  }

  private AlgebraConverter()
  {
  }
}
