package dascosa.algebra;

import dascosa.Dascosa;
import dascosa.algebra.scan.TableAccessNode;
import dascosa.data.*;
import dascosa.message.ComplexMessage;
import dascosa.site.query.Executor;
import dascosa.site.query.optimizer.Option;
import dascosa.site.query.optimizer.ResultSetIdentifier;
import dascosa.util.FilteredOutput;
import rice.p2p.commonapi.Message;
import rice.p2p.commonapi.NodeHandle;

import javax.sql.DataSource;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 25.apr.2007
 * Time: 14:31:55
 * To change this template use File | Settings | File Templates.
 */
public abstract class AlgebraNode implements ComplexMessage {

  private static long ID;

  // Self
  private final long nodeID; // Unique ID
  private long operationID; // Nodes performing the same operation (distributed) have same ID. Only changed during planning
  private NodeHandle assignedSite; // Only changed during planning
  protected PartialResult result; // Only used after site has been received
  transient int[] startTags;
  protected transient int tuplesToSkip;
  private transient boolean halted;
  private int expectedCardinality;
  //private NodeHandle[] optionalSites; // Origins

  // Input
  private final int numberOfOperands;
  protected transient Operand[] operands;
  private final ArrayList<AlgebraNode> operandNodes = new ArrayList<AlgebraNode>();

  // Output
  private final long resultNodeIDs[]; // Only changed during planning
  private final transient AlgebraNode resultNodes[]; // Only used during planning  
  private final NodeHandle resultSites[]; // Only changed during planning
  private double plannedCost;


  static {
    ID = new Random(System.currentTimeMillis()).nextLong() / 2;
  }

  public NodeHandle[] getResultSites() {
    return resultSites;
  }

  private static void addExternalOperands(List<AlgebraNode> externalOperands, AlgebraNode algebraNode, NodeHandle site) {
    for (AlgebraNode operandNode : algebraNode.operandNodes)
      if (operandNode.getAssignedSite() == site)
        addExternalOperands(externalOperands, operandNode, site);
      else
        externalOperands.add(operandNode);
  }

  protected AlgebraNode(AlgebraNode resultNodes[]) {
    this(resultNodes, 1);
  }

  protected AlgebraNode(int index, AlgebraNode resultNode) {
    this(index, resultNode, 1);
  }
  protected AlgebraNode(AlgebraNode resultNode) {
    this(new AlgebraNode[]{resultNode}, 1);
  }

  protected AlgebraNode(AlgebraNode resultNode, int numberOfOperands) {
    this(new AlgebraNode[]{resultNode}, numberOfOperands);
  }

  protected AlgebraNode(int index, AlgebraNode resultNode, int numberOfOperands) {
    this(index, new AlgebraNode[]{resultNode}, numberOfOperands);
  }

  private AlgebraNode(AlgebraNode[] resultNodes, int numberOfOperands) {
    this(-1, resultNodes, numberOfOperands);
  }

  private AlgebraNode(int index, AlgebraNode[] resultNodes, int numberOfOperands) {
    expectedCardinality = -1; //No estimate made yet
    this.resultNodes = resultNodes;
    this.numberOfOperands = numberOfOperands;
    resultSites = new NodeHandle[resultNodes.length];
    nodeID = ID++;
    operationID = nodeID; // Default
    resultNodeIDs = new long[resultNodes.length];
    for (int i = 0; i < resultNodes.length; i++) {
      resultNodeIDs[i] = resultNodes[i].getNodeID();
      resultSites[i] = resultNodes[i].getAssignedSite();
    }
    for (AlgebraNode resultNode : resultNodes)
      resultNode.operandNodes.add(index==-1?resultNode.operandNodes.size():index, this);
    //Dascosa.out.println(FilteredOutput.DEBUG, "Created AlgebraNode instance with nodeId: "+nodeID+ " type: "+toString());
  }

  // Constructor used by CacheNode
  protected AlgebraNode(long nodeID, long operationID, long resultNodeID, NodeHandle resultSite) {
    this.nodeID = nodeID;
    resultNodes = new AlgebraNode[0];
    resultNodeIDs = new long[1];
    resultNodeIDs[0] = resultNodeID;
    resultSites = new NodeHandle[1];
    resultSites[0] = resultSite;
    numberOfOperands = 0;
  }

// --------------------- Begin Interface ComplexMessage ---------------------
  public void prepareMessage() {
    operands = null;
  }

// --------------------- End Interface ComplexMessage ---------------------
// --------------------- Begin Interface Message ---------------------
  public int getPriority() {
    return Message.LOW_PRIORITY;
  }

// --------------------- End Interface Message ---------------------

  // Self
  /**
   *
   * @param executor
   * @param dataSource
   * @return True if the operation was halted to wait for packet reciept.
   */
  public abstract boolean doOperation(Executor executor, DataSource dataSource);

  public abstract PartialResult simulate();

  public long getNodeID() {
    return nodeID;
  }

  public long getOperationID() {
    return operationID;
  }

  public void setOperationID(long operationID) {
    this.operationID = operationID;
  }

  public NodeHandle getAssignedSite() {
    return assignedSite;
  }

  public void setAssignedSite(NodeHandle site) {
    assignedSite = site;
    for (AlgebraNode operandNode : operandNodes)
      for (int i = 0; i < operandNode.getResultNodeCount(); i++)
        if (operandNode.resultNodeIDs[i] == nodeID)
          operandNode.resultSites[i] = site;
  }

  public void removeFailedSite(NodeHandle failedSite) {
    if (assignedSite == failedSite) {
      setAssignedSite(null);
      for (AlgebraNode operandNode : operandNodes)
        operandNode.removeFailedSite(failedSite);
    }
  }

  public PartialResult getResult() {
    return result;
  }

  public void clearResult() {
    if (result != null)
      result.clear();
  }

  public void setStartTags(int startTags[]) {
    this.startTags = startTags;
  }

  public void setTuplesToSkip(int tuplesToSkip) {
    this.tuplesToSkip = tuplesToSkip;
  }

  public boolean isHalted() {
    return halted;
  }

  public void setHalted(boolean halted) {
    this.halted = halted;
  }

  // Input
  public int getNumberOfOperands() {
    return numberOfOperands;
  }

  public ArrayList<AlgebraNode> getOperandNodes() {
    return operandNodes;
  }

  public void removeOperandNode(AlgebraNode operandNode) {
    operandNodes.remove(operandNode);
  }

  public List<AlgebraNode> getExternalOperandNodes() {
    List<AlgebraNode> externalOperands = new ArrayList<AlgebraNode>();
    AlgebraNode.addExternalOperands(externalOperands, this, assignedSite);
    return externalOperands;
  }

  // Output
  public int getResultNodeCount() {
    return resultNodeIDs.length;
  }

  public long getResultNodeID(int index) {
    return resultNodeIDs[index];
  }

  void setResultNode(int index, AlgebraNode resultNode) {
    resultNodes[index] = resultNode;
    resultNodeIDs[index] = resultNode.getNodeID();
  }

  /**
   * Can only be used during planning
   */
  public AlgebraNode getResultNode(int index) {
    return resultNodes[index];
  }

  public NodeHandle getResultSite(int index) {
    return resultSites[index];
  }

  // Operand handling
  public synchronized void deliverOperandHeader(PacketHeader packetHeader) {
    if (operands == null)
      operands = new Operand[numberOfOperands];
    for (int i = 0; i < operands.length; i++)
      if (operands[i] == null) {
        List<Long> sourceAlgebraNodeIDs = new ArrayList<Long>();
        for (AlgebraNode operandNode : operandNodes)
          if (operandNode.getOperationID() == packetHeader.getSourceAlgebraOperationID())
            sourceAlgebraNodeIDs.add(operandNode.getNodeID());
        long array[] = new long[sourceAlgebraNodeIDs.size()];
        for (int j = 0; j < array.length; j++)
          array[j] = sourceAlgebraNodeIDs.get(j);
        operands[i] = new Operand(packetHeader.getSourceAlgebraOperationID(), array, startTags); // Assumes one operand, startTags won't be valid otherwise
        operands[i].addPacketHeader(packetHeader);

        if (operands[numberOfOperands - 1] != null) {
          // Sort operands to get deterministic results
          final Map<Long,Integer> indexMap = new HashMap<Long,Integer>();
          for (int index = 0; index < operandNodes.size(); index++)
            indexMap.put(operandNodes.get(index).getOperationID(), index);
          Arrays.sort(operands, new Comparator<Operand>() {
            public int compare(Operand o1, Operand o2) {
              return indexMap.get(o1.getSourceAlgebraOperationID()) - indexMap.get(o2.getSourceAlgebraOperationID());
            }
          });
        }
        return;
      } else if (operands[i].getSourceAlgebraOperationID() == packetHeader.getSourceAlgebraOperationID()) {
        operands[i].addPacketHeader(packetHeader);
        return;
      }
    System.err.println(this + " - Too many data headers delivered");
  }

  public synchronized void deliverOperandPacket(Packet packet) {
    for (Operand operand : operands)
      if (operand != null &&
              operand.getSourceAlgebraOperationID() == packet.getSourceAlgebraOperationID()) {
        operand.addPacket(packet);
        return;
      }
    System.err.println(this + " - Too many data packets delivered");
  }

  // Other
  public List<TableAccessNode> getUnassignedTableNodes() {
    List<TableAccessNode> tableNodes = new ArrayList<TableAccessNode>();
    addUnassignedTableNodes(tableNodes);
    return tableNodes;
  }

  protected void addUnassignedTableNodes(List<TableAccessNode> tableNodes) {
    for (AlgebraNode algebraNode : operandNodes)
      algebraNode.addUnassignedTableNodes(tableNodes);
  }

  public boolean allTableNodesAssigned() {
    for (AlgebraNode algebraNode : operandNodes)
      if (!algebraNode.allTableNodesAssigned())
        return false;
    return true;
  }

  public String canonicalForm() {
    StringBuilder sb = new StringBuilder();
    canonicalForm(sb);
    if (sb.length() > 0)
      sb.deleteCharAt(0);
    return sb.toString();
  }

  public int[] getLastPacketTags(AlgebraNode operandNode) {
    if (operands == null)
      return new int[0];
    for (Operand operand : operands)
      if (operand != null && operand.getSourceAlgebraOperationID() == operandNode.getOperationID())
        return operand.getLastPacketTags(operandNode.getNodeID());
    return new int[0];
  }

  public int getReceivedTupleCount(AlgebraNode operandNode) {
    if (operands == null)
      return 0;
    for (Operand operand : operands)
      if (operand != null && operand.getSourceAlgebraOperationID() == operandNode.getOperationID())
        return operand.getReceivedTupleCount(operandNode.getNodeID());
    return 0;
  }

  public void removeCrashNodes(NodeHandle failedSite) {
    for (int i = 0; i < operandNodes.size(); i++) {
      AlgebraNode node = operandNodes.get(i);
      if (node instanceof CrashNode && node.getAssignedSite() == failedSite) {
        AlgebraNode operandNode = node.getOperandNodes().get(0);
        operandNodes.set(i, operandNode);
        operandNode.resultSites[0] = assignedSite;
        for (int j = 0; j < operandNode.resultNodeIDs.length; j++)
          if (operandNode.resultNodeIDs[j] == node.getNodeID())
            operandNode.resultNodeIDs[j] = getNodeID();
        if (operands != null)
          for (Operand operand : operands)
            if (operand != null && operand.getSourceAlgebraOperationID() == node.getOperationID())
              operand.replaceOperandNode(node, operandNode);
      }
    }
    for (AlgebraNode operandNode : operandNodes)
      if (operandNode.getAssignedSite() == failedSite)
        operandNode.removeCrashNodes(failedSite);
  }

  public void removeOperandsFrom(AlgebraNode algebraNode) {
    if (operands != null)
      for (Operand operand : operands)
        if (operand != null)
          operand.removeOperandsFrom(algebraNode);
  }

  public void setPlannedCost(double cost) {
    this.plannedCost = cost;
  }

  public double getPlannedCost() {
    return plannedCost;
  }

  protected void canonicalForm(StringBuilder sb) {
    sb.append(" {");
    sb.append(toString());
    for (AlgebraNode operand : operandNodes)
      if (operand instanceof CacheNode)
        operand.getOperandNodes().get(0).canonicalForm(sb);
      else
        operand.canonicalForm(sb);
    sb.append('}');
  }

  protected void copyResultSites(AlgebraNode algebraNode) {
    System.arraycopy(algebraNode.resultSites, 0, resultSites, 0, resultSites.length);
  }

  /**
   * Get (estimated) computational cost of this node
   *
   * @return Computational cost
   */
  public abstract double getCostEstimate();

  /**
   * Get (estimated) size of result.
   *
   * @return Size of result
   */
  public abstract double getSizeEstimate();

  public void setExpectedCardinallity(int i) {
    expectedCardinality = i;
  }

  public int getExpectedCardinality() {
    if (expectedCardinality == -1)
      throw new Error("No estimate given for this node");
    return expectedCardinality;
  }

  /**
   * Get node's position/depth in algebra tree. Leaf nodes are at level 1.
   * Internal nodes are at
   *
   * @return node depth
   */
  public int getDepth() {
    int depth = 0;

    for (AlgebraNode node : getOperandNodes())
      if (node.getDepth() > depth)
        depth = node.getDepth();
    depth++;

    return depth;
  }

  public List<ResultSetIdentifier> buildResultSetIdentifier() {
    LinkedList<ResultSetIdentifier> rsis = new LinkedList<ResultSetIdentifier>();
    if (isPJS()){
      ResultSetIdentifier rsi = new ResultSetIdentifier();
      rsi.setCanonicalForm(canonicalForm());
      buildResultSetIdentifier(rsi);
      rsis.add(rsi);
    }else
      for (AlgebraNode operandNode : getOperandNodes())
        rsis.addAll(operandNode.buildResultSetIdentifier());

    return rsis;
  }

  protected void buildResultSetIdentifier(ResultSetIdentifier rsi) {
    //List<AlgebraNode> operands = this.getOperandNodes();
    for (AlgebraNode operand : getOperandNodes())
      operand.buildResultSetIdentifier(rsi);
  }

  /**
   * Decides if the Query represented by this tree contains
   * only select, project, join or table access operators.
   *
   * The default implementation given by this abstract class
   * simply returns false. Select, project, join and tablescan
   * operators should override this method.
   *
   * @return If the algebranode and its children represents an SPJ-Query
   */
  public boolean isPJS() {
    return false;
  }

  public int getTuplesCreated(){
    return result.getTuplesAdded();
  }
}
