package dascosa.algebra;

import dascosa.Dascosa;
import dascosa.data.*;
import dascosa.site.cache.ResultCache;
import dascosa.site.query.Executor;

import dascosa.site.query.optimizer.ResultSetIdentifier;
import dascosa.util.FilteredOutput;
import javax.sql.DataSource;
import java.util.*;

/**
 * An algebra node that replaces any cached algebra subtree and provides data from cache.
 *
 * When executing, the CacheNode will serve data from cache and pretend to be the replaced algebra node.
 */
public final class CacheNode extends AlgebraNode implements Pauseable {

  private transient ResultCache resultCache;
  private String canonicalForm;
  private ResultSetIdentifier resultSetIdentifier;
  private final String toString;
  private transient List<Packet> packets;
  private transient PacketHeader resultHeader;
  private transient Iterator<Packet> packetIterator;
  private boolean isInitalized = false;
  private boolean useAlternatePlan = false;
  private boolean logged = false;

  /**
   * Construct a new cache node based on the algebra node it is replacing.
   * The resultCache must already contain the result of replacedNode.
   *
   * @param resultCache The site's ResultCache object
   * @param replacedNode The algebra node this node is replacing
   */
  public CacheNode(ResultCache resultCache, AlgebraNode replacedNode) {
    super(replacedNode.getNodeID(), replacedNode.getOperationID(), replacedNode.getResultNodeID(0), replacedNode.getResultSite(0));
    System.out.println("Using cached version of " + replacedNode.canonicalForm());
    this.resultCache = resultCache;
    canonicalForm = replacedNode.canonicalForm();
    toString = replacedNode.toString();
    setAssignedSite(replacedNode.getAssignedSite());
    this.resultSetIdentifier = replacedNode.buildResultSetIdentifier().get(0);
    PacketHeader oldHeader = resultCache.getPacketHeader(canonicalForm);
    List<String> columnNames = new LinkedList<String>();
    for (int i = 0; i < oldHeader.getColumnCount(); i++)
      columnNames.add(oldHeader.getColumnName(i));
    resultHeader = new PacketHeader(columnNames, replacedNode.getOperationID(), replacedNode.getNodeID(), replacedNode.getResultNodeID(0));
    packets = resultCache.getPackets(canonicalForm());
    getOperandNodes().add(replacedNode);
  }

  public CacheNode(int index, AlgebraNode resultNode, ResultSetIdentifier resultSetIdentifier) {
    super(index, resultNode);
    this.resultSetIdentifier = resultSetIdentifier;
    toString = "CacheNode()";
  }

  public List<ResultSetIdentifier> buildResultSetIdentifier() {
    List<ResultSetIdentifier> list = new LinkedList<ResultSetIdentifier>();
    list.add(resultSetIdentifier);
    return list;
  }

  public void init(ResultCache resultCache) {
    isInitalized = true;
    this.resultCache = resultCache;

    if (resultCache.isCached(resultSetIdentifier.getCanonicalForm())) {

      PacketHeader oldHeader = resultCache.getPacketHeader(resultSetIdentifier.getCanonicalForm());
      List<String> columnNames = new LinkedList<String>();
      for (int i = 0; i < oldHeader.getColumnCount(); i++)
        columnNames.add(oldHeader.getColumnName(i));
      resultHeader = new PacketHeader(columnNames, getOperationID(), getNodeID(), getResultNodeID(0));
      //result = new PartialResult(1);
      //result.addPacketHeader(null, resultHeader);
      packets = resultCache.getPackets(resultSetIdentifier.getCanonicalForm());
    } else {
      Dascosa.out.println(FilteredOutput.DEBUG, "INFO: Cache is missing - Using Alternate Plan for " + resultSetIdentifier.toString());
      useAlternatePlan = true;
    }
  }

  public boolean getUseAlternatePlan() {
    return useAlternatePlan;
  }

  /**
   * Get the canonical form of the subtree represented by this node
   *
   * @return The canonical form of the replaced node
   */
  @Override
  public String canonicalForm() {
    return canonicalForm;
  }

  /**
   * Fake toString of replaced node
   *
   * @return toString of replaced node
   */
  @Override
  public String toString() {
    return toString;
  }

  /**
   * Deliver data from cache, pretending to be the replaced node
   *
   * @param executor
   * @param dataSource Not used
   */
  public boolean doOperation(Executor executor, DataSource dataSource) {

    if (result == null) {
      if (useAlternatePlan) {
        if (!logged) Dascosa.out.println(FilteredOutput.CACHELOG, "Miss! " + resultSetIdentifier.toString() + " at site " + getAssignedSite().toString());


        if (operands == null || operands[0] == null) {
          return false;
        }

        Dascosa.out.println(FilteredOutput.DEBUG, "INFO: Executing alternate plan");

        List<String> columnNames = new LinkedList<String>();
        for (int i = 0; i < operands[0].getColumnCount(); i++)
          columnNames.add(operands[0].getColumnName(i));
        resultHeader = new PacketHeader(columnNames, getOperationID(), getNodeID(), getResultNodeID(0));

        result = new PartialResult(1);
        result.addPacketHeader(executor, resultHeader);

      } else {
        result = new PartialResult(1);
        result.addPacketHeader(executor, resultHeader);

        packetIterator = packets.iterator();
        if (!logged) Dascosa.out.println(FilteredOutput.CACHELOG, "Hit! " + (resultSetIdentifier == null ? canonicalForm : resultSetIdentifier) + " at site " + getAssignedSite().toString());
      }
      if (resultSetIdentifier == null)
        resultCache.updateUsage(canonicalForm());
      else
        resultCache.updateUsage(resultSetIdentifier.getCanonicalForm());
    }

    if (useAlternatePlan) {
      packetIterator = operands[0].unprocessedPacketIterator();
      canonicalForm = resultSetIdentifier.getCanonicalForm();
    }

    logged = true;

    boolean pauseProduction = false;
    while (!pauseProduction && packetIterator.hasNext()) {
      Packet nextPacket = packetIterator.next();
      //nextPacket.removeLastPacketTag();
      pauseProduction = result.addTuplesFromPacket(executor, resultHeader, nextPacket, useAlternatePlan);
    }

    if (useAlternatePlan)
      if (operands[0].isComplete() && !operands[0].hasUnprocessedPackets() && !packetIterator.hasNext()) {
        result.lastTupleAdded(executor, resultHeader);
        Dascosa.out.println(FilteredOutput.DEBUG|FilteredOutput.CACHEINVESTMENT, "Cachenode has added last tuple, to alternate plan");
        return false;
      }

    if (!packetIterator.hasNext()) {
      return false;
    } else{
      return pauseProduction;
    }
  }

  /**
   * Return a cost estimate for executing this operator.
   * This is the true cost, not faked to be like the replaced node.
   *
   * But should it really be?
   *
   * @return Estimated cost of operator
   */
  public double getCostEstimate() {
    try {
      return result.getTupleCount();
    } catch (NullPointerException e) {
      return getOperandNodes().get(0).getCostEstimate();
    }
  }

  /**
   * Return a size estimate of the result of this operator.
   * Since the result is cached, this should be the true size.
   *
   * @return Size of result
   */
  public double getSizeEstimate() {
    try {
      return result.getColumnCount() * result.getTupleCount();
    } catch (NullPointerException e) {
      return getOperandNodes().get(0).getSizeEstimate();
    }
  }

  /**
   * Should never be run for this node type.
   *
   * @return Not used
   */
  public PartialResult simulate() {
    System.err.println("Simulating CacheNode.");
    return result;
  }

  @Override
  public boolean isPJS(){
    return getOperandNodes().get(0).isPJS();
  }
}
