package dascosa.algebra;

import dascosa.Dascosa;
import dascosa.algebra.constraints.*;
import dascosa.data.*;
import dascosa.site.query.Executor;

import dascosa.site.query.optimizer.ResultSetIdentifier;
import dascosa.util.FilteredOutput;
import javax.sql.DataSource;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 15.feb.2008
 * Time: 17:28:29
 * To change this template use File | Settings | File Templates.
 */
public final class ConstraintNode extends AlgebraNode implements PartialRestartable, Pauseable
{
  private static final double SIZE_FACTOR = 0.001;
  private final Constraint constraints[][];
  private final boolean conjunctive;

  private static int getNumberNested(Constraint constraints[][])
  {
    int number = 0;
    for (Constraint[] constraintBlock : constraints)
      for (Constraint constraint : constraintBlock)
        if (constraint instanceof NestedConstraint)
          number++;
    return number;
  }

  public ConstraintNode(int index, AlgebraNode resultNode, Constraint andBlock[])
  {
    this(index, resultNode, new Constraint[][] {andBlock}, false);
  }

  public ConstraintNode(AlgebraNode resultNode, Constraint andBlock[])
  {
    this(resultNode, new Constraint[][] {andBlock}, false);
  }

  public ConstraintNode(AlgebraNode resultNode, Constraint constraint)
  {
    this(resultNode, new Constraint[][] {new Constraint[] {constraint}}, true);
  }

  public ConstraintNode(AlgebraNode resultNode, Constraint constraints[][])
  {
    this(resultNode, constraints, true);
  }

  public ConstraintNode(AlgebraNode resultNode, Constraint constraints[][], boolean conjunctive){
    this(-1, resultNode, constraints, conjunctive);
  }

  public ConstraintNode(int index, AlgebraNode resultNode, Constraint constraints[][], boolean conjunctive)
  {
    super(index, resultNode, ConstraintNode.getNumberNested(constraints) + 1);
    this.constraints = constraints;
    this.conjunctive = conjunctive;
  }

  public String toString()
  {
    return "Constraint(" + Arrays.deepToString(constraints) + ')';
  }

  public boolean doOperation(Executor executor, DataSource dataSource)
  {
    if (operands == null)
      return false;

    for (int i = 1; i < operands.length; i++)
      if (operands[i] == null || !operands[i].isComplete()) // Nested queries must have completed
        return false;

    PacketHeader resultHeader;
    if (result == null) {
      result = new PartialResult(1, startTags);
      resultHeader = new PacketHeader(operands[0], getOperationID(), getNodeID(),
                                      getResultNodeID(0));
      result.addPacketHeader(executor, resultHeader);
      int operandNr = 1;
      for (Constraint[] orBlock : constraints)
        for (Constraint constraint : orBlock) {
          if (constraint instanceof SimpleConstraint)
            ((SimpleConstraint)constraint).init(operands[0]);
          else if (constraint instanceof NestedConstraint)
            ((NestedConstraint)constraint).init(operands[0], operands[operandNr++]);
        }
    }
    else
      resultHeader = result.getPacketHeader(0);

    boolean pauseProduction = false;
    PacketIterator iterator = operands[0].unprocessedIterator();
    while (!pauseProduction && iterator.hasNext()) {
      Object tuple[] = iterator.next();

      boolean match = false;
      if (conjunctive) {
        for (Constraint[] orBlock : constraints) {
          match = false;
          for (Constraint constraint : orBlock)
            match = match || constraint == null || constraint.evaluate(tuple);
          if (!match)
            break;
        }
      }
      else { // disjunctive
        for (Constraint[] andBlock : constraints) {
          match = true;
          for (Constraint constraint : andBlock)
            match = match && (constraint == null || constraint.evaluate(tuple));
          if (match)
            break;
        }
      }

      if (match)
        pauseProduction = result.addTuple(executor, resultHeader, tuple, iterator.getCurrentTags());
    }

    if (operands[0].isComplete() && !operands[0].hasUnprocessedPackets() && !iterator.hasNext()) {
      result.lastTupleAdded(executor, resultHeader);
      return false;
    }
    else
      return pauseProduction;
  }

  public double getCostEstimate()
  {
    double cost = 0.0;
    double size = 0.0;

    for (AlgebraNode n : getOperandNodes()) {
      cost += n.getCostEstimate();
      size += n.getSizeEstimate();
    }
    cost += size;

    return cost;
  }

  public Set<SimpleConstraint> getConjunctiveSimpleConstraintsSet() {
    if (ConstraintNode.getNumberNested(constraints) > 0)
      throw new RuntimeException("ConstraintNode: SQL not supported for nested constraints");

    Set<SimpleConstraint> queryConstraints = new HashSet<SimpleConstraint>();

    if (conjunctive || constraints.length == 1)
      for (Constraint[] orBlock : constraints) {
        if (orBlock.length > 1 && conjunctive)
          throw new RuntimeException("ConstraintNode: OR not supported");
        for (Constraint c : orBlock)
          queryConstraints.add((SimpleConstraint) c);
      }
    else
      throw new RuntimeException("ConstraintNode: Only CNF supported");

    return queryConstraints;
  }

  @Override
  public void buildResultSetIdentifier(ResultSetIdentifier rsi) {
    rsi.addConstraints(getConjunctiveSimpleConstraintsSet());
    super.buildResultSetIdentifier(rsi);
  }

  public String getSQLConstraint()
  {
    if (ConstraintNode.getNumberNested(constraints) > 0)
      throw new RuntimeException("ConstraintNode: SQL not supported for nested constraints");
    if (conjunctive) {
      return ConstraintNode.conjunctiveSQLBuilder(constraints);
    }
    else {
      return ConstraintNode.disjunctiveSQLBuilder(constraints);
    }
  }

  public static String disjunctiveSQLBuilder(Constraint[][] constraints){
    StringBuilder sb = new StringBuilder();
    for (Constraint[] andBlock : constraints) {
      if (sb.length() == 0)
        sb.append('(');
      else
        sb.append(") OR (");
      for (int i = 0; i < andBlock.length; i++) {
        if (i > 0)
          sb.append(" AND ");
        sb.append(((SimpleConstraint)andBlock[i]).getSQLConstraint());
      }
    }
    sb.append(')');
    return sb.toString();
  }

  public static String conjunctiveSQLBuilder(Constraint[][] constraints)
  {
    StringBuilder sb = new StringBuilder();
    for (Constraint[] orBlock : constraints) {
      if (sb.length() == 0)
        sb.append('(');
      else
        sb.append(") AND (");
      for (int i = 0; i < orBlock.length; i++) {
        if (i > 0)
          sb.append(" OR ");
        sb.append(((SimpleConstraint)orBlock[i]).getSQLConstraint());
      }
    }
    sb.append(')');
    return sb.toString();
  }

  public double getSizeEstimate()
  {
    double size = 0.0;

    for (AlgebraNode n : getOperandNodes()) {
      size += n.getSizeEstimate();
    }
    size *= SIZE_FACTOR;

    return size;
  }

  public PartialResult simulate()
  {
    if (operands == null)
      return null;

    for (int i = 1; i < operands.length; i++)
      if (operands[i] == null) // Nested queries must have completed
        return null;

    PartialResult simulation = new PartialResult(1);
    PacketHeader resultPart = new PacketHeader(operands[0], getOperationID(), getNodeID(),
                                               getResultNodeID(0));
    simulation.addPacketHeader(null, resultPart);

    int operandNr = 1;
    for (Constraint[] orBlock : constraints)
      for (Constraint constraint : orBlock) {
        if (constraint instanceof SimpleConstraint)
          ((SimpleConstraint)constraint).init(operands[0]);
        else if (constraint instanceof NestedConstraint)
          ((NestedConstraint)constraint).init(operands[0], operands[operandNr++]);
      }

    operands = null;
    return simulation;
  }
  @Override
  public boolean isPJS(){
    if (getNumberNested(constraints)>0)
      return false;
    for (AlgebraNode node : getOperandNodes())
      if (!node.isPJS())
        return false;
    return true;
  }
}
