package dascosa.algebra;

import dascosa.algebra.scan.TableAccessNode;
import dascosa.data.*;
import dascosa.site.query.Executor;
import dascosa.site.query.optimizer.JoinNode;

import javax.sql.DataSource;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 14.feb.2008
 * Time: 15:49:28
 * To change this template use File | Settings | File Templates.
 */
public final class EquiJoinNode extends AlgebraNode implements JoinNode
{
  private static final double SIZE_FACTOR = 0.001;
  private static final int ALTERNATE_FREQUENCY = 100;
  private final String joinAttribute1;
  private final String joinAttribute2;
  private transient List<Map<Object, List<Object[]>>> tuples;
  private transient Map<Long, Double> operandSizes;
  private transient Object[][] tempTuples;
  private transient int joinColumn[];
  private transient int activeOperand;

  public EquiJoinNode(AlgebraNode resultNode, String joinAttribute1, String joinAttribute2)
  {
    super(resultNode, 2);
    this.joinAttribute1 = joinAttribute1;
    this.joinAttribute2 = joinAttribute2;
  }

  public String toString()
  {
    return "EquiJoin( " + joinAttribute1 + '=' + joinAttribute2 + ')';
  }

  public boolean doOperation(Executor executor, DataSource dataSource)
  {
    if (operands == null || operands[1] == null)
      return false;

    if (result == null) {
      List<String> columnNames = new ArrayList<String>();
      joinColumn = new int[] {-1, -1};
      for (int i = 0; i < operands[0].getColumnCount(); i++) {
        String columnName = operands[0].getColumnName(i);
        if (columnName.equals(joinAttribute1))
          joinColumn[0] = i;
        columnNames.add(columnName);
      }
      for (int i = 0; i < operands[1].getColumnCount(); i++) {
        String columnName = operands[1].getColumnName(i);
        if (columnName.equals(joinAttribute2))
          joinColumn[1] = i;
        columnNames.add(columnName);
      }

      result = new PartialResult(1, tuplesToSkip);
      PacketHeader resultHeader = new PacketHeader(columnNames, getOperationID(), getNodeID(), getResultNodeID(0));
      result.addPacketHeader(executor, resultHeader);
      tuples = new ArrayList<Map<Object, List<Object[]>>>();
      tuples.add(new HashMap<Object, List<Object[]>>());
      tuples.add(new HashMap<Object, List<Object[]>>());
      tempTuples = new Object[2][];
    }

    PacketIterator iterators[] = new PacketIterator[] {
      operands[0].unprocessedIterator(),
      operands[1].unprocessedIterator()
    };

    while (iterators[0].hasNext() || iterators[1].hasNext()) {
      int inactiveOperand = 1 - activeOperand;
      int counter = 0;
      while (counter < ALTERNATE_FREQUENCY && iterators[activeOperand].hasNext()) { // TODO: This is not deterministic
        counter++;
        Object activeTuple[] = iterators[activeOperand].next();
        Object activeKeyValue = activeTuple[joinColumn[activeOperand]];
        Map<Object, List<Object[]>> activeMap = tuples.get(activeOperand);
        if (activeMap != null) { // Still interested in storing activeTuple for future joins
          List<Object[]> activeList = activeMap.get(activeKeyValue);
          if (activeList == null) {
            activeList = new ArrayList<Object[]>();
            activeMap.put(activeKeyValue, activeList);
          }
          activeList.add(activeTuple);
        }

        List<Object[]> inactiveList = tuples.get(inactiveOperand).get(activeKeyValue);
        if (inactiveList != null) {
          tempTuples[activeOperand] = activeTuple;
          for (Object[] inactiveTuple : inactiveList) {
            if (activeTuple[joinColumn[activeOperand]].equals(inactiveTuple[joinColumn[inactiveOperand]])) {
              tempTuples[inactiveOperand] = inactiveTuple;
              joinTuples(executor);
            }
          }
        }
      }
      activeOperand = inactiveOperand;
    }

    if (operands[0].isComplete() && !operands[0].hasUnprocessedPackets())
      tuples.set(1, null); // No need to continue to store tuples from operand 1 as no more tuples from operand 0 will arrive
    else if (operands[1].isComplete() && !operands[1].hasUnprocessedPackets())
      tuples.set(0, null);
    
    if (operands[0].isComplete() && !operands[0].hasUnprocessedPackets() &&
        operands[1].isComplete() && !operands[1].hasUnprocessedPackets()) {
      result.lastTupleAdded(executor, result.getPacketHeader(0));
      tuples = null;
    }
    return false;
  }

  public double getCostEstimate()
  {
    double cost = 0.0;
    double size = 0.0;

    for (double d : getOperandSizes().values()) {
      size += d;
    }
    for (AlgebraNode n : getOperandNodes()) {
      cost += n.getCostEstimate();
    }
    cost += size;

    return cost;
  }

  public double getSizeEstimate()
  {
    double size = 1.0;

    for (double d : getOperandSizes().values()) {
      size *= d;
    }
    size *= SIZE_FACTOR;

    return size;
  }

  public PartialResult simulate()
  {
    if (operands == null || operands[1] == null)
      return null;

    List<String> columnNames = new ArrayList<String>();
    for (int i = 0; i < operands[0].getColumnCount(); i++)
      columnNames.add(operands[0].getColumnName(i));
    for (int i = 0; i < operands[1].getColumnCount(); i++) {
      String columnName = operands[1].getColumnName(i);
      if (!columnName.equals(joinAttribute1) || !columnName.equals(joinAttribute2))
        columnNames.add(columnName);
    }

    if (operands[0].getColumn(joinAttribute1) == -1)
      throw new RuntimeException("EquiJoin: Unable to find " + joinAttribute1);
    if (operands[1].getColumn(joinAttribute2) == -1)
      throw new RuntimeException("EquiJoin: Unable to find " + joinAttribute2);

    PartialResult simulation = new PartialResult(1);
    PacketHeader resultPart = new PacketHeader(columnNames, getOperationID(), getNodeID(),
                                               getResultNodeID(0));
    simulation.addPacketHeader(null, resultPart);
    operands = null;

    return simulation;
  }

  Map<Long, Double> getOperandSizes()
  {
    if (operandSizes == null) {
      operandSizes = new HashMap<Long, Double>();
      for (AlgebraNode n : getOperandNodes()) {
        double size = operandSizes.containsKey(n.getOperationID()) ? operandSizes.get(n.getOperationID()) : 0.0;
        operandSizes.put(n.getOperationID(), size + n.getSizeEstimate());
      }
    }
    return operandSizes;
  }

  private void joinTuples(Executor executor)
  {
    Object resultTuple[] = new Object[result.getColumnCount()];
    System.arraycopy(tempTuples[0], 0, resultTuple, 0, tempTuples[0].length);
    System.arraycopy(tempTuples[1], 0, resultTuple, tempTuples[0].length, tempTuples[1].length);
    result.addTuple(executor, result.getPacketHeader(0), resultTuple);
  }

  private String findTableName(AlgebraNode operand, String joinAttribute)
  {

    if (operand instanceof TableAccessNode) {
      return ((TableAccessNode) operand).getTableName();
    } else {
      //JoinNode joinNode =  operand;
      List<AlgebraNode> liste = new LinkedList<AlgebraNode>();
      liste.addAll(operand.getOperandNodes());

      while (!liste.isEmpty()) {
        AlgebraNode current = liste.remove(0);
        if (current instanceof TableAccessNode) {
          TableAccessNode tn = (TableAccessNode) current;
          String[] attributes = tn.getAttributeNames();
          for (String attribute : attributes) {
            if (attribute.equals(joinAttribute))
              return tn.getTableName();
          }
        } else {
          liste.addAll(current.getOperandNodes());
        }
      }
    }
    throw new RuntimeException("Tablename not found for joinattribute: "+joinAttribute);
  }
//--------------------- Begin Interface JoinNode ---------------------

  public String[] getInnerRelationName()
  {
    //In a future version with better alias support this has to be reimplemented
    String joinAttribute = joinAttribute1;
    if (joinAttribute1.contains("."))
      joinAttribute = joinAttribute.substring(joinAttribute.indexOf('.')+1);
    AlgebraNode operand = getOperandNodes().get(0);
    String name = findTableName(operand, joinAttribute);
    if (name == null){
      List<AlgebraNode> nodes = getOperandNodes();
      name = findTableName(nodes.get(nodes.size()-1), joinAttribute);
    }
    return new String[] {name, joinAttribute};
  }

  public String[] getOuterRelationName()
  {
    //In a future version with better alias support this has to be reimplemented
    String joinAttribute = joinAttribute2;
    if (joinAttribute1.contains("."))
      joinAttribute = joinAttribute.substring(joinAttribute.indexOf('.')+1);

    List<AlgebraNode> nodes = getOperandNodes();
    String name = findTableName(nodes.get(nodes.size()-1), joinAttribute);
    if (name == null)
      name = findTableName(nodes.get(0), joinAttribute);
    //Dascosa.out.println(Dascosa.FilteredOutput.optimizer, name + "." + joinAttribute);
    return new String[] {name, joinAttribute};
  }

// --------------------- End Interface JoinNode ---------------------

  @Override
  public boolean isPJS(){
    for (AlgebraNode node : getOperandNodes())
      if (!node.isPJS())
        return false;
    return true;
  }
}
