/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dascosa.algebra;

import dascosa.data.PacketHeader;
import dascosa.data.PartialResult;
import dascosa.site.query.Executor;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javax.sql.DataSource;

/**
 * This class implements the node to deals with skyline of clause.
 * @author joao
 */
public class SkylineNode extends AlgebraNode {
    public static enum SkylineParams  {MIN, MAX, DIFF};
    private final String attributeNames[];
    private final SkylineParams[] skyParams;
    private final boolean distinct;

    /**
     * Stores the partial skylines
     */
    private transient List<Object[]> skyline;

    public SkylineNode(AlgebraNode resultNode, String[] attributeNames, SkylineParams[] skyParams, boolean distinct){
        super(resultNode);
        this.attributeNames = attributeNames;
        this.skyParams = skyParams;
        this.distinct = distinct;
    }

    public String toString(){
        StringBuffer buff = new StringBuffer();

        buff.append("Skyline(");
        buff.append(this.distinct?"DISTINCT ":"");
        for(int i=0;i<this.attributeNames.length;i++){
            buff.append(this.attributeNames[i]);
            buff.append(' ');
            switch(this.skyParams[i]){
                case MIN:
                    buff.append("MIN");
                    break;
                case MAX:
                    buff.append("MAX");
                    break;
                case DIFF:
                    buff.append("DIFF");
                    break;
            }
            if(i+1<this.attributeNames.length){
                buff.append(' ');
            }
        }
        buff.append(')');
        return buff.toString();
    }

    @Override
    public boolean doOperation(Executor executor, DataSource dataSource) {
        if (operands == null){
            return false;
        }

        PacketHeader resultHeader;
        if (result == null) {
            result = new PartialResult(1, tuplesToSkip);
            resultHeader = new PacketHeader(operands[0], getOperationID(), getNodeID(),
                                          getResultNodeID(0));
            result.addPacketHeader(executor, resultHeader);
            skyline = new ArrayList<Object[]>();
        }
        else{
          resultHeader = result.getPacketHeader(0);
        }
        
        int[] skyIndexes = new int[attributeNames.length];
        for (int i = 0; i < skyIndexes.length; i++){
            skyIndexes[i] = resultHeader.getColumn(attributeNames[i]);
        }
                
        for (Iterator<Object[]> iterator = operands[0].unprocessedIterator(); iterator.hasNext();){
            skylineComparison(iterator.next() , skyline, skyIndexes);
        }
        
        if(operands[0].isComplete()){
            for (Object[] tuple : skyline){
                result.addTuple(executor, resultHeader, tuple);
            }
            result.lastTupleAdded(executor, resultHeader);
          skyline = null;
        }
        return false;
    }

    /**
     * Compares a tuple against a window. All elements of the window dominated
     * by the given are removed. If there is no element in the window which
     * dominates the given tuple, it is inserted in the windows.
     * @param theTuple
     * @param window
     * @param skyIndexes
     */
    private void skylineComparison(Object[] theTuple, List<Object[]> window, int[] skyIndexes){
        for(int i=0; i< window.size(); i++){
            if(isDominated(theTuple, window.get(i), skyIndexes)){
                return;
            }else if(isDominated(window.get(i), theTuple, skyIndexes)){
                //remove the windows's tuple which is dominated by theTuple
                window.remove(i--);
            }
        }
        window.add(theTuple);
    }

    private final boolean isDominated(Object[] aTuple, Object[] another, int[] skyIndexes){

        //Chech if both tulples are equals
        boolean equals = true;
        if(aTuple.length==another.length){
            for(int i=0; i< skyIndexes.length; i++){
                int index = skyIndexes[i];

                int res = ((Comparable)aTuple[index]).compareTo(another[index]);

                switch(this.skyParams[i]){
                    case MIN:
                        if(res<0){
                            return false;
                        } else if(res!=0){
                            equals = false;
                        }
                        break;
                    case MAX:
                        if(res>0){
                            return false;
                        } else if(res!=0){
                            equals = false;
                        }

                        break;
                    case DIFF:
                        if(res!=0){
                            return false;
                        } 

                        break;
                }
                
            }
        }else{
            throw new RuntimeException("The tuples are not comparable!");
        }

        //if the distinct param was not provided and the tuples are equals.
        if(!distinct && equals){
            return false;
        }
        return true;
    }

  public double getCostEstimate()
  {
    double cost = 0.0;
    double size = 0.0;

    for (AlgebraNode n : getOperandNodes()) {
      cost += n.getCostEstimate();
      size += n.getSizeEstimate();
    }
    cost += size * Math.log(cost);

    return cost;
  }

  public double getSizeEstimate()
  {
    double size = 0.0;

    for (AlgebraNode n : getOperandNodes()) {
      size += n.getSizeEstimate();
    }

    return size;
  }

  public PartialResult simulate()
  {
    if (operands == null){
      return null;
    }
  
    for (String attributeName : attributeNames){
        if (operands[0].getColumn(attributeName) == -1){
            throw new RuntimeException("Skyline: Unable to find " + attributeName);
        }
    }
   
    PartialResult simulation = new PartialResult(1);
    PacketHeader resultPart = new PacketHeader(operands[0], getOperationID(), getNodeID(),
                                               getResultNodeID(0));
    simulation.addPacketHeader(null, resultPart);

    operands = null;
    return simulation;
  }

}
