package dascosa.data;

import dascosa.site.query.Executor;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 20.aug.2007
 * Time: 18:15:59
 * To change this template use File | Settings | File Templates.
 */
public final class PartialResult {

  private final Packet activePackets[]; // One for each header, packets begun but not sent
  private final PacketHeader headers[];
  private final int tuplesToSkip;
  private final boolean[] resultCompleted; // One for each header
  private int packetSize;
  private int startTags[];
  private int tupleCounter;
  private int packetsSent = 0;
  private static final int sizeFactor = 190000; //based on serialized size

  public PartialResult(int numberOfHeaders) // AlgebraNode(s)
  {
    this(numberOfHeaders, 0);
  }

  public PartialResult(int numberOfHeaders, int startTags[]) // AlgebraNode(s)
  {
    this(numberOfHeaders, 0);
    this.startTags = startTags;
  }

  public PartialResult(int numberOfHeaders, int tuplesToSkip) // AlgebraNode(s)
  {
    this.tuplesToSkip = tuplesToSkip;
    headers = new PacketHeader[numberOfHeaders];
    activePackets = new Packet[numberOfHeaders];
    resultCompleted = new boolean[numberOfHeaders];
    packetSize = Packet.MAX_TUPLES_IN_PACKET - (tuplesToSkip % Packet.MAX_TUPLES_IN_PACKET);
  }

  public void addPacketHeader(Executor executor, PacketHeader header) // AlgebraNode(s)
  {
    for (int i = 0; i < headers.length; i++)
      if (headers[i] == null) {
        headers[i] = header;
        break;
      }
    if (executor != null)
      executor.sendHeader(this, header);
  }

  public boolean addTuple(Executor executor, PacketHeader header, Object tuple[]) // AlgebraNode(s)
  {
    return addTuple(executor, header, tuple, null);
  }

  public boolean addTuple(Executor executor, PacketHeader header, Object tuple[], int tags[]) // ConstraintNode + this
  {
    tupleCounter++;
    if (tupleCounter <= tuplesToSkip)
      return false;
    int index = getPacketHeaderNumber(header);
    Packet activePacket = activePackets[index];
    if (activePacket == null) {
      activePacket = new Packet(header);
      activePackets[index] = activePacket;
    }
    activePacket.addTuple(tuple);
    if (activePacket.getTupleCount() >= packetSize) {
      if (tags != null) {
        int tagCopy[] = new int[tags.length];
        System.arraycopy(tags, 0, tagCopy, 0, tags.length);
        if (startTags != null)
          for (int i = 0; i < tags.length; i++)
            tags[i] += startTags[i];
        activePacket.setTags(tagCopy);
      }
      executor.sendPacket(this, header, activePacket);
      packetsSent++;
      activePackets[index] = null;
      packetSize = Packet.MAX_TUPLES_IN_PACKET;
      return activePacket.isReceiptRequested();
    } else
      return false;
  }

  public boolean addTuplesFromPacket(Executor executor, PacketHeader header, Packet sourcePacket) {
    return addTuplesFromPacket(executor, header, sourcePacket, false);
  }

  public boolean addTuplesFromPacket(Executor executor, PacketHeader header, Packet sourcePacket, boolean hasMultipleSources) // Cache & CrashNode
  {
    tupleCounter += sourcePacket.getTupleCount();
    int index = getPacketHeaderNumber(header);
    Packet packet = new Packet(header, sourcePacket, !hasMultipleSources);
    Packet lastPacket = activePackets[index];
    if (lastPacket != null)
      executor.sendPacket(this, header, lastPacket);
    executor.sendPacket(this, header, packet);
    activePackets[index] = null;
    if (sourcePacket.isLastPacket() && !hasMultipleSources)
      resultCompleted[index] = true;
    return packet.isReceiptRequested();
  }

  public void clear() // AlgebraNode
  {
    if (activePackets == null)
      return;
    for (int i = 0; i < activePackets.length; i++) {
      activePackets[i] = null;
      resultCompleted[i] = false;
    }
  }

  public int getColumnCount() // AlgebraNode(s)
  {
    return headers[0].getColumnCount();
  }

  public String getColumnName(int col) // ProjectNode
  {
    return headers[0].getColumnName(col);
  }

  public PacketHeader getPacketHeader(int number) // QueryPlanner & AlgebraNode(s)
  {
    return headers[number];
  }

  public int getPacketHeaderCount() // QueryPlanner
  {
    return headers.length;
  }

  public int getPacketHeaderNumber(PacketHeader header) // Executor & this
  {
    for (int i = 0; i < headers.length; i++)
      if (headers[i] == header)
        return i;
    return -1;
  }

  public int getTuplesAdded(){
    return tupleCounter;
  }

  public int getTupleCount() // MainCommand, CacheNode, SQLPanel
  {
    int count = 0;
    for (PacketHeader header : headers)
      for (int i = 0; i < header.getPacketCount(); i++)
        count += header.getPacket(i).getTupleCount();
    return count;
  }

  public boolean hasCompleted() // Executor
  {
    for (boolean headerCompleted : resultCompleted)
      if (!headerCompleted)
        return false;
    return true;
//    for (Packet packet : activePackets)
//      if (packet == null || !packet.isLastPacket())
//        return false;
//    return true;
  }

  public void lastTupleAdded(Executor executor, PacketHeader header) // AlgebraNode(s)
  {
    int index = getPacketHeaderNumber(header);
    Packet lastPacket = activePackets[index];
    if (lastPacket == null) { // No tuples produced
      lastPacket = new Packet(header);
      activePackets[index] = lastPacket;
    }
    lastPacket.setLastPacket();
    resultCompleted[index] = true;
    executor.sendPacket(this, header, lastPacket);
  }

  public int getSize() {
    if (hasCompleted())
      return packetsSent * sizeFactor;
    else
      throw new RuntimeException("Asking for result size before it has completed");
  }
}
