package dascosa.site;

import dascosa.message.flood.TestRunnerRequestMsg;
import dascosa.Dascosa;
import dascosa.algebra.*;
import dascosa.data.Packet;
import dascosa.data.PacketHeader;
import dascosa.message.*;
import dascosa.message.flood.AbstractFloodRequestMsg;
import dascosa.message.flood.AbstractFloodResponseMsg;
import dascosa.message.request.*;
import dascosa.message.restart.*;
import dascosa.site.cache.ResultCache;
import dascosa.site.fault.FaultDetector;
import dascosa.site.fault.RestartCache;
import dascosa.site.index.Index;
import dascosa.site.localdb.*;
import dascosa.site.query.QueryHandler;
import dascosa.site.query.optimizer.DataStatistics;
import dascosa.site.update.*;
import dascosa.util.FilteredOutput;
import rice.p2p.commonapi.Message;
import rice.p2p.commonapi.NodeHandle;

import javax.sql.DataSource;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 04.okt.2007
 * Time: 14:06:40
 * To change this template use File | Settings | File Templates.
 */
final class MessageHandler
{
  private static final int MAX_UNPROCESSED_MESSAGES = 10;
  private final CommitManager commitManager;
  private final FaultDetector faultDetector;
  private final FloodManager floodManager;
  private final FragmentCopier fragmentCopier;
  private final FragmentSender fragmentSender;
  private final FragmentSplitter fragmentSplitter;
  private final Index index;
  private final InsertHandler insertHandler;
  private final List<Message> unprocessedMessagesLow = new LinkedList<Message>();
  private final List<Message> unprocessedMessagesMedium = new LinkedList<Message>();
  private final LocalFragmentHandler localFragmentHandler;
  private final LockManager lockManager;
  private final Map<PacketReceiptMsg, NodeHandle> waitingForReceipt = new HashMap<PacketReceiptMsg, NodeHandle>();
  private final NetworkLayer networkLayer;
  private final RestartCache restartCache;
  private final QueryHandler queryHandler;
  private final RequestResponseManager requestResponseManager;
  private final ResultCache resultCache;
  private final SiteManager siteManager;

  MessageHandler(SiteManager siteManager, NetworkLayer networkLayer, RequestResponseManager requestResponseManager,
                 Index index, QueryHandler queryHandler, FragmentCopier fragmentCopier,
                 CommitManager commitManager, InsertHandler insertHandler, FloodManager floodManager,
                 LocalFragmentHandler localFragmentHandler, FragmentSplitter fragmentSplitter,
                 FragmentSender fragmentSender, LockManager lockManager, ResultCache resultCache)
  {
    this.siteManager = siteManager;
    this.networkLayer = networkLayer;
    this.requestResponseManager = requestResponseManager;
    this.index = index;
    this.queryHandler = queryHandler;
    this.fragmentCopier = fragmentCopier;
    this.commitManager = commitManager;
    this.insertHandler = insertHandler;
    this.floodManager = floodManager;
    this.localFragmentHandler = localFragmentHandler;
    this.fragmentSplitter = fragmentSplitter;
    this.fragmentSender = fragmentSender;
    this.lockManager = lockManager;
    this.resultCache = resultCache;
    faultDetector = queryHandler.getFaultDetector();
    restartCache = queryHandler.getPacketCache();
  }

  void messageReceived(Message message) // Pastry thread (network) or MessageHandler thread (local)
  {
    if (message.getPriority() == Message.LOW_PRIORITY) {
      synchronized (unprocessedMessagesLow) {
        unprocessedMessagesLow.add(message);
        unprocessedMessagesLow.notifyAll();
      }
    }
    else {
      synchronized (unprocessedMessagesMedium) {
        unprocessedMessagesMedium.add(message);
        unprocessedMessagesMedium.notifyAll();
      }
    }
  }

  void removeMessages()
  {
    synchronized (unprocessedMessagesLow) {
      unprocessedMessagesLow.clear();
    }
    synchronized (unprocessedMessagesMedium) {
      unprocessedMessagesMedium.clear();
    }
    synchronized (waitingForReceipt) {
      waitingForReceipt.clear();
    }
  }

  Thread startThread(DataSource dataSource, int priority) 
  {
    Thread thread = new Thread(new WorkerThread(dataSource, priority));
    thread.start();
    return thread;
  }

// -------------------------- INNER CLASSES --------------------------

  private final class WorkerThread implements Runnable
  {
    private final DataSource dataSource;
    private final List<Message> queue;

    private WorkerThread(DataSource dataSource, int priority)
    {
      this.dataSource = dataSource;
      if (priority == Message.LOW_PRIORITY)
        queue = unprocessedMessagesLow;
      else
        queue = unprocessedMessagesMedium;
    }

    public void run()
    {
      while (true) {
        Message message;
        synchronized (queue) {
          while (queue.isEmpty()) {
            try {
              queue.wait();
            } catch (InterruptedException e) {
            }
          }
          message = queue.remove(0);
        }
        if (queue == unprocessedMessagesLow && queue.size() < MAX_UNPROCESSED_MESSAGES / 2) {
          synchronized (waitingForReceipt) {
            for (PacketReceiptMsg msg: waitingForReceipt.keySet()) {
              networkLayer.sendMessage(msg, waitingForReceipt.get(msg));
//              System.out.println("MessageHandler$WorkerThread.run: Sending " + msg);
            }
            waitingForReceipt.clear();
          }
        }
        processMessage(message);
      }
    }

    private void processMessage(Message message)
    {
      if (message instanceof PyxidaMessage)
        siteManager.getPyxidaHandler().processMessage((PyxidaMessage) message);
      else if (message instanceof HistogramsRequestMsg) {
        HistogramsRequestMsg msg = (HistogramsRequestMsg)message;
        Dascosa.out.println(FilteredOutput.DATA_STATISTICS, "Received histogramRequest with: "+msg.identifier.toString());
        if(msg.requestCreation)
          localFragmentHandler.processHistogramMsg(msg, dataSource);
        else
          index.doLookup(msg);
      }
      else if (message instanceof CoordinateRequestMsg) {
        CoordinateRequestMsg request = (CoordinateRequestMsg)message;
        CoordinateResponseMsg response = new CoordinateResponseMsg(request, siteManager.getLocalCoords());
        networkLayer.sendMessage(response, request.getReturnToSite());
      }
      else if (message instanceof HistogramResponseMessage)
    	  DataStatistics.getInstance().responseReceived((HistogramResponseMessage)message, null, dataSource);
      else if (message instanceof AbstractResponseMsg)
        requestResponseManager.responseReceived((AbstractResponseMsg)message, dataSource);
      else if (message instanceof AbstractFloodRequestMsg)
        floodManager.requestReceived((AbstractFloodRequestMsg)message);
      else if (message instanceof AbstractFloodResponseMsg)
        floodManager.responseReceived((AbstractFloodResponseMsg)message);
      else if (message instanceof PublishTableMsg)
        index.addEntry((PublishTableMsg)message);
      else if (message instanceof PublishCacheMsg)
        index.addCacheEntry((PublishCacheMsg)message);
      else if (message instanceof PublishResultSetUsageMsg)
        index.cacheInvestment.processMsg((PublishResultSetUsageMsg)message);
      else if (message instanceof MasterNotificationMsg)
        localFragmentHandler.processMasterMsg((MasterNotificationMsg)message, lockManager, dataSource);
      else if (message instanceof LookupRequestMsg)
        index.doLookup((LookupRequestMsg)message);
      else if (message instanceof LookupPKRequestMsg)
        index.doLookup((LookupPKRequestMsg)message);
      else if (message instanceof QueryCompletedMsg)
        queryHandler.queryCleanup((QueryCompletedMsg)message);
      else if (message instanceof QueryStatsMsg)
        queryHandler.queryStatsReceived((QueryStatsMsg)message);
      else if (message instanceof AliveRequestMsg) {
        if (siteManager.isCrashed())
          Dascosa.out.println("Alive check: Dead");
        else {
          AliveRequestMsg requestMsg = (AliveRequestMsg)message;
          networkLayer.sendMessage(new AliveResponseMsg(requestMsg), requestMsg.getReturnToSite());
        }
      }
      else if (message instanceof RestartSiteMsg)
        siteManager.restartNode();
      else if (message instanceof CrashNotifiyMsg)
        siteManager.stopMessageSending(((CrashNotifiyMsg)message).getAlgebraNodeID());

      else if (!siteManager.isCrashed()) { // These messages should not be processed if site is crashed
        if (message instanceof AlgebraNode) {
          if (message instanceof CacheNode) {
            CacheNode cacheNode = (CacheNode)message;
            cacheNode.init(resultCache);
            queryHandler.algebraNodeReceived((AlgebraNode) message, dataSource);

            if (cacheNode.getUseAlternatePlan()) {
              for (AlgebraNode algebraNode : ((AlgebraNode) message).getOperandNodes()) {
                networkLayer.sendMessage(algebraNode, algebraNode.getAssignedSite());
                if (algebraNode.getAssignedSite() != siteManager.getLocalSite())
                  faultDetector.addWatch(algebraNode);
              }
            }

          }
          else if (Dascosa.isPropertySet(Dascosa.USE_SEMANTIC_CACHE) &&
              resultCache.isCached(((AlgebraNode)message).canonicalForm())) {
            // Spørringen er cachet
            CacheNode cacheNode = new CacheNode(resultCache, (AlgebraNode)message);
            resultCache.addCacheNode(((AlgebraNode)message).getNodeID(), cacheNode);
            queryHandler.algebraNodeReceived(cacheNode, dataSource);
          }
          else {
            queryHandler.algebraNodeReceived((AlgebraNode)message, dataSource);
            for (AlgebraNode algebraNode : ((AlgebraNode)message).getOperandNodes()) {
              networkLayer.sendMessage(algebraNode, algebraNode.getAssignedSite());
              if (algebraNode.getAssignedSite() != siteManager.getLocalSite())
                faultDetector.addWatch(algebraNode);
            }
          }
        }
        else if (message instanceof PacketHeader) {
          PacketHeader msg = (PacketHeader)message;
          AlgebraNode algebraNode = queryHandler.getOperandAlgebraNode(msg.getSourceAlgebraNodeID());
          if (algebraNode != null) {
            faultDetector.messageReceivedFrom(algebraNode.getAssignedSite());
            queryHandler.operandHeaderReceived(msg);
          }
          else {
            System.err.println("NPE(PacketHeader) - Can't find algebra node for " + msg.getSourceAlgebraNodeID());
          }
        }
        else if (message instanceof Packet) {
          Packet msg = (Packet)message;
          AlgebraNode algebraNode = queryHandler.getOperandAlgebraNode(msg.getSourceAlgebraNodeID());
          if (algebraNode != null) {
            faultDetector.messageReceivedFrom(algebraNode.getAssignedSite()); 
            if (msg.isLastPacket())  // Don't care about this algebra node anymore
              faultDetector.removeWatch(msg.getSourceAlgebraNodeID());
            if (msg.isReceiptRequested() && algebraNode.getAssignedSite() != siteManager.getLocalSite()) {
              PacketReceiptMsg packetReceiptMsg = new PacketReceiptMsg(msg.getSourceAlgebraNodeID());
              if (unprocessedMessagesLow.size() < MAX_UNPROCESSED_MESSAGES)
                networkLayer.sendMessage(packetReceiptMsg, algebraNode.getAssignedSite());
              else {
                synchronized (waitingForReceipt) {
//                  System.out.println("MessageHandler$WorkerThread.processMessage: Not sending " + packetReceiptMsg);
                  waitingForReceipt.put(packetReceiptMsg, algebraNode.getAssignedSite());
                }
              }
            }
            queryHandler.operandPacketReceived(msg, dataSource);
          }
          else {
            System.err.println("NPE (Packet) - Can't find algebra node for " + msg.getSourceAlgebraNodeID());
          }
        }
        else if (message instanceof PartialRestartFailedMsg) { // Stateless node crashed
          PartialRestartFailedMsg msg = (PartialRestartFailedMsg)message;
          AlgebraNode algebraNode = msg.getAlgebraNode();
          List<AlgebraNode> operandNodes = algebraNode.getOperandNodes();
          int[] tags = msg.getPacketTags();
          if (tags == null || tags.length == 0)
            tags = new int[operandNodes.size()];
          algebraNode.setStartTags(tags);
          Dascosa.out.println("Restart " + algebraNode + " on a new site");
          queryHandler.algebraNodeReceived(algebraNode, dataSource);
          for (int i = 0; i < operandNodes.size(); i++) {
            AlgebraNode operandNode = operandNodes.get(i);
            networkLayer.sendMessage(new PartialRestartOperandMsg(operandNode, tags[i]), operandNode.getAssignedSite());
            if (operandNode.getAssignedSite() != siteManager.getLocalSite())
              faultDetector.addWatch(operandNode);
          }
        }
        else if (message instanceof PartialRestartOperandMsg) {
          PartialRestartOperandMsg msg = (PartialRestartOperandMsg)message;
          AlgebraNode algebraNode = msg.getAlgebraNode();
          if (!restartCache.lookup(algebraNode, msg.getTuplesToSkip())) { // No cache hit
            algebraNode.setTuplesToSkip(msg.getTuplesToSkip());
            Dascosa.out.println("Restart " + algebraNode + ", skipping first " + msg.getTuplesToSkip());
            queryHandler.algebraNodeReceived(algebraNode, dataSource);
          }
        }
        else if (message instanceof CompleteRestartFailedMsg) { // Stateful node crashed
          CompleteRestartFailedMsg msg = (CompleteRestartFailedMsg)message;
          AlgebraNode algebraNode = msg.getAlgebraNode();
          algebraNode.setTuplesToSkip(msg.getTuplesToSkip());
          Dascosa.out.println("Restart " + algebraNode + " on a new site, skipping first " + msg.getTuplesToSkip());
          queryHandler.algebraNodeReceived(algebraNode, dataSource);
          for (AlgebraNode operandNode : algebraNode.getOperandNodes()) {
            networkLayer.sendMessage(new CompleteRestartOperandMsg(operandNode), operandNode.getAssignedSite());
            if (operandNode.getAssignedSite() != siteManager.getLocalSite())
              faultDetector.addWatch(operandNode);
          }
        }
        else if (message instanceof CompleteRestartOperandMsg) {
          CompleteRestartOperandMsg msg = (CompleteRestartOperandMsg)message;
          if (!restartCache.lookup(msg.getAlgebraNode(), 0)) { // No cache hit
            Dascosa.out.println("Restart " + msg.getAlgebraNode());
            queryHandler.algebraNodeReceived(msg.getAlgebraNode(), dataSource);
            for (AlgebraNode algebraNode : msg.getAlgebraNode().getOperandNodes()) {
              networkLayer.sendMessage(algebraNode, algebraNode.getAssignedSite());
              if (algebraNode.getAssignedSite() != siteManager.getLocalSite())
                faultDetector.addWatch(algebraNode);
            }
          }
        }
        else if (message instanceof FragmentCopyRequestMsg)
          fragmentCopier.sendCopy((FragmentCopyRequestMsg)message, dataSource);
        else if (message instanceof FragmentSplitRequestMsg)
          fragmentSplitter.splitRequested((FragmentSplitRequestMsg)message);
        else if (message instanceof FragmentReceivedMsg)
          fragmentSplitter.fragmentReceived((FragmentReceivedMsg)message, dataSource);
        else if (message instanceof DeleteReplicateRequestMsg)
          localFragmentHandler.processDeletionMsg((DeleteReplicateRequestMsg)message, dataSource);
        else if (message instanceof StatementRequestMsg)
          commitManager.statementRequested((StatementRequestMsg)message, dataSource);
        else if (message instanceof CommitRequestMsg)
          commitManager.executeCommit((CommitRequestMsg)message);
        else if (message instanceof InsertRequestMsg)
          insertHandler.insertRequested((InsertRequestMsg)message, dataSource);
        else if (message instanceof LockRequestMsg)
          lockManager.lockRequested((LockRequestMsg)message);
        else if (message instanceof FragmentSendRequestMsg)
          fragmentSender.copyReceived((FragmentSendRequestMsg)message, dataSource);
        else if (message instanceof PacketReceiptMsg)
          queryHandler.receiptReceived(((PacketReceiptMsg)message).getSourceAlgebraNodeID(), dataSource);
        else if (message instanceof TestCompleteMsg) {
          siteManager.getTestRunner().receiveMessage((TestCompleteMsg)message);
        }
      }
    }
  }
}
