package dascosa.site;

import dascosa.message.ComplexMessage;
import dascosa.Dascosa;
import rice.p2p.commonapi.*;
import rice.pastry.socket.SocketNodeHandle;

import java.io.*;
import java.net.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 23.apr.2007
 * Time: 10:09:35
 * To change this template use File | Settings | File Templates.
 */
public final class NetworkLayer implements Application, Runnable
{
  private static final int MAX_QUEUE = 30;
  private final Collection<MessageListener> messageListeners = new ArrayList<MessageListener>();
  private final IdFactory idFactory;
  private final MessageNotification notification = new MessageNotification();
  private Endpoint endpoint;
  private MessageHandler messageHandler;
  private int serverPort;
  private AtomicLong queueCounter = new AtomicLong(0);

  private static void sendMessageSocket(Message message, InetSocketAddress targetSite)
  {
    synchronized (targetSite) {
      try {
        Socket socket = new Socket(targetSite.getAddress(), targetSite.getPort() * 2);
//      System.out.println("Sending: " + message + " to " + socket);
        ObjectOutputStream output = new ObjectOutputStream(new BufferedOutputStream(socket.getOutputStream()));
        output.writeObject(message);
        output.close();
        socket.close();
      } catch (IOException e) {
        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      }
    }
  }

  NetworkLayer(IdFactory idFactory)
  {
    this.idFactory = idFactory;
  }

// --------------------- Begin Interface Application ---------------------

  public boolean forward(RouteMessage message)
  {
    return true;
  }

  public void deliver(Id id, Message message)
  {
    if (id == null)
      for (MessageListener messageListener : messageListeners)
        messageListener.messageDirect(message);
    else
      for (MessageListener messageListener : messageListeners)
        messageListener.messageReceived(message);
    messageHandler.messageReceived(message);
  }

  public void update(NodeHandle handle, boolean joined)
  {
  }

// --------------------- End Interface Application ---------------------


// --------------------- Begin Interface Runnable ---------------------

  public void run()
  {
    try {
      ServerSocket server = new ServerSocket(serverPort);
      while (true) {
        Socket socketToClient = server.accept();
        ObjectInputStream input = new ObjectInputStream(new BufferedInputStream(socketToClient.getInputStream()));
        try {
          Message message = (Message)input.readObject();
//          System.out.println("Message received: " + message + " from " + socketToClient);
          for (MessageListener messageListener : messageListeners)
            messageListener.messageReceived(message);
          messageHandler.messageReceived(message);
        } catch (ClassNotFoundException e) {
          e.printStackTrace();
        }
        input.close();
        socketToClient.close();
      }
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

// --------------------- End Interface Runnable ---------------------

  public void addMessageListener(MessageListener messageListener)
  {
    messageListeners.add(messageListener);
  }

  public void sendMessage(Message message, NodeHandle targetSite)
  {
    if (message instanceof ComplexMessage)
      ((ComplexMessage)message).prepareMessage();
    if (targetSite == endpoint.getLocalNodeHandle())
      deliver(null, message); // Direct delivery
    else {
      for (MessageListener messageListener : messageListeners)
        messageListener.messageSending(message);
      if (!Dascosa.isPropertySet(Dascosa.USE_PASTRY_COMMUNICATION) &&
          targetSite instanceof SocketNodeHandle) {
        SocketNodeHandle socketNodeHandle = (SocketNodeHandle)targetSite;
        NetworkLayer.sendMessageSocket(message, socketNodeHandle.getInetSocketAddress());
      }
      else {
        queueCounter.incrementAndGet();
        while (message.getPriority() == Message.LOW_PRIORITY && hasQueueOverload()) {
          try {
            synchronized (this) {
              wait();
            }
          } catch (InterruptedException e) {
          }
        }
        endpoint.route(null, message, targetSite, notification);
      }
    }
  }

  public void sendMessage(Message message, String string)
  {
    Id id = idFactory.buildId(string.toLowerCase());
    for (MessageListener messageListener : messageListeners)
      messageListener.messageSending(message);
    queueCounter.incrementAndGet();
    while (message.getPriority() == Message.LOW_PRIORITY && hasQueueOverload()) {
      try {
        synchronized (this) {
          wait();
        }
      } catch (InterruptedException e) {
      }
    }
    endpoint.route(id, message, null, notification);
  }

  void connect(Node node, MessageHandler messageHandler)
  {
    this.messageHandler = messageHandler;
    endpoint = node.buildEndpoint(this, "dascosa");
    endpoint.register();
    if (!Dascosa.isPropertySet(Dascosa.USE_PASTRY_COMMUNICATION) &&
        endpoint.getLocalNodeHandle() instanceof SocketNodeHandle) {
      SocketNodeHandle socketNodeHandle = (SocketNodeHandle)endpoint.getLocalNodeHandle();
      serverPort = socketNodeHandle.getInetSocketAddress().getPort() * 2;
      new Thread(this).start();
    }
  }

  private boolean hasQueueOverload()
  {
    return queueCounter.get() - notification.sentCounter.get() > MAX_QUEUE;
  }

  private synchronized void messageSent()
  {
    notifyAll();
  }

// -------------------------- INNER CLASSES --------------------------

  private final class MessageNotification implements DeliveryNotification
  {
    private AtomicLong sentCounter = new AtomicLong(0);

    public void sent(MessageReceipt msg)
    {
      sentCounter.incrementAndGet();
      messageSent();
    }

    public void sendFailed(MessageReceipt msg, Exception reason)
    {
      sentCounter.incrementAndGet();
      System.err.println("Send failed: " + msg.toString());
      System.err.println("Reason:      " + reason);
    }
  }
}
