package dascosa.site;

import dascosa.Dascosa;
//import dascosa.util.ParetoRandom;
import dascosa.algebra.*;
import dascosa.message.flood.*;
import dascosa.message.request.*;
import dascosa.message.restart.CrashNotifiyMsg;
import dascosa.parser.*;
import dascosa.site.cache.CacheTester;
import dascosa.site.cache.ResultCache;
import dascosa.site.dynfrag.DynamicAlgorithm;
import dascosa.site.fault.CacheListener;
import dascosa.site.fault.RestartCache;
import dascosa.site.index.Index;
import dascosa.site.index.IndexListener;
import dascosa.site.localdb.*;
import dascosa.site.query.QueryHandler;
import dascosa.site.query.QueryListener;
import dascosa.site.query.optimizer.DataStatistics;
import dascosa.site.test.CrashTester;
import dascosa.site.test.TestRunner;
import dascosa.site.tpch.TpchTester;
import dascosa.site.update.*;
import dascosa.util.FilteredOutput;
import dascosa.util.ParetoRandom;
import edu.harvard.syrah.pyxida.nc.lib.Coordinate;
import edu.harvard.syrah.pyxida.nc.lib.NCClient;
import heggland.probability.ExponentialDistribution;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.derby.jdbc.EmbeddedConnectionPoolDataSource;


import rice.p2p.commonapi.*;
import rice.pastry.PastryNode;
import rice.pastry.socket.SocketPastryNodeFactory;

import javax.sql.DataSource;
import java.io.*;
import java.net.InetSocketAddress;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 20.apr.2007
 * Time: 14:49:18
 * To change this template use File | Settings | File Templates.
 */
public final class SiteManager
{
  public static final int BINDPORT = 9001;
  public static final int BOOTPORT = 9001;
  public static final int NUM_DIMS = 5;
  private final Collection<DebugEventListener> debugEventListeners = new ArrayList<DebugEventListener>();
  private final CopyLookupClient copyLookupClient = new CopyLookupClient();
  private final CreateHandler createHandler;
  private final FloodClient setPropertyClient = new SetPropertyClient();
  private final FloodClient tableListClient = new TableListClient();
  private final FloodClient setCacheParametersClient = new SetCacheParametersClient();
  private final FloodManager floodManager;
  private final FragmentCopier fragmentCopier;
  private final FragmentSplitter fragmentSplitter;
  private final InetSocketAddress bootAddress;
  private final InsertHandler insertHandler;
  private final LocalFragmentHandler localFragmentHandler;
  private final MessageHandler messageHandler;
  private final NCClient ncClient;
  private final NetworkLayer networkLayer;
  private final PyxidaHandler pyxidaHandler;
  private final TestRunner testRunner;
  private final QueryHandler queryHandler;
  private final RequestResponseManager requestResponseManager;
  private final SocketPastryNodeFactory factory;
  private final String database;
  private final UpdateHandler updateHandler;
  private PastryNode node;
  private boolean isCrashed;
  private final ExponentialDistribution distribution = new ExponentialDistribution(3000);
  private Set<TestListener> testListeners = new HashSet<TestListener>();
  private AtomicInteger waiting = new AtomicInteger(0);

  static {
    new SqlScript(new StringReader("SELECT * FROM A;"));
  }

  public SiteManager(String database, SocketPastryNodeFactory factory,
                     InetSocketAddress bootAddress, IdFactory idFactory,
                     DebugEventListener debugEventListener,
                     MessageListener messageListener,
                     DatabaseListener databaseListener,
                     IndexListener indexListener,
                     List<CacheListener> cacheListeners,
                     QueryListener queryListener) throws Exception
  {
    this.database = database;
    this.factory = factory;
    this.bootAddress = bootAddress;

    ncClient = new NCClient(NUM_DIMS);

    if (debugEventListener != null)
      addDebugEventListener(debugEventListener);

    EmbeddedConnectionPoolDataSource dataSource = new EmbeddedConnectionPoolDataSource();
    dataSource.setDatabaseName(database);

    networkLayer = new NetworkLayer(idFactory);
    if (messageListener != null)
      networkLayer.addMessageListener(messageListener);
    requestResponseManager = new RequestResponseManager(this, networkLayer);

    InvalidateReplicatesHandler invalidateReplicatesHandler =
      new InvalidateReplicatesHandler(requestResponseManager, this);

    CommitManager commitManager = new CommitManager(this, networkLayer, requestResponseManager,
                                                    invalidateReplicatesHandler, dataSource);
    localFragmentHandler = new LocalFragmentHandler(this, networkLayer, databaseListener, dataSource);
    fragmentCopier = new FragmentCopier(networkLayer, requestResponseManager, localFragmentHandler);

    fragmentSplitter = new FragmentSplitter(dataSource, requestResponseManager, this,
                                            invalidateReplicatesHandler, localFragmentHandler,
                                            networkLayer);

    Index index = new Index(networkLayer, this);
    if (indexListener != null)
      index.addListener(indexListener);
    ResultCache resultCache = new ResultCache(this, networkLayer);
    if (cacheListeners != null) {
      for (CacheListener cacheListener: cacheListeners)
        resultCache.addCacheListener(cacheListener);
    }
    floodManager = new FloodManager(this, networkLayer, index, resultCache);

    insertHandler = new InsertHandler(this, networkLayer, requestResponseManager,
                                      invalidateReplicatesHandler, dataSource, localFragmentHandler);
    RestartCache restartCache = new RestartCache(networkLayer);
    if (cacheListeners != null) {
      for (CacheListener cacheListener: cacheListeners)
        restartCache.addCacheListener(cacheListener);
    }
    queryHandler = new QueryHandler(this, networkLayer, requestResponseManager, restartCache, resultCache);
    if (queryListener != null)
      queryHandler.addQueryListener(queryListener);
    queryHandler.addQueryListener(localFragmentHandler);

    LockManager lockManager = new LockManager(localFragmentHandler, networkLayer);

    RefragmentHandler refragmentHandler = new RefragmentHandler(requestResponseManager,
                                                                invalidateReplicatesHandler,
                                                                localFragmentHandler, networkLayer,
                                                                this, lockManager);

    messageHandler = new MessageHandler(this, networkLayer, requestResponseManager, index,
                                        queryHandler, fragmentCopier, commitManager, insertHandler,
                                        floodManager, localFragmentHandler, fragmentSplitter,
                                        refragmentHandler.getFragmentSender(), lockManager, resultCache);

    createHandler = new CreateHandler(this, localFragmentHandler, requestResponseManager);
    updateHandler = new UpdateHandler(commitManager, this, requestResponseManager);

    messageHandler.startThread(dataSource, Message.LOW_PRIORITY);
    messageHandler.startThread(dataSource, Message.MEDIUM_PRIORITY);

    bootNode();

    new DynamicAlgorithm(getLocalSite(), localFragmentHandler, refragmentHandler);

    pyxidaHandler = new PyxidaHandler(networkLayer, this);
    testRunner = new TestRunner(this, networkLayer);

    new DataStatistics(requestResponseManager);
  }

  public void decrementWaiting() {
    int remaining = waiting.decrementAndGet();
    Dascosa.out.println(FilteredOutput.DEBUG, "Sitemanager: Still waiting for "+ remaining +" ctests");
  }
  
  public void addTestListener(TestListener testListener) {
    testListeners.add(testListener);
  }

  public void removeTestListener(TestListener testListener) {
    testListeners.remove(testListener);
  }

  public RequestResponseManager getRequestResponseManager() {
    return requestResponseManager;
  }

  public QueryHandler getQueryHandler(){
    return queryHandler;
  }

  public String toString()
  {
    return "SiteManager @ " + database;
  }

  public void bootNode() throws IOException, InterruptedException
  {
    node = factory.newNode(factory.getNodeHandle(bootAddress));
    synchronized (node) {
      while (!node.isReady() && !node.joinFailed()) {
        node.wait(500);
        if (node.joinFailed())
          throw new IOException("Could not join the FreePastry ring. Reason: " + node.joinFailedReason());
      }
    }
    networkLayer.connect(node, messageHandler);
    localFragmentHandler.startPublish();
  }

  public void crashSite(AlgebraNode crashNode)
  {
    Dascosa.out.println("Crashing site: " + System.currentTimeMillis());
    isCrashed = true;
    queryHandler.crashSite();
    messageHandler.removeMessages();
    List<AlgebraNode> externalOperands = crashNode.getExternalOperandNodes();
    for (AlgebraNode operandNode : externalOperands)
      networkLayer.sendMessage(new CrashNotifiyMsg(operandNode.getNodeID()), operandNode.getAssignedSite());
  }

  public void executeSQL(String input) // Called by GUI-thread
  {
    String sql = input.trim();//.toLowerCase(); //blame!
    if (sql.startsWith("create table"))
      createHandler.executeCreate(sql);
    else if (sql.startsWith("get"))
      copyLookupClient.copyFragment(sql);
    else if (sql.startsWith("split"))
      fragmentSplitter.splitFragment(sql);
    else if (sql.startsWith("list"))
      floodManager.initiateFlood(new TableListRequestMsg(), tableListClient, null);
    else if (sql.startsWith("set")) {
      String property = input.substring(3).trim();
      boolean value = Dascosa.isPropertySet(property);
      floodManager.initiateFlood(new SetPropertyRequestMsg(property, !value), setPropertyClient, null);
    }
    else if (sql.startsWith("caching")) {
      try {
        StringTokenizer st = new StringTokenizer(sql);
        st.nextToken();
        boolean flush = st.nextToken().equals("flush");
        int cacheSize = 1000000*Integer.parseInt(st.nextToken());
        double sizeFactor = Double.parseDouble(st.nextToken());
        double costFactor = Double.parseDouble(st.nextToken());
        double depthFactor = Double.parseDouble(st.nextToken());
        double lastUsedFactor = Double.parseDouble(st.nextToken());
        floodManager.initiateFlood(new SetCacheParametersMsg(flush, cacheSize, sizeFactor, costFactor, depthFactor, lastUsedFactor), setCacheParametersClient, new Object[0]);
      } catch (Exception e) {
        fireDebugErrorEvent("Error parsing CACHING command.\n");
      }
    }
    else if (sql.startsWith("tpch")) {
      StringTokenizer st = new StringTokenizer(sql.substring(4));
      if (st.hasMoreTokens()) {
        int number = Integer.parseInt(st.nextToken());
        TpchTester tpchTester = new TpchTester(this, networkLayer, requestResponseManager);
        tpchTester.executeTPCH(number);
      }
      else
        executeAllTPCH();
    }
    else if (sql.startsWith("crash")) {
      StringTokenizer st = new StringTokenizer(sql.substring(5));
      int number = -1;
      if (st.hasMoreTokens())
        number = Integer.parseInt(st.nextToken());
      int crashNumber = -1;
      if (st.hasMoreTokens())
        crashNumber = Integer.parseInt(st.nextToken());
      executeCrashTest(number, crashNumber);
    }
    else if (sql.startsWith("ctest")) {
      /*
       * ctest <seed> <numQueries> <tests...>
       */
      StringTokenizer st = new StringTokenizer(sql.substring(5));
      int seed = 0;
      if (st.hasMoreTokens())
        seed = Integer.parseInt(st.nextToken());
      int numQueries = 0;
      if (st.hasMoreTokens())
        numQueries = Integer.parseInt(st.nextToken());
      List<Integer> tests = new ArrayList<Integer>();
      while (st.hasMoreTokens()) {
        tests.add(Integer.parseInt(st.nextToken()));
      }
      executeCacheTest(seed, numQueries, tests);
    }
    else if (sql.startsWith("script ")) {
      executeScript(sql.substring(7));
    }
    else if (sql.startsWith("sleep")) {
      StringTokenizer st = new StringTokenizer(sql.substring(5));
      int sleep = Integer.parseInt(st.nextToken());
      try {
        Thread.sleep(sleep);
      } catch (InterruptedException e) {
        // Ignore
      }
    }
    else if (sql.startsWith("exponential sleep")) {
//      StringTokenizer st = new StringTokenizer(sql.substring(5));
//      int sleep = Integer.parseInt(st.nextToken());
      try {
        Thread.sleep(8000);
      } catch (InterruptedException e) {
        // Ignore
      }
    }
//    else if (sql.startsWith("exponential sleep")) {
//      //TODO: exp delay
//      int sleep = (int)distribution.draw();
//      while (sleep > 0){
//        try {
//          sleep -= 1000;
//          System.err.println("Have to sleep for: " + sleep);
//          Thread.sleep(1000);
//        } catch (InterruptedException e) {
//          // Ignore
//        }
//      }
//    }
    else if (sql.startsWith("qodtest")) {
      //starts the distributed testrunner
      Pattern p = Pattern.compile("^qodtest (\\d+) (.+)$");
      Matcher m = p.matcher(sql);
      if (m.matches()){
        waiting.addAndGet(Integer.parseInt(m.group(1)));

        floodManager.initiateFlood(new TestRunnerRequestMsg(m.group(2), getLocalSite()), testRunner, null);
      }else{
        throw new IllegalArgumentException("qodtest <int:Number of sites running ctest> <String:Command>");
      }
      
    }
    else if (sql.startsWith("qodwait")) {
      // brukes til distribuert tester
      while (waiting.intValue()>0) {
        Thread.yield();
      }
    }
    
    else if (sql.startsWith("print ")) {
      System.out.println(sql.substring(6));
    }

    else {
      try {
        SqlScript.ReInit(new StringReader(sql));
        SimpleNode simpleNode = SqlScript.SQLScript();
        simpleNode.dump("");
        int sqlID = ((SimpleNode)simpleNode.jjtGetChild(0)).getId();
        if (sqlID == SqlScriptTreeConstants.JJTQUERYSTATEMENT)
          queryHandler.executeQuery(AlgebraConverter.convertAlgebra(simpleNode));
        else if (sqlID == SqlScriptTreeConstants.JJTINSERTSTATEMENT)
          insertHandler.executeInsert(sql, simpleNode);
        else if (sqlID == SqlScriptTreeConstants.JJTUPDATESTATEMENT ||
                 sqlID == SqlScriptTreeConstants.JJTDELETESTATEMENT) {
          updateHandler.executeUpdate(sql, simpleNode);
        }
        else
          fireDebugErrorEvent("Unsupported SQL statement.\n");
      } catch (ParseException e) {
        fireDebugErrorEvent(e.getMessage());
      }
      synchronized (SiteManager.this) {
        try {
            SiteManager.this.wait();
          } catch (InterruptedException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
          }
          System.out.println("Test finished: " + System.currentTimeMillis());
        }
        try {
          Thread.sleep(1000); // Give some extra time to cleanup
        } catch (InterruptedException e) {
          e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }
  }



  public void fireDebugErrorEvent(String errorMsg)
  {
    for (DebugEventListener debugEventListener : debugEventListeners)
      debugEventListener.showError(errorMsg);
  }

  public void fireDebugMessageEvent(String message)
  {
    for (DebugEventListener debugEventListener : debugEventListeners)
      debugEventListener.showMessage(message);
  }

  public NodeHandle getLocalSite()
  {
    return node.getLocalNodeHandle();
  }

  public NCClient getNCClient()
  {
    return ncClient;
  }

  public PyxidaHandler getPyxidaHandler()
  {
    return pyxidaHandler;
  }

  public boolean isCrashed()
  {
    if (isCrashed)
      return true;
    CrashNode crashNode = queryHandler.getCrashNode();
    if (crashNode != null && crashNode.isCrashed()) {
      crashSite(crashNode);
      return true;
    }
    else
      return false;
  }

  public void killNode()
  {
    localFragmentHandler.stopPublish();
    node.destroy();
  }

  public synchronized void queryCompleted()
  {
    notifyAll();
  }

  void addDebugEventListener(DebugEventListener debugEventListener)
  {
    debugEventListeners.add(debugEventListener);
  }

  public LocalFragmentHandler getLocalFragmentHandler() {
    return localFragmentHandler;
  }

  public TestRunner getTestRunner() {
    return testRunner;
  }

  private void executeAllTPCH()
  {
    final TpchTester tpchTester = new TpchTester(this, networkLayer, requestResponseManager);
    // Big test
    Runnable bigtester = new Runnable() {
      public void run()
      {
        int[] queries;
        if (Dascosa.isPropertySet(Dascosa.USE_OPTIMIZER))
          queries = new int[]{101, 102, 103, 104, 106, 110, 111, 112, 113, 114, 116, 117, 118, 120, 121};
        else
          queries = new int[]{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22};
        for (int i : queries) {
          fireDebugMessageEvent("TPC-H query " + i + ".\n");
          tpchTester.executeTPCH(i);
          waitForCompletion();
        }
      }

      private void waitForCompletion()
      {
        synchronized (SiteManager.this) {
          try {
            SiteManager.this.wait();
          } catch (InterruptedException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
          }
        }
        try {
          Thread.sleep(2000); // Give some extra time to cleanup
        } catch (InterruptedException e) {
          e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
      }
    };
    new Thread(bigtester).start();
  }


  private void executeCacheTest(final int seed, final int numQueries, final List<Integer> tests)
  {
    final CacheTester cacheTester = new CacheTester(this, networkLayer, requestResponseManager, seed);
    //tpchTester.executeTPCH(number, crashNumber);
    // Big test
    Runnable bigtester = new Runnable()
    {
      public void run()
      {
        //int tests[] = {102,103,4,105,6,7,8,9,110,111,112,13,14,15,16,19,20,121,22}; // 5 fail, 10 fail w/cache, 17 bug, 21 ? // Lokalt: 5, 17, 21: OOM, 18: ?, 20: FaultHandler timeout
        //int tests[] = {2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,19,20,21,22}; // 5 fail, 10 fail w/cache, 17 bug, 21 ? // Lokalt: 5, 17, 21: OOM, 18: ?, 20: FaultHandler timeout

        /*// Randomize list
        Random r = new Random(seed);
        List<Integer> tmptests = new ArrayList<Integer>(tests);
        List<Integer> unorderedTests = new ArrayList<Integer>();
        while (tmptests.size() > 0) {
          unorderedTests.add(tmptests.remove(r.nextInt(tmptests.size())));
        }*/

        Random rng = new ParetoRandom(seed);
//        Random rng = new Random(seed);
        System.out.println("*** BIG TEST STARTING " + System.currentTimeMillis());
        System.out.println("seed=" + seed + " numQueries=" + numQueries + " tests.size()=" + tests.size());
        for (int i = 0; i < numQueries; i++) {
          cacheTester.executeTest(tests.get(rng.nextInt(tests.size())));
          waitForCompletion();
          try {
            Thread.sleep(5000);
          } catch (InterruptedException ex) {
            Logger.getLogger(SiteManager.class.getName()).log(Level.SEVERE, null, ex);
          }
        }
        for (TestListener testListener : testListeners) {
          testListener.testComplete();
        }

        System.out.println("*** BIG TEST COMPLETED " + System.currentTimeMillis());
      }

      private void waitForCompletion()
      {
        synchronized (SiteManager.this) {
          try {
            SiteManager.this.wait();
          } catch (InterruptedException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
          }
          System.out.println("Test finished: " + System.currentTimeMillis());
        }
        try {
          Thread.sleep(1000); // Give some extra time to cleanup
        } catch (InterruptedException e) {
          e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
      }
    };
    //new Thread(bigtester).start();
    bigtester.run();
  }

  private void executeCrashTest(int number, int crashNumber)
  {
    CrashTester crashTester = new CrashTester(this, networkLayer, requestResponseManager);
    crashTester.executeCrash(number, crashNumber);
//    // Big test
//    Runnable bigtester = new Runnable() {
//      public void run()
//      {
//        System.out.println("*** BIG TEST STARTING " + System.currentTimeMillis());
//        for (int i = 0; i < 20; i++) {
//          crashTester.executeCrash(number, crashNumber);
//          waitForCompletion();
//        }
//        System.out.println("*** BIG TEST COMPLETED " + System.currentTimeMillis());
//      }
//
//      private void waitForCompletion()
//      {
//        synchronized (this) {
//          try {
//            wait();
//          } catch (InterruptedException e) {
//            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
//          }
//        }
//        try {
//          Thread.sleep(10000); // Give some extra time to cleanup
//        } catch (InterruptedException e) {
//          e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
//        }
//      }
//    };
//    new Thread(bigtester).start();
  }

  private void executeScript(final String file)
  {
    Runnable script = new Runnable() {
      public void run()
      {
        BufferedReader reader;
        try {
        reader = new BufferedReader(new FileReader(file));
        } catch (FileNotFoundException e) {
          System.err.println("Script file '" + file + "' not found.");
          return;
        }
        System.out.println("*** SCRIPT '" + file + "' STARTING ***");
        try {
          String line;
          while((line = reader.readLine()) != null) {
              if (!line.trim().startsWith("#") && line.trim().length() != 0){
                executeSQL(line);
                //waitForCompletion();
              }
          }
        } catch (IOException e) {
          e.printStackTrace();
        }
        System.out.println("*** SCRIPT '" + file + "' COMPLETED ***");
      }
      private void waitForCompletion()
      {

      }
    };
    new Thread(script).start();
  }

  PastryNode getPastryNode()
  {
    return node;
  }

  void restartNode()
  {
    Dascosa.out.println("Restarting site: " + System.currentTimeMillis());
    isCrashed = false;
  }

  void stopMessageSending(long algebraNodeID)
  {
    AlgebraNode algebraNode = queryHandler.getLocalAlgebraNode(algebraNodeID);
    if (algebraNode != null)
      algebraNode.setHalted(true);
  }

// -------------------------- INNER CLASSES --------------------------

  private final class CopyLookupClient implements MessageClient, FragmentCopyClient
  {
    private void copyFragment(String sql)
    {
      StringTokenizer st = new StringTokenizer(sql);
      st.nextToken();
      String tableName = st.nextToken();
      int fragmentNr = Integer.parseInt(st.nextToken());
      AbstractRequestMsg message = new LookupRequestMsg(tableName, fragmentNr);
      requestResponseManager.sendRequest(message, tableName, copyLookupClient, new Object[0]);
    }

    public void responseReceived(AbstractResponseMsg msg, Object parameters[], DataSource dataSource)
    {
      LookupResponseMsg message = (LookupResponseMsg)msg;
      fragmentCopier.requestCopy(this, message.getTableName(), message.getFragmentNr(),
                                 message.getSites()[0][0], new Object[0]);
    }

    public void fragmentCopied(String tableName, int fragmentNr, Object parameters[])
    {
      fireDebugMessageEvent("Fragment copied to local site.\n");
    }
  }

  private final class SetCacheParametersClient implements FloodClient
  {
    public void responseReceived(AbstractFloodResponseMsg message, Object parameter[])
    {
      SetCacheParametersResponseMsg msg = (SetCacheParametersResponseMsg)message;
      fireDebugMessageEvent("cacheSize=" + msg.getCacheSize() + ".\n");
    }

    public void floodCompleted(Object parameter[])
    {
      fireDebugMessageEvent("Cache parameters complete.\n");
    }
  }

  private final class SetPropertyClient implements FloodClient
  {
    public void floodCompleted(Object[] parameter)
    {
      fireDebugMessageEvent("Set property completed.\n");
    }

    public void responseReceived(AbstractFloodResponseMsg message, Object[] parameter)
    {
    }
  }

  private final class TableListClient implements FloodClient
  {
    public void responseReceived(AbstractFloodResponseMsg message, Object parameter[])
    {
      TableListResponseMsg msg = (TableListResponseMsg)message;
      for (String tableName : msg.getTableNames())
        fireDebugMessageEvent(tableName + '\n');
    }

    public void floodCompleted(Object parameter[])
    {
      fireDebugMessageEvent("List complete.\n");
    }
  }

  public Coordinate getLocalCoords() {
    return ncClient.getApplicationCoords();
  }
}
