package dascosa.site.cache;

import dascosa.data.PacketHeader;
import dascosa.data.Packet;

import dascosa.site.query.optimizer.ResultSetIdentifier;
import dascosa.Dascosa;

import java.util.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
 * A cache entry as stored by ResultCache
 */
final class CacheEntry
{
  /** Cache is still filling up */
  static final int STATUS_CACHING = 2;
  /** Caching of this result is complete */
  static final int STATUS_COMPLETE = 3;

  /** The canonical form of the cached query */
  private final String canonicalForm;
  /** Cost estimate calculated before caching starts */
  private final double costEstimate;
  /** Size estimate calculated before caching starts */
  private final double sizeEstimate;
  /** Depth of algebra node in query tree, leaves are at level 1 */
  private final int depth;
  /** The cached packets for this entry */
  private final List<Packet> packets = new LinkedList<Packet>();
  /** The cached packet header for this entry */
  private final PacketHeader packetHeader;
  /** The current status of this entry */
  private int status;
  /** The current size of this entry */
  private int size;
  /** The table hash for this entry */
  private final String tableHash;
  /** The count of tuples in this entry, is updated for each new packet of tuples added */

  private final ResultSetIdentifier resultSetIdentifier;

  /**
   * Create a new cache entry
   *
   * @param canonicalForm Canonical form of the query
   * @param packetHeader Packet header for result
   * @param costEstimate Cost estimate from algebra node
   * @param sizeEstimate Size estimate from algebra node
   * @param depth Algebra node depth in tree (from algebra node)
   */
  CacheEntry(String canonicalForm, ResultSetIdentifier resultSetIdentifier, PacketHeader packetHeader,
             double costEstimate, double sizeEstimate, int depth)
  {
    this.canonicalForm = canonicalForm;
    this.resultSetIdentifier = resultSetIdentifier;
    status = STATUS_CACHING;
    size = 0;
    this.packetHeader = packetHeader;
    this.costEstimate = costEstimate;
    this.sizeEstimate = sizeEstimate;
    this.depth = depth;
    tableHash = CacheEntry.tableHash(canonicalForm).toLowerCase();
  }

  /**
   * Select one of the tables involved in the query
   *
   * @param query Canonical form of query
   * @return Name of one of the involved tables
   */
  private static String tableHash(String query)
  {
    if (Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER)) {
      SortedSet<String> tables = new TreeSet<String>();
      Pattern p = Pattern.compile("LogicalTableAccess\\(([^)]+)\\)|SelProjTableAccess\\(([^)(\\s]+)_\\d+ : \\[[\\w,\\s]*\\] : \\(");
      Matcher m = p.matcher(query);
      while (m.find()) {
        tables.add(m.group(1)!=null?m.group(1):m.group(2));
      }

      return tables.first();

    } else {

      ArrayList<String> tables = new ArrayList<String>();

      //Pattern p = Pattern.compile("SimpleTableAccess\\(([^)]+)\\)");
      Pattern p = Pattern.compile("LogicalTableAccess\\(([^)]+)\\)");
      Matcher m = p.matcher(query);
      while (m.find()) {
        tables.add(m.group(1));
      }
      p = Pattern.compile("SelProjTableAccess\\(([^)]+)_");
      m = p.matcher(query);
      while (m.find()) {
        tables.add(m.group(1));
      }
      //TODO: Should probably sort tables alphabetically (or is this taken care of by the canonical form?)
      //String fragment = tables.get(Math.abs(query.hashCode() % tables.size()));
      //if (fragment.indexOf('_') > 0)
      //  fragment = fragment.substring(0, fragment.indexOf('_'));
      return tables.get(Math.abs(query.hashCode() % tables.size()));
    }
  }

  synchronized void addPacket(Packet packet)
  {
    packets.add(packet);
    size += packet.getPacketSize();
  }

  String getCanonicalForm()
  {
    return canonicalForm;
  }

  double getCostEstimate()
  {
    return costEstimate;
  }

  ResultSetIdentifier getResultSetIdentifier() {
    return resultSetIdentifier;
  }

  int getTupleCount() {
      return resultSetIdentifier.getTupleCount();
  }

  int getDepth()
  {
    return depth;
  }

  PacketHeader getPacketHeader()
  {
    return packetHeader;
  }

  synchronized List<Packet> getPackets()
  {
    return new ArrayList<Packet>(packets);
  }

  int getSize()
  {
    return size;
  }

  double getSizeEstimate()
  {
    return sizeEstimate;
  }

  int getStatus()
  {
    return status;
  }

  void setStatus(int status)
  {
    this.status = status;
  }

  public String getTableHash()
  {
    return tableHash;
  }

  public void increaseTupleCount(int newTuples) {
    if (Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER)) {
      resultSetIdentifier.setTupleCount(resultSetIdentifier.getTupleCount() + newTuples);
    }
    //System.err.println("NEW TUPLECOUNT IS: " + tupleCount);
  }
}
