package dascosa.site.cache;

import dascosa.algebra.AlgebraNode;
import dascosa.algebra.CacheNode;
import dascosa.algebra.scan.TableAccessNode;
import dascosa.data.Packet;
import dascosa.data.PacketHeader;
import dascosa.message.PublishCacheMsg;
import dascosa.site.index.Index;
import dascosa.site.SiteManager;
import dascosa.site.NetworkLayer;
import dascosa.site.fault.CacheListener;
import dascosa.Dascosa;
import dascosa.site.query.optimizer.ResultSetIdentifier;
import dascosa.util.FilteredOutput;
import rice.p2p.commonapi.Message;

import javax.swing.Timer;
import java.awt.event.*;
import java.util.*;
import java.util.Map.Entry;

/**
 *  Main class of semantic caching. Contains publishing timer, list of cache entries, replacement policy, etc.
 */
public final class ResultCache
{
  /** LRU-liste av cachebruk, canonicalform */
  private final List<String> lruList = new LinkedList<String>();
  /** AlgebraNodeID => CacheNode, Hvilke ordinære noder har blitt erstattet av hvilken cachenode */
  private final Map<Long, CacheNode> replacedNodes = new HashMap<Long, CacheNode>();
  /** Cache entries, indexed by canonical form of query */
  private final Map<String, CacheEntry> cacheEntries = new HashMap<String, CacheEntry>();
  /** Cachereplacement-kø, inneholder faktiske algebranoder. Er det nødvendig? */
  private final Comparator<CacheEntry> cacheComparator = new CacheComparator();
  private final List<CacheEntry> replacementQueue = new ArrayList<CacheEntry>();
  private double sizeFactor;
  private double costFactor;
  private double depthFactor;
  private double lastUsedFactor;
  /** Maks størrelse på cache */
  private int cacheSize;
  private final ActionListener publishListener;
  private final List<CacheListener> cacheListeners = new ArrayList<CacheListener>();

  private SiteManager siteManager;

  /**
   * Create a new result cache.
   *
   * @param siteManager Site manager for this site
   * @param networkLayer Network layer for this site
   */
  public ResultCache(final SiteManager siteManager, final NetworkLayer networkLayer)
  {
    this.siteManager = siteManager;
    publishListener = new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        if (Dascosa.isPropertySet(Dascosa.USE_SEMANTIC_CACHE)) {
          synchronized(ResultCache.this) {
            HashMap<String, ArrayList<CacheEntry>> allEntries = new HashMap<String, ArrayList<CacheEntry>>();
            for (CacheEntry ent : cacheEntries.values()) {
              if (ent.getStatus() == CacheEntry.STATUS_COMPLETE) {
                String tableHash = ent.getTableHash();
                ArrayList<CacheEntry> hashBucket = allEntries.get(tableHash);
                if (hashBucket == null){
                  hashBucket = new ArrayList<CacheEntry>();
                  allEntries.put(tableHash, hashBucket);
                }
                hashBucket.add(ent);
                //Message message = new PublishCacheMsg(siteManager.getLocalSite(), ent.getResultSetIdentifier(), ent.getCanonicalForm(), siteManager.getNCClient().getApplicationCoords());
                //networkLayer.sendMessage(message, ent.getTableHash());
              }

            }
            for (Entry<String, ArrayList<CacheEntry>> hashBucket : allEntries.entrySet()){
              ArrayList<CacheEntry> entries = hashBucket.getValue();
              ResultSetIdentifier[] rsis = new ResultSetIdentifier[entries.size()];
              String[] canonicals = new String[entries.size()];
              int i = 0;
              for (CacheEntry entry : entries){
                rsis[i] = entry.getResultSetIdentifier();
                canonicals[i++] = entry.getCanonicalForm();
              }
              Message message = new PublishCacheMsg(siteManager.getLocalSite(), rsis, canonicals, siteManager.getNCClient().getApplicationCoords());
              networkLayer.sendMessage(message, hashBucket.getKey());
            }
          }
        }
      }
    };
    Timer publishTimer = new Timer(Index.INDEX_PUBLISH_INTERVAL, publishListener);
    publishTimer.setRepeats(true);
    publishTimer.start();
  }

  /**
   * Register a cache node for delivering cached data.
   *
   * @param replacedNodeID Node ID of the algebra node that is replaced by the cache node
   * @param cacheNode The cache node that replaces the algebra node
   */
  public synchronized void addCacheNode(long replacedNodeID, CacheNode cacheNode)
  {
    //synchronized (replacedNodes) {
      replacedNodes.put(replacedNodeID, cacheNode);
    //}
  }

  /**
   * Potentially add a packet to the cache. The packet is only added if we have decided to cache the result.
   * Otherwise it is ignored.
   *
   * If adding the packet makes the cache too big (getUsedSize() > cacheSize), something has to be removed
   * from the cache.
   *
   * @param node Algebra node that produced the packet
   * @param packet The packet
   */
  public synchronized void addPacket(AlgebraNode node, Packet packet)
  {
    String query = node.canonicalForm();
    CacheEntry cacheEntry = cacheEntries.get(query);
    if (cacheEntry == null || cacheEntry.getStatus() != CacheEntry.STATUS_CACHING || !node.isPJS())
      return;

    cacheEntry.addPacket(packet);
    cacheEntry.increaseTupleCount(packet.getTupleCount());
    if (getUsedSize() > cacheSize) {
      Collections.sort(replacementQueue, cacheComparator);
      CacheEntry removeEntry = replacementQueue.isEmpty() ? null : replacementQueue.get(0);
      if (removeEntry != null) {
        //System.err.println("DEBUG: Cache overflow, removing '" + removeEntry.getCanonicalForm() + "'.");
        remove(removeEntry);
      }
    }

    if (packet.isLastPacket()) {
      if (cacheEntries.containsKey(query)) { // might be removed in above block
        cacheEntry.setStatus(CacheEntry.STATUS_COMPLETE);
        assert lruList.indexOf(query) == -1; // New node, should not be in list
        lruList.add(0, query);
        // Reinsert in priority queue with updated score (real vs. estimated size)
//        replacementQueue.remove(cacheEntry); // BUG!!!!
//        replacementQueue.add(cacheEntry);
        System.err.println("DEBUG: Caching complete: estimated size=" + cacheEntry.getSizeEstimate() +
                           ", real size=" + cacheEntries.get(query).getSize() + "\n       (query='" + query + "')");
        Dascosa.out.println(FilteredOutput.CACHELOG, "Added " + cacheEntry.getResultSetIdentifier() + " to site " + siteManager.getLocalSite().toString());
      }
    }

    //Trigger catalog publishing
    publishListener.actionPerformed(null);
  }

  /**
   * Potentially add a packet header to the cache. The decision to cache a query result or not is made here.
   *
   * @param node The algebra node that produced the packet header
   * @param packetHeader The packet header
   */
  public synchronized void addPacketHeader(AlgebraNode node, PacketHeader packetHeader)
  {
    if (node instanceof CacheNode && node.getOperandNodes().size() == 0)
      return; // Don't cache cache results
    String query = node.canonicalForm();
    if (node instanceof CacheNode && node.getOperandNodes().size() > 0)
      query = node.getOperandNodes().get(0).canonicalForm();
    // Decide to cache or not
    //System.err.println("Packet header for " + node.canonicalForm() + ": " + packetHeader);
    if (cacheEntries.get(query) == null && shouldCache(node)) {
      ResultSetIdentifier rsi = Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER) && node.isPJS() ? node.buildResultSetIdentifier().get(0): null;
      //System.err.println("Caching packet header for " + node.canonicalForm() + ": " + packetHeader);

      CacheEntry cacheEntry = new CacheEntry(query, rsi, packetHeader, node.getCostEstimate(), node.getSizeEstimate(), node.getDepth());
      cacheEntries.put(query, cacheEntry);
      replacementQueue.add(cacheEntry);
      for (CacheListener cacheListener: cacheListeners)
        cacheListener.semanticCacheEntryAdded(query);
    }
  }

  /**
   * Flush cache, clear all records.
   */
  public synchronized void flush()
  {
    replacementQueue.clear();
    cacheEntries.clear();
    lruList.clear();
    for (CacheListener cacheListener: cacheListeners)
      cacheListener.semanticCacheCleared();
    //TODO: Trigger catalog publishing timer
  }

  /**
   * Get packet header for a cached query.
   *
   * @param canonicalForm Canonical form of cached query
   * @return Packet header
   */
  public synchronized PacketHeader getPacketHeader(String canonicalForm)
  {
    return cacheEntries.get(canonicalForm).getPacketHeader();
  }

  /**
   * Get list of packets for a cached query.
   *
   * @param canonicalForm Canonical form of cached query
   * @return List of packets
   */
  public synchronized List<Packet> getPackets(String canonicalForm)
  {
    return cacheEntries.get(canonicalForm).getPackets();
  }

  /**
   * Get size of current cache entries
   *
   * @return Size of current cache entries
   */
  public synchronized int getUsedSize()
  {
    int size = 0;
    for (CacheEntry e : cacheEntries.values())
      size += e.getSize();
    return size;
  }

  /**
   * Is there a complete (finished) cache of this query?
   *
   * @param canonicalForm Campmocal form of query
   * @return True if a completed cache exists
   */
  public synchronized boolean isCached(String canonicalForm)
  {
    return cacheEntries.get(canonicalForm) != null &&
           cacheEntries.get(canonicalForm).getStatus() == CacheEntry.STATUS_COMPLETE;
  }

  /**
   * Remove a node from mapping between replaced nodes' IDs and cache nodes.
   *
   * @param replacedNode Replaced node
   * @return The cache node that was removed
   */
  public synchronized AlgebraNode removeCacheNode(AlgebraNode replacedNode)
  {
    //synchronized (replacedNodes) {
      return replacedNodes.remove(replacedNode.getNodeID());
    //}
  }

  /**
   * Set caching parameters.
   *
   * @param cacheSize Maximum cache size
   * @param sizeFactor Size factor for priority queue calculations
   * @param costFactor Cost factor for priority queue calculations
   * @param depthFactor Depth factor for priority queue calculations
   * @param lastUsedFactor LRU factor for priority queue calculations
   */
  public synchronized void setParameters(int cacheSize, double sizeFactor, double costFactor, double depthFactor,
                            double lastUsedFactor)
  {
    if (getUsedSize() > cacheSize)
      flush();
    this.cacheSize = cacheSize;
    this.sizeFactor = sizeFactor;
    this.costFactor = costFactor;
    this.depthFactor = depthFactor;
    this.lastUsedFactor = lastUsedFactor;
    System.out.println("DEBUG: Cache parameters set.");
  }

  /**
   * Update LRU list when an entry has been used
   *
   * @param query Canonical form of cached query that has been used (which caused this update)
   */
  public synchronized void updateUsage(String query)
  {
    lruList.remove(query);
    lruList.add(0, query);

      // Reinsert all nodes in queue to update LRU score
//      CacheEntry entries[] = replacementQueue.toArray(new CacheEntry[replacementQueue.size()]);
//      replacementQueue.clear();
//      replacementQueue.addAll(Arrays.asList(entries));
  }

  /**
   * Remove a cache entry
   *
   * @param cacheEntry Canonical form of entry to remove
   */
  private synchronized void remove(CacheEntry cacheEntry)
  {
    cacheEntries.remove(cacheEntry.getCanonicalForm());
    Dascosa.out.println(FilteredOutput.CACHELOG, "Removed " + cacheEntry.getResultSetIdentifier() + " from site " + siteManager.getLocalSite().toString());
    replacementQueue.remove(cacheEntry);
    lruList.remove(cacheEntry.getCanonicalForm());
    for (CacheListener cacheListener: cacheListeners)
      cacheListener.semanticCacheEntryRemoved(cacheEntry.getCanonicalForm());
    //TODO: Trigger catalog publishing timer
  }

  /**
   * Decide if a node should be cached or not.
   *
   * The decision is made based on a priority queue. The node is compared to the head of the queue (the least
   * "cache worthy" element in cache).
   *
   * @param node The algebra node which result might be cached
   * @return True if the result should be cached, false if not
   */
  private synchronized boolean shouldCache(AlgebraNode node)
  {
    //if (node instanceof CacheNode) return false;
    if (!node.isPJS())
      return false;

//    if (node instanceof CacheNode && node.getOperandNodes().size() == 1)
//      node = node.getOperandNodes().get(0);

    String canonicalForm;
    if (node instanceof CacheNode)
      canonicalForm = node.getOperandNodes().get(0).canonicalForm();
    else
      canonicalForm = node.canonicalForm();

    int usedSize = getUsedSize();
    double estimatedSize = node.getSizeEstimate();
    double cost = 0; //node.getCostEstimate();
    int depth = node.getDepth();
    int lruIndex = lruList.indexOf(canonicalForm);

    //System.err.println("DEBUG: cacheSize=" + cacheSize + ", usedSize=" + usedSize + ", estimatedSize=" +
    //                   estimatedSize + ", cost=" + cost + ", depth=" + depth + ", lruIndex=" + lruIndex);

    // Is caching turned off?
    if (cacheSize <= 0) {
      //System.err.println("DEBUG: Caching is off.");
      return false;
    }

    assert cacheSize > 0 : cacheSize;

    if (node instanceof TableAccessNode) {
      TableAccessNode tableAccessNode = (TableAccessNode)node;
      if (siteManager.getLocalFragmentHandler().existsIn(tableAccessNode.getTableName())) {
        return false;
      }
    }

//    if (node instanceof CacheNode)
//      return false;

    // Ignore table access and cache nodes:
//    if (node instanceof TableAccessNode || node instanceof CacheNode) //sjekke at TableAccessNode ikke er lokal
//      return false;

    // Don't cache if estimated size is larger than cache size
    if (estimatedSize > cacheSize) {
      //System.err.println("DEBUG: Decided NOT to cache " + node.canonicalForm() +
      //                   " Reason: Too large to fit in cache (" + estimatedSize + " > " + cacheSize + ").");
      return false;
    }

    assert estimatedSize < cacheSize : estimatedSize;

    // Don't cache if it cache to small, and node scores worse than the worst already cached

    ResultSetIdentifier rsi = Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER) && node.isPJS() ? node.buildResultSetIdentifier().get(0): null; //This is safe because topNode is PJS
    CacheEntry cacheEntry = new CacheEntry(canonicalForm, rsi, null, cost, estimatedSize, depth);

    Collections.sort(replacementQueue, cacheComparator);
    if (estimatedSize > (cacheSize - usedSize) && !replacementQueue.isEmpty() &&
        cacheComparator.compare(cacheEntry, replacementQueue.get(0)) < 0) {
      //System.err.println("DEBUG: Decided NOT to cache " + node.canonicalForm() +
      //                   " Reason: Scores worse than worst cached.");
      return false;
    }

    // Check against all that will be removed to make it fit
    int replacedSize = 0;
    Iterator<CacheEntry> queueIterator = replacementQueue.iterator();
    while (cacheSize - (usedSize - replacedSize) < estimatedSize) {
      if (!queueIterator.hasNext()) {
        //System.err.println("DEBUG: Decided NOT to cache " + node.canonicalForm() + " Reason: Bug in algorithm. cacheSize=" + cacheSize + ", usedSize=" + usedSize + ", replacedSize=" + replacedSize);
        return false;
      }
      CacheEntry replacedEntry = queueIterator.next();
      if (cacheComparator.compare(cacheEntry, replacedEntry) < 0) {
        //System.err.println("DEBUG: Decided NOT to cache " + node.canonicalForm() +
        //                   " Reason: Scores worse than replaced.");
        return false;
      }
      replacedSize += replacedEntry.getSizeEstimate();
    }

    // At this point we should have rejected all don't-cache situations

    //Flyttet til addPacketHeader(): replacementQueue.add(new CacheEntry(node.canonicalForm(), CacheEntry.STATUS_CACHING, 0, null, null, node.getCostEstimate(), node.getSizeEstimate(), node.getDepth()));
    //System.err.println("DEBUG: Decided to cache " + node.canonicalForm());
    return true;
  }

  public void addCacheListener(CacheListener cacheListener)
  {
    cacheListeners.add(cacheListener);
  }

  public void removeCacheListener(CacheListener cacheListener)
  {
    cacheListeners.remove(cacheListener);
  }

// -------------------------- INNER CLASSES --------------------------

  /**
   * Compares cache entries to find out which is more important to have in cache.
   *
   * This is where the cache replacement policy is carried out.
   */
  private final class CacheComparator implements Comparator<CacheEntry>
  {
    /**
     * Compare two cache entries.
     *
     * If e1<e2, e1 is removed first from the cache (of the two).
     *
     * @param e1 First cache entry
     * @param e2 Second cache entry
     * @return Negative if e1<e2, 0 if e1 and e2 have an equal score, or positive if e1>e2
     */
    public int compare(CacheEntry e1, CacheEntry e2)
    {
      // e1<e2 ==> e1 forsvinner først fra cache
      // returverdi: e1<e2: -, e1=e2: 0, e1>e2: +
      // lru index of non-cached queries: -1 (automatically from .indexOf)
      double d1;
      double d2;
      String query1 = e1.getCanonicalForm();
      String query2 = e2.getCanonicalForm();
      synchronized (ResultCache.this) {
        double size1;
        if (cacheEntries.containsKey(query1))
          size1 = cacheEntries.get(query1).getSize();
        else
          size1 = e1.getSizeEstimate();
        double size2;
        if (cacheEntries.containsKey(query2))
          size2 = cacheEntries.get(query2).getSize();
        else
          size2 = e2.getSizeEstimate();
        d1 = sizeFactor * size1 + costFactor * e1.getCostEstimate() + depthFactor * e1.getDepth() + lastUsedFactor * lruList.indexOf(e1.getCanonicalForm());
        d2 = sizeFactor * size2 + costFactor * e2.getCostEstimate() + depthFactor * e2.getDepth() + lastUsedFactor * lruList.indexOf(e2.getCanonicalForm());
      }
      return Double.compare(d1, d2);
    }
  }
}
