package dascosa.site.dynfrag;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 22.jul.2008
 * Time: 15:18:33
 * To change this template use File | Settings | File Templates.
 */
final class BucketSet
{
  static final int WIDTH_INCREASE_FACTOR = 10;
  static final int MAX_BUCKETS = 100;
  private final AccessLogger logger;
  private final Map<Integer, Bucket> buckets = new HashMap<Integer, Bucket>(); // BucketNumber => Count
  private final List<Bucket> freeBucketList = new LinkedList<Bucket>();

  BucketSet(AccessLogger logger)
  {
    this.logger = logger;
  }

  /**
   * Produces a new bucket set from joining several old ones. Note: Old buckets have same or smaller width,
   * so no bucket splitting is necessary.
   */
  BucketSet(AccessLogger logger, BucketSet oldBucketSets[])
  {
    this(logger);
    for (BucketSet oldBucketSet : oldBucketSets) {
      if (oldBucketSet != null) {
        for (int oldBucketNumber : oldBucketSet.buckets.keySet()) {
          int newBucketNumber = logger.computeBucketNumber(oldBucketSet.logger.computeID(oldBucketNumber));
          addToBucket(newBucketNumber, oldBucketSet.buckets.get(oldBucketNumber));
        }
        freeBucketList.addAll(oldBucketSet.buckets.values()); // NB! Destroys oldBucketSets
      }
    }
  }

//  /** Produces a new bucket set from a subset of an old one. Note: Assumes same bucketWidth */
//  private BucketSet(AccessLogger logger, BucketSet oldBucketSet, int oldStartBucketNumber, int oldEndBucketNumber)
//  {
//    this(logger);
//    for (int oldBucketNumber = oldStartBucketNumber; oldBucketNumber <= oldEndBucketNumber; oldBucketNumber++)
//      if (oldBucketSet.buckets.containsKey(oldBucketNumber)) {
//        int newBucketNumber = logger.computeBucketNumber(oldBucketSet.logger.computeID(oldBucketNumber));
//        addToBucket(newBucketNumber, oldBucketSet.buckets.get(oldBucketNumber));
//      }
//  }

  public String toString()
  {
    return buckets.toString();
  }

  private void addToBucket(int bucketNumber, Bucket bucket)
  {
    addToBucket(bucketNumber, bucket.reads, bucket.writes);
  }

  private void addToBucket(int bucketNumber, int reads, int writes)
  {
    Bucket newBucket = buckets.get(bucketNumber);
    if (newBucket == null) {
      if (!freeBucketList.isEmpty()) {
        newBucket = freeBucketList.remove(0);
        newBucket.reads = reads;
        newBucket.writes = writes;
      }
      else
        newBucket = new Bucket(reads, writes);
      buckets.put(bucketNumber, newBucket);
    }
    else {
      newBucket.reads += reads;
      newBucket.writes += writes;
    }
  }

  Collection<Integer> getBucketNumbers()
  {
    return buckets.keySet();
  }

  int getBucketsUsed()
  {
    return buckets.size();
  }

  int getHighestBucketNumber()
  {
    int num = Integer.MIN_VALUE;
    for (int bn : buckets.keySet())
      num = Math.max(num, bn);
    return num;
  }

  int getLowestBucketNumber()
  {
    int num = Integer.MAX_VALUE;
    for (int bn : buckets.keySet())
      num = Math.min(num, bn);
    return num;
  }

  int getReads(List<Integer> bucketNumbers, int startIndex, int endIndex)
  {
    int reads = 0;
    for (int n = startIndex; n <= endIndex; n++) {
      Bucket bucket = buckets.get(bucketNumbers.get(n));
      if (bucket != null)
        reads += bucket.reads;
    }
    return reads;
  }

  int getWrites(List<Integer> bucketNumbers, int startIndex, int endIndex)
  {
    int writes = 0;
    for (int n = startIndex; n <= endIndex; n++) {
      Bucket bucket = buckets.get(bucketNumbers.get(n));
      if (bucket != null)
        writes += bucket.writes;
    }
    return writes;
  }

  void makeBiggerBuckets(int increaseFactor)
  {
    Map<Integer, Bucket> oldBuckets = new HashMap<Integer, Bucket>(buckets);
    buckets.clear();
    for (Integer oldBucketNumber : oldBuckets.keySet()) {
      int bucketNumber = oldBucketNumber / increaseFactor;
      addToBucket(bucketNumber, oldBuckets.get(oldBucketNumber));
    }
    freeBucketList.addAll(oldBuckets.values());
  }

  void makeSmallerBuckets(int reduceFactor)
  {
    Map<Integer, Bucket> oldBuckets = new HashMap<Integer, Bucket>(buckets);
    buckets.clear();
    for (Integer oldBucketNumber : oldBuckets.keySet()) {
      int newStartBucketNumber = oldBucketNumber * reduceFactor; // Including
      int newEndBucketNumber = (oldBucketNumber + 1) * reduceFactor - 1; // Including
      newEndBucketNumber = Math.min(newEndBucketNumber, logger.computeBucketNumber((int)logger.getMaxAssigned()));
      int oldBucketReadCount = oldBuckets.get(oldBucketNumber).reads;
      int newBucketReadCount = oldBucketReadCount / (newEndBucketNumber - newStartBucketNumber + 1); // Introduces error
      if (newBucketReadCount == 0 && oldBucketReadCount > 0)
        newBucketReadCount = 1; // To avoid that tuples disappear
      int oldBucketWriteCount = oldBuckets.get(oldBucketNumber).writes;
      int newBucketWriteCount = oldBucketWriteCount / (newEndBucketNumber - newStartBucketNumber + 1); // Introduces error
      if (newBucketWriteCount == 0 && oldBucketWriteCount > 0)
        newBucketWriteCount = 1; // To avoid that tuples disappear
      for (int newBucketNumber = newStartBucketNumber; newBucketNumber <= newEndBucketNumber; newBucketNumber++) {
        addToBucket(newBucketNumber, newBucketReadCount, newBucketWriteCount);
      }
    }
    freeBucketList.addAll(oldBuckets.values());
  }

  void recordAccess(int id, boolean isWrite)
  {
    int bucketNumber = logger.computeBucketNumber(id);
    if (logger.computeID(bucketNumber) > logger.getMaxAssigned())
      throw new RuntimeException("ERROR: Bucketnumber outside BucketSet");
    if (buckets.containsKey(bucketNumber)) {
      Bucket oldBucket = buckets.get(bucketNumber);
      if (isWrite)
        oldBucket.writes++;
      else
        oldBucket.reads++;
    }
    else if (buckets.size() < MAX_BUCKETS)
      addToBucket(bucketNumber, isWrite ? 0 : 1, isWrite ? 1 : 0);
    else {
      int bucketFactor = 1;
      while (buckets.size() >= MAX_BUCKETS) {
        bucketFactor *= WIDTH_INCREASE_FACTOR;
        makeBiggerBuckets(WIDTH_INCREASE_FACTOR);
      }
      bucketNumber = logger.computeBucketNumber(id, logger.getBucketWidth() * bucketFactor);
      addToBucket(bucketNumber, isWrite ? 0 : 1, isWrite ? 1 : 0);
      logger.makeOtherBucketsBigger(this, bucketFactor);
    }
  }

// -------------------------- INNER CLASSES --------------------------

  private static final class Bucket
  {
    private int reads;
    private int writes;

    private Bucket(int reads, int writes)
    {
      this.writes = writes;
      this.reads = reads;
    }

    public String toString()
    {
      return "R:" + reads + " W:" + writes;
    }
  }
}
