package dascosa.site.fault;

import dascosa.algebra.*;
import dascosa.algebra.scan.TableAccessNode;
import dascosa.message.request.*;
import dascosa.message.restart.CompleteRestartFailedMsg;
import dascosa.message.restart.PartialRestartFailedMsg;
import dascosa.site.*;
import dascosa.site.query.QueryPlanner;
import dascosa.Dascosa;
import rice.p2p.commonapi.Message;
import rice.p2p.commonapi.NodeHandle;

import javax.sql.DataSource;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 10.sep.2007
 * Time: 15:41:41
 * To change this template use File | Settings | File Templates.
 */
final class FaultHandler implements MessageClient
{
  private final Map<AlgebraNode, Message> restartMsg = new HashMap<AlgebraNode, Message>();
  private final NetworkLayer networkLayer;
  private final RequestResponseManager requestResponseManager;

  FaultHandler(NetworkLayer networkLayer, RequestResponseManager requestResponseManager)
  {
    this.networkLayer = networkLayer;
    this.requestResponseManager = requestResponseManager;
  }

// --------------------- Begin Interface MessageClient ---------------------

  public void responseReceived(AbstractResponseMsg msg, Object parameters[], DataSource dataSource)
  {
    LookupResponseMsg message = (LookupResponseMsg)msg;
    AlgebraNode topNode = (AlgebraNode)parameters[0];
    TableAccessNode tableNode = (TableAccessNode)parameters[1];
    NodeHandle failedSite = (NodeHandle)parameters[2];
    tableNode.setAttributeNames(message.getAttributeNames());
    boolean foundLiveSite = false;
    for (NodeHandle site: message.getSites()[0]) {
      if (!site.equals(failedSite)) {
        foundLiveSite = true;
        tableNode.setAssignedSite(site);
      }
    }
    if (!foundLiveSite) {
      System.err.println("Unable to find live replicate"); 
      return;
    }
    synchronized (topNode) {
      if (topNode.allTableNodesAssigned())
        restartQuery(topNode);
    }
  }

// --------------------- End Interface MessageClient ---------------------

  synchronized void handleFault(AlgebraNode localNode, AlgebraNode failedNode, NodeHandle failedSite)
  {
    System.out.println("FAULT DETECTED:");
    System.out.println("  Detector: " + localNode);
    System.out.println("  Failed: " + failedNode);
    int receivedTuples = localNode.getReceivedTupleCount(failedNode);
    int tags[] = localNode.getLastPacketTags(failedNode);
    AlgebraNode restartNode = failedNode;
    if (failedNode instanceof CrashNode)
      restartNode = failedNode.getOperandNodes().get(0);
    localNode.removeCrashNodes(failedSite); // Remove crash nodes

    if (!Dascosa.isPropertySet(Dascosa.USE_COMPLETE_RESTART)) { // Partial restart
      if (restartNode instanceof PartialRestartable)
        restartMsg.put(restartNode, new PartialRestartFailedMsg(restartNode, tags));
      else
        restartMsg.put(restartNode, new CompleteRestartFailedMsg(restartNode, receivedTuples));
    }
    else
      localNode.removeOperandsFrom(restartNode);

    restartNode.removeFailedSite(failedSite);
    List<TableAccessNode> failedAccessNodes = restartNode.getUnassignedTableNodes();
    if (failedAccessNodes.isEmpty())
      restartQuery(restartNode);
    else {
      for (TableAccessNode tableNode : failedAccessNodes) {
        AbstractRequestMsg message = new LookupRequestMsg(tableNode.getTableName(), tableNode.getFragmentNr());
        requestResponseManager.sendRequest(message, tableNode.getTableName(), this,
                                           new Object[] {restartNode, tableNode, failedSite});
      }
    }
  }

  private synchronized void restartQuery(AlgebraNode topNode)
  {
    QueryPlanner.assignNodes(topNode);
    if (Dascosa.isPropertySet(Dascosa.USE_COMPLETE_RESTART))
      networkLayer.sendMessage(topNode, topNode.getAssignedSite());
    else
      networkLayer.sendMessage(restartMsg.remove(topNode), topNode.getAssignedSite());
  }
}
