package dascosa.site.fault;

import dascosa.algebra.AlgebraNode;
import dascosa.data.Packet;
import dascosa.data.PacketHeader;
import dascosa.site.NetworkLayer;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 28.aug.2007
 * Time: 21:21:44
 * To change this template use File | Settings | File Templates.
 */
public final class RestartCache
{
  private final Collection<CacheListener> cacheListeners = new ArrayList<CacheListener>();
  private final Map<Long, List<Packet>> packetCache = new HashMap<Long, List<Packet>>();
  private final Map<Long, PacketHeader> packetHeaderCache = new HashMap<Long, PacketHeader>();
  private final NetworkLayer networkLayer;

  public RestartCache(NetworkLayer networkLayer)
  {
    this.networkLayer = networkLayer;
  }

  public void addCacheListener(CacheListener cacheListener)
  {
    cacheListeners.add(cacheListener);
  }

  public synchronized void addPacket(Packet packet)
  {
    List<Packet> packetList = packetCache.get(packet.getSourceAlgebraNodeID());
    if (packetList == null) {
      packetList = new ArrayList<Packet>();
      packetCache.put(packet.getSourceAlgebraNodeID(), packetList);
    }
    packetList.add(packet);
    for (CacheListener cacheListener : cacheListeners)
      cacheListener.restartPacketAdded(packet);
  }

  public synchronized void addPacketHeader(PacketHeader packetHeader)
  {
    packetHeaderCache.put(packetHeader.getSourceAlgebraNodeID(), packetHeader);
    for (CacheListener cacheListener : cacheListeners)
      cacheListener.restartPacketAdded(packetHeader);
  }

  public synchronized void clearCache()
  {
    packetCache.clear();
    packetHeaderCache.clear();
    for (CacheListener cacheListener : cacheListeners)
      cacheListener.restartPacketsCleared();
  }

  public synchronized boolean lookup(AlgebraNode algebraNode, int tuplesToSkip)
  {
    PacketHeader packetHeader = packetHeaderCache.get(algebraNode.getNodeID());
    if (packetHeader == null)
      return false;
    System.out.println("Tuple cache hit on " + algebraNode);
    int index = -1;
    for (int i = 0; i < algebraNode.getResultNodeCount(); i++)
      if (algebraNode.getResultNodeID(i) == packetHeader.getTargetAlgebraNodeID())
        index = i;
    networkLayer.sendMessage(packetHeader, algebraNode.getResultSite(index));
    List<Packet> packetList;
    if (tuplesToSkip == 0)
      packetList = packetCache.get(algebraNode.getNodeID());
    else
      packetList = getPackets(algebraNode, tuplesToSkip);
    if (packetList != null)
      for (Packet packet : packetList)
        networkLayer.sendMessage(packet, algebraNode.getResultSite(index));
    return true;
  }

  public synchronized void removeFrom(AlgebraNode algebraNode)
  {
    PacketHeader packetHeader = packetHeaderCache.remove(algebraNode.getNodeID());
    if (packetHeader != null)
      for (CacheListener cacheListener : cacheListeners)
        cacheListener.restartPacketRemoved(packetHeader);
    List<Packet> packetList = packetCache.remove(algebraNode.getNodeID());
    if (packetList != null) {
      for (Packet packet : packetList)
        for (CacheListener cacheListener : cacheListeners)
          cacheListener.restartPacketRemoved(packet);
    }
  }

  synchronized List<Packet> getPackets(AlgebraNode algebraNode, int tuplesToSkip)
  {
    List<Packet> packetList = packetCache.get(algebraNode.getNodeID());
    if (packetList == null)
      return null;
    List<Packet> newList = new ArrayList<Packet>();
    for (Packet packet : packetList) {
      if (tuplesToSkip > 0 && tuplesToSkip < packet.getTupleCount())
        newList.add(new Packet(packet, tuplesToSkip)); // Partial packet
      else
        newList.add(packet);
    }
    return newList;
  }
}
