package dascosa.site.index;


import dascosa.Dascosa;
import dascosa.message.*;
import dascosa.message.request.*;
import dascosa.site.NetworkLayer;
import dascosa.site.SiteManager;
import dascosa.site.cache.investment.CacheInvestment;
import dascosa.site.cache.investment.Candidate;
import dascosa.site.query.optimizer.DataStatistics;
import dascosa.site.query.optimizer.LinkStatistics;
import dascosa.site.query.optimizer.ResultSetIdentifier;
import dascosa.util.FilteredOutput;
import edu.harvard.syrah.pyxida.nc.lib.Coordinate;
import rice.p2p.commonapi.NodeHandle;

import javax.swing.Timer;
import java.awt.event.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 23.apr.2007
 * Time: 10:13:47
 * To change this template use File | Settings | File Templates.
 */
public final class Index
{
  public static final int INDEX_CLEAR_INTERVAL = 30000;
  public static final int INDEX_PUBLISH_INTERVAL = 4000;
  private final Collection<IndexListener> listeners = new ArrayList<IndexListener>();
  private final Map<String, IndexTable> index = new HashMap<String, IndexTable>();
  private final Map<String, List<IndexCache>> cacheIndex = new HashMap<String, List<IndexCache>>();
  public final CacheInvestment cacheInvestment;
  private final NetworkLayer networkLayer;
  private final LinkStatistics linkStatistics;

  public Index(NetworkLayer networkLayer, SiteManager siteManager)
  {
    this.networkLayer = networkLayer;
    this.linkStatistics = LinkStatistics.getInstance(siteManager);
    this.cacheInvestment =  new CacheInvestment(siteManager, this);
    Timer clearTimer = new Timer(INDEX_CLEAR_INTERVAL, new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        // Remove old table index entries
        synchronized (index) {
          for (Iterator<IndexTable> iterator = index.values().iterator(); iterator.hasNext();) {
            IndexTable table = iterator.next();
            if (!table.hasValidNodes()) {
              iterator.remove();
              for (IndexListener listener : listeners)
                listener.indexEntryRemoved(table);
            }
          }
        }
        // Remove old cache index entries
        synchronized (cacheIndex) {
          for (Iterator<String> iterator = cacheIndex.keySet().iterator(); iterator.hasNext();) {
            String cacheName = iterator.next();
            List<IndexCache> indexCaches = cacheIndex.get(cacheName);
            for (Iterator<IndexCache> entryIterator = indexCaches.iterator(); entryIterator.hasNext();) {
              IndexCache indexCache = entryIterator.next();
              if (!indexCache.isValid()) {
                entryIterator.remove();
                for (IndexListener listener : listeners)
                  listener.indexEntryRemoved(indexCache);
              }
            }
            if (indexCaches.isEmpty()) {
              iterator.remove();
            }
          }
        }
      }
    });
    clearTimer.setRepeats(true);
    clearTimer.start();
  }

  public void addCacheEntry(PublishCacheMsg message)
  {
    String[] queries = message.getQueries();
    ResultSetIdentifier[] rsis = message.getResultSetIdentifier();
    for (int i = 0; i < queries.length; i++){
      IndexCache indexCache = new IndexCache(queries[i], rsis[i], message.getSite(), message.getCoordinate());
      synchronized (cacheIndex) {
        List<IndexCache> indexCaches = cacheIndex.get(queries[i]);
        if (indexCaches == null || indexCaches.isEmpty()) { // no existing caches of this query
          indexCaches = new LinkedList<IndexCache>();
          indexCaches.add(indexCache);
          cacheIndex.put(queries[i], indexCaches);
          for (IndexListener listener : listeners)
            listener.indexEntryAdded(indexCache);
        }
        else { // a list of caches already exists
          for (IndexCache ic : indexCaches) {
            if (ic.getSite() == indexCache.getSite()) {
              indexCache = ic;
              break;
            }
          }
        }
        indexCache.refresh();
      }
    }
  }

  public void addEntry(PublishTableMsg message)
  {
    synchronized (index) {
      IndexTable table = index.get(message.getTableName());
      if (table == null) {
        table = new IndexTable(message.getTableName(), message.getPrimaryKey(), message.getAttributeNames());
        index.put(message.getTableName(), table);
        for (IndexListener listener : listeners)
          listener.indexEntryAdded(table);
      }
      IndexFragment fragment = table.publish(message.getSite(), message.getFragmentNr(),
                                             message.getMinUsed(), message.getMaxUsed(), message.getTupleCount(),
                                             message.getHistograms(), message.getCoordinate());
      if (fragment.getMaster() == null) {
        fragment.selectNewMaster();
        for (IndexFragment f : table.getFragments()) {
          if (f.getMaster() != null) // Should not be necessary
            networkLayer.sendMessage(new MasterNotificationMsg(table.getTableName(), f.getFragmentNr(),
                                                               f.getPkAssigned()),
                                     f.getMaster().getSite());
        }
      }
    }
  }

  public void addListener(IndexListener listener)
  {
    synchronized (index) {
      listeners.add(listener);
    }
  }

  public void doLookup(LookupRequestMsg message) {
    Map<String, Set<NodeHandle>> caches = null;
    Map<ResultSetIdentifier, Set<NodeHandle>> cachesByRSI = null;

    Map<NodeHandle, Coordinate> cacheCoordinates = null;

    String query = message.getQuery();
    ResultSetIdentifier[] rsis = message.getResultSetIdentifier();

    if (Dascosa.isPropertySet(Dascosa.USE_SEMANTIC_CACHE) && query != null) {
      Set<Candidate> hints = cacheInvestment.getHints(rsis);

      caches = new HashMap<String, Set<NodeHandle>>();
      cachesByRSI = new HashMap<ResultSetIdentifier, Set<NodeHandle>>();
      cacheCoordinates = new HashMap<NodeHandle, Coordinate>();
      synchronized (cacheIndex) {
        for (String q : cacheIndex.keySet())
          if (Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER)) {
            List<IndexCache> indexCaches = cacheIndex.get(q);
            ResultSetIdentifier cacheRSI = indexCaches.get(0).getResultSetIdentifier();
            for (ResultSetIdentifier rsi : rsis)
              if (rsi.contains(cacheRSI)) {
                Set<NodeHandle> sites = new HashSet<NodeHandle>();
                for (IndexCache ic : cacheIndex.get(q)) {
                  sites.add(ic.getSite());
                  cacheCoordinates.put(ic.getSite(), ic.getCoordinate());
                }
                caches.put(q, sites);
                cachesByRSI.put(cacheIndex.get(q).get(0).getResultSetIdentifier(), sites);
              }
          } else if (query.lastIndexOf(q) != -1) { // A cached subquery is found
            Set<NodeHandle> sites = new HashSet<NodeHandle>();
            for (IndexCache ic : cacheIndex.get(q)) {
              sites.add(ic.getSite());
              cacheCoordinates.put(ic.getSite(), ic.getCoordinate());
            }
            caches.put(q, sites);
            cachesByRSI.put(cacheIndex.get(q).get(0).getResultSetIdentifier(), sites);
          }
      }
      for (Candidate candidate : hints) {
        Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: Suggesting candidate for " + candidate.getRSID().toString() + " at site " + candidate.getSite().toString());
        Set<NodeHandle> sites = cachesByRSI.get(candidate.getRSID());
        if (sites == null) {
          sites = new HashSet<NodeHandle>();
          cachesByRSI.put(candidate.getRSID(), sites);
        }
        if (!sites.contains(candidate.getRSID())) {
          sites.add(candidate.getSite());
          cacheCoordinates.put(candidate.getSite(), linkStatistics.getCoordinates(candidate.getSite()));
        }
      }
    }

    if (cachesByRSI != null)
      Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: CacheLookup -> RSID response " + cachesByRSI.toString());
    else
      Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: CacheLookup -> RSID response is NULL");

    IndexTable table;
    synchronized (index) {
      table = index.get(message.getTableName());
    }
    if (table == null) {
      networkLayer.sendMessage(new LookupResponseMsg(message, message.getTableName()), message.getReturnToSite());
    }
    else {
      String attributeNames[] = table.getAttributeNames();
      String primaryKey = table.getPrimaryKey();

      NodeHandle[][] sites = table.getDatabaseSites();
      Coordinate[][] coordinates = table.getSiteCoordinates();
      int fragmentTupleCounts[] = table.getFragmentTupleCounts();
      int masterIndex[] = table.getMasterIndex();
      int fragmentNumbers[] = table.getFragmentNumbers();
      if (message.getFragmentNr() == LookupRequestMsg.ALL_FRAGMENTS)
        networkLayer.sendMessage(new LookupResponseMsg(message, message.getTableName(), attributeNames,
                                                       primaryKey, sites, masterIndex,
                                                       coordinates, fragmentTupleCounts, caches, cachesByRSI,
                                                       fragmentNumbers, cacheCoordinates),
                                 message.getReturnToSite());
      else {
        int fragmentIndex = -1;
        for (int i = 0; i < fragmentNumbers.length; i++)
          if (fragmentNumbers[i] == message.getFragmentNr())
            fragmentIndex = i;
        if (fragmentIndex == -1) {
          System.err.println("Index.doLookup: Could not find fragment " + message.getFragmentNr() +
                             " of " + message.getTableName() );
          networkLayer.sendMessage(new LookupResponseMsg(message, message.getTableName()), message.getReturnToSite());
        }
        else {
          sites = new NodeHandle[][] {sites[fragmentIndex]};
          coordinates = new Coordinate[][] {coordinates[fragmentIndex]};
          fragmentTupleCounts = new int[] {fragmentTupleCounts[fragmentIndex]};
          masterIndex = new int[] {masterIndex[fragmentIndex]};
          fragmentNumbers = new int[] { fragmentNumbers[fragmentIndex] };
          networkLayer.sendMessage(new LookupResponseMsg(message, message.getTableName(),
                                                         message.getFragmentNr(), attributeNames,
                                                         primaryKey, sites, masterIndex,
                                                         coordinates, fragmentTupleCounts, caches, cachesByRSI,
                                                         fragmentNumbers, cacheCoordinates),
                                   message.getReturnToSite());
        }
      }
    }
  }

  public void doLookup(LookupPKRequestMsg message)
  {
    IndexTable table;
    synchronized (index) {
      table = index.get(message.getTableName());
    }
    if (table == null) {
      networkLayer.sendMessage(new LookupPKResponseMsg(message, message.getTableName(), -1,
                                                       null, -1),
                               message.getReturnToSite());
    }
    else {
      IndexFragment tableFragment = table.getAssignedFragment(message.getPkValue());
      NodeHandle sites[] = table.getDatabaseSites()[tableFragment.getFragmentNr() - 1];
      int masterIndex = table.getMasterIndex()[tableFragment.getFragmentNr() - 1];
      networkLayer.sendMessage(new LookupPKResponseMsg(message, message.getTableName(),
                                                       tableFragment.getFragmentNr(), sites, masterIndex),
                               message.getReturnToSite());
    }
  }

  public void doLookup(HistogramsRequestMsg message)
  {
    IndexTable table;
    synchronized (index) {
      table = index.get(message.identifier.tables.iterator().next());//TODO: Assumes only one table
    }
    if (DataStatistics.getInstance().hasHistogram(message.identifier)) {
      HistogramResponseMessage responseMsg = new HistogramResponseMessage(message, DataStatistics.getInstance().getHistogram(message.identifier));
      networkLayer.sendMessage(responseMsg, message.getReturnToSite());
    }
    else {
      IndexFragment fragment = table.getFragment(message.identifier.fragmentnr);
      NodeHandle site = fragment.getMaster().getSite(); //TODO: Select closest replikate instead of master
      message.requestCreation = true;
      message.originatingNode = message.getReturnToSite();
      networkLayer.sendMessage(message, site);
    }
    //NodeHandle site = table.getFragment(message.identifier.fragmentnr).getMaster().getSite();
  }

  public String[] getTableNames()
  {
    synchronized (index) {
      String tableNames[] = new String[index.size()];
      int i = 0;
      for (String tableName : index.keySet())
        tableNames[i++] = tableName;
      return tableNames;
    }
  }
}
