package dascosa.site.index;

import dascosa.site.query.optimizer.Histogram;
import edu.harvard.syrah.pyxida.nc.lib.Coordinate;
import rice.p2p.commonapi.NodeHandle;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 20.apr.2007
 * Time: 15:14:01
 * To change this template use File | Settings | File Templates.
 */
final class IndexTable
{
  private final Map<Integer, IndexFragment> fragments = new HashMap<Integer, IndexFragment>();
  private final String tableName;
  private final String attributeNames[];
  private final String primaryKey;
  private final int pkMin;
  private final int pkMax;
  
  IndexTable(String tableName, String primaryKey, String attributeNames[])
  {
    this.tableName = tableName;
    this.primaryKey = primaryKey;
    this.attributeNames = attributeNames;
    pkMin = Integer.MIN_VALUE; // TODO: Assuming INTEGER as primary key datatype
    pkMax = Integer.MAX_VALUE;
  }

  public synchronized String toString()
  {
    return tableName + " => " + fragments.values();
  }

  public String getPrimaryKey()
  {
    return primaryKey;
  }

  synchronized IndexFragment getAssignedFragment(int pkValue)
  {
    for (IndexFragment fragment : fragments.values())
      if (fragment.getPkAssigned().contains(pkValue))
        return fragment;
    return null; // TODO: Find a good fragment or make one
  }

  String[] getAttributeNames()
  {
    return attributeNames;
  }

  synchronized NodeHandle[][] getDatabaseSites()
  {
    NodeHandle[][] databaseSites = new NodeHandle[fragments.size()][];
    int index = 0;
    for (IndexFragment fragment : fragments.values())
      databaseSites[index++] = fragment.getDatabaseSites();
    return databaseSites;
  }

  synchronized IndexFragment getFragment(int fragmentNr)
  {
    return fragments.get(fragmentNr);
  }

  synchronized int[] getFragmentTupleCounts()
  {
    int result[] = new int[fragments.size()];
    int index = 0;
    for (IndexFragment f : fragments.values())
      result[index++] = f.getTupleCount();
    return result;
  }

  synchronized int[] getFragmentNumbers()
  {
    int numbers[] = new int[fragments.size()];
    int index = 0;
    for (int fragmentNr: fragments.keySet())
      numbers[index++] = fragmentNr;
    return numbers;
  }

  synchronized Collection<IndexFragment> getFragments()
  {
    return fragments.values();
  }

  synchronized int[] getMasterIndex()
  {
    int masterIndex[] = new int[fragments.size()];
    int index = 0;
    for (IndexFragment fragment: fragments.values())
      masterIndex[index++] = fragment.getMasterIndex();
    return masterIndex;
  }

  synchronized Coordinate[][] getSiteCoordinates()
  {
    Coordinate[][] siteCoordinates = new Coordinate[fragments.size()][];
    int index = 0;
    for (IndexFragment fragment: fragments.values())
      siteCoordinates[index++] = fragment.getSiteCoordinates();
    return siteCoordinates;
  }

  String getTableName()
  {
    return tableName;
  }

  synchronized boolean hasValidNodes()
  {
    for (Iterator<IndexFragment> iterator = fragments.values().iterator(); iterator.hasNext();) {
      IndexFragment fragment = iterator.next();
      if (!fragment.hasValidNodes())
        iterator.remove();
    }
    return !fragments.isEmpty();
  }

  synchronized IndexFragment publish(NodeHandle site,  int fragmentNr, Integer minUsed, Integer maxUsed, Integer tupleCount, HashMap<String, Histogram> histograms, Coordinate coordinate)
  {
    IndexFragment fragment = fragments.get(fragmentNr);
    if (fragment == null) { // New fragment
      fragment = new IndexFragment(fragmentNr, minUsed, maxUsed, tupleCount, histograms);
      fragments.put(fragment.getFragmentNr(), fragment);
      reassignFragmentValuespaces();
    }
    else { // Existing fragment
      if (fragment.hasReplicate(site)) { // Update
        boolean changed = fragment.update(minUsed, maxUsed);
        if (changed)
          reassignFragmentValuespaces();
      }
    }
    fragment.addReplicate(site, coordinate);
    return fragment;
  }

  synchronized void reassignFragmentValuespaces()
  {
    List<IndexFragment> tempFragments = new ArrayList<IndexFragment>(fragments.values());
    Collections.sort(tempFragments);

    // Check for overlaps
    IndexFragment fragment1 = tempFragments.get(0);
    for (int i = 1; i < tempFragments.size(); i++) {
      IndexFragment fragment2 = tempFragments.get(i);
      Interval interval1 = fragment1.getPkUsed();
      Interval interval2 = fragment2.getPkUsed();
      if (interval1 != null && interval2 != null && interval1.intersects(interval2)) {
        System.err.println("Fragment overlap detected: " + tableName);
        System.err.println(fragment1 + " overlaps " + fragment2);
        return;
      }
      fragment1 = fragment2;
    }

    // Algorithm
    fragment1 = tempFragments.get(0);
    int min = pkMin;
    for (int i = 1; i < tempFragments.size(); i++) {
      IndexFragment fragment2 = tempFragments.get(i);
      Interval interval1 = fragment1.getPkUsed();
      Interval interval2 = fragment2.getPkUsed();
      if (interval1 != null && interval2 != null) {
        double diff = (interval2.getMin() - interval1.getMax()) / 2.0d;
        int max = (int)Math.floor(interval1.getMax() + diff);
        fragment1.setPkAssigned(new Interval(min, max));
        min = max + 1;
      }
      fragment1 = fragment2;
    }
    fragment1.setPkAssigned(new Interval(min, pkMax));
  }
}
