package dascosa.site.localdb;

import dascosa.message.FragmentReceivedMsg;
import dascosa.message.request.*;
import dascosa.site.*;
import dascosa.site.update.InvalidateReplicatesClient;
import dascosa.site.update.InvalidateReplicatesHandler;
import rice.p2p.commonapi.NodeHandle;

import javax.sql.DataSource;
import java.io.*;
import java.sql.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 09.jul.2008
 * Time: 12:01:42
 * To change this template use File | Settings | File Templates.
 */
public final class FragmentSplitter implements MessageClient, InvalidateReplicatesClient
{
  private final DataSource dataSource;
  private final InvalidateReplicatesHandler invalidateReplicatesHandler;
  private final LocalFragmentHandler localFragmentHandler;
  private final Map<String, Integer> oldFragments = new HashMap<String, Integer>();
  private final Map<String, OutputStream> openFiles = new HashMap<String, OutputStream>();
  private final NetworkLayer networkLayer;
  private final RequestResponseManager requestResponseManager;
  private final SiteManager siteManager;
  private final byte[] buffer = new byte[FragmentCopier.PACKET_SIZE];

  public FragmentSplitter(DataSource dataSource, RequestResponseManager requestResponseManager,
                          SiteManager siteManager, InvalidateReplicatesHandler invalidateReplicatesHandler,
                          LocalFragmentHandler localFragmentHandler, NetworkLayer networkLayer)
  {
    this.dataSource = dataSource;
    this.requestResponseManager = requestResponseManager;
    this.siteManager = siteManager;
    this.invalidateReplicatesHandler = invalidateReplicatesHandler;
    this.localFragmentHandler = localFragmentHandler;
    this.networkLayer = networkLayer;
  }

// --------------------- Begin Interface InvalidateReplicatesClient ---------------------

  // Executed at Master

  public void replicatesInvalidated(String tableName, int fragmentNr, Object parameters[], DataSource dataSource)
  {
    executeSplitMigrate((FragmentSplitRequestMsg)parameters[0]);
  }

// --------------------- End Interface InvalidateReplicatesClient ---------------------


// --------------------- Begin Interface MessageClient ---------------------

  public void responseReceived(AbstractResponseMsg msg, Object parameters[], DataSource dataSource)
  {
    if (msg instanceof LookupResponseMsg) { // Initiator site
      LookupResponseMsg message = (LookupResponseMsg)msg;
      int oldFragmentNr = (Integer)parameters[0];
      long minValue = (Long)parameters[1];
      int masterIndexes[] = message.getMasterIndex();
      int fragmentIndex = message.getFragmentIndex(oldFragmentNr);
      AbstractRequestMsg request = new FragmentSplitRequestMsg(message.getTableName(), oldFragmentNr,
                                                               masterIndexes.length + 1,
                                                               minValue, message.getSites()[fragmentIndex]);
      NodeHandle masterSite = message.getSites()[fragmentIndex][masterIndexes[fragmentIndex]];
      requestResponseManager.sendRequest(request, masterSite, this, new Object[] {masterSite});
    }
    else if (msg instanceof LockResponseMsg) { // Master site
      LockResponseMsg message = (LockResponseMsg)msg;
      FragmentSplitRequestMsg request = (FragmentSplitRequestMsg)parameters[0];
      if (message.isGranted()) {
        if (request.getReplicateSites().length > 1) { // Must invalidate
          invalidateReplicatesHandler.invalidateReplicates(message.getTableName(), message.getFragmentNr(),
                                                           request.getReplicateSites(), this,
                                                           parameters);
        }
        else
          executeSplitMigrate(request);
      }
      else {
        System.err.println("FragmentSplitter: Lock denied " + request);
      }
    }
    else if (msg instanceof FragmentSplitResponseMsg) { // Initiator site
      FragmentSplitResponseMsg message = (FragmentSplitResponseMsg)msg;
      try {
        String physicalName = message.getTableName() + '_' + message.getFragmentNr();
        synchronized (openFiles) {
          OutputStream os = openFiles.get(physicalName);
          if (os == null) {
            os = new BufferedOutputStream(new FileOutputStream(physicalName + ".imp"));
            openFiles.put(physicalName, os);
          }
          os.write(message.getData());
          if (message.isLast()) {
            os.close();
            openFiles.remove(physicalName);
            Connection connection = dataSource.getConnection();
            Statement statement = connection.createStatement();
            statement.executeUpdate(message.getDefinition());
            statement.close();
            if (new File(physicalName + ".imp").length() > 0) {
              PreparedStatement importSt = connection.prepareStatement("CALL SYSCS_UTIL.SYSCS_IMPORT_TABLE (?,?,?,?,?,?,?)");
              importSt.clearParameters();
              importSt.setString(1, null);
              importSt.setString(2, physicalName.toUpperCase());
              importSt.setString(3, physicalName + ".imp");
              importSt.setString(4, "%");
              importSt.setString(5, null);
              importSt.setString(6, null);
              importSt.setInt(7, 1);
              importSt.execute();
              importSt.close();
            }
            connection.close();
            networkLayer.sendMessage(new FragmentReceivedMsg(message.getTableName(), message.getFragmentNr(), true),
                                     (NodeHandle)parameters[0]);
            localFragmentHandler.initializeTableFragment(message.getTableName(), message.getFragmentNr(), dataSource);
            siteManager.fireDebugMessageEvent(message.getTableName() + " fragment " + message.getFragmentNr() + " written locally.\n");
          }
        }
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      } catch (IOException e) {
        e.printStackTrace();
      } catch (SQLException e) {
        e.printStackTrace();
      }
    }
  }

// --------------------- End Interface MessageClient ---------------------


  // Executed at master
  public void fragmentReceived(FragmentReceivedMsg message, DataSource dataSource)
  {
    if (message.isSuccess()) {
      // Delete
      try {
        Connection connection = dataSource.getConnection();
        Statement statement = connection.createStatement();
        statement.executeUpdate("DROP TABLE " + message.getTableName() + '_' + message.getFragmentNr());
        statement.close();
        connection.close();
      } catch (SQLException e) {
        System.err.println("FragmentSplitter: Error deleting old fragment");
        e.printStackTrace();
      }

      int oldFragmentNr;
      synchronized (oldFragments) {
        oldFragmentNr = oldFragments.remove(message.getTableName() + '_' + message.getFragmentNr());
      }

      // Republish
      try {
        localFragmentHandler.initializeTableFragment(message.getTableName(), oldFragmentNr, dataSource);
      } catch (SQLException e) {
        e.printStackTrace();
      }

      // Unlock
      AbstractRequestMsg unlockMsg = new LockRequestMsg(message.getTableName(), oldFragmentNr, false);
      unlockMsg.setReturnToSite(siteManager.getLocalSite());
      networkLayer.sendMessage(unlockMsg, siteManager.getLocalSite());
    }
    else {
      System.err.println("FragmentSplitter: Sending of fragment failed");
    }
  }

  // Entry point - initiator
  public void splitFragment(String sql)
  {
    StringTokenizer st = new StringTokenizer(sql);
    st.nextToken();
    String tableName = st.nextToken();
    int fragmentNr = Integer.parseInt(st.nextToken());
    long minValue = Long.parseLong(st.nextToken());
    AbstractRequestMsg message = new LookupRequestMsg(tableName, LookupRequestMsg.ALL_FRAGMENTS);
    requestResponseManager.sendRequest(message, tableName, this, new Object[] {fragmentNr,
                                                                               minValue});
  }

  // Executed at master
  public void splitRequested(FragmentSplitRequestMsg message)
  {
    AbstractRequestMsg lockRequest = new LockRequestMsg(message.getTableName(), message.getOldFragmentNr(),
                                                        true);
    requestResponseManager.sendRequest(lockRequest, siteManager.getLocalSite(), this, new Object[] {message});
  }

  // Executed at master

  private void executeSplitMigrate(FragmentSplitRequestMsg message)
  {
    TableFragment oldTableFragment = localFragmentHandler.getTableFragment(message.getTableName(),
                                                                           message.getOldFragmentNr());
    try {
      // Split
      String definition = splitFragment(oldTableFragment, message.getNewFragmentNr(), message.getMinValue());
      sendCopy(message, dataSource, definition);
      synchronized (oldFragments) {
        oldFragments.put(message.getTableName() + '_' + message.getNewFragmentNr(), message.getOldFragmentNr());
      }
    } catch (SQLException e) {
      e.printStackTrace();
    }
  }

  // Executed at master
  private void sendCopy(FragmentSplitRequestMsg message, DataSource dataSource, String definition)
  {
    try {
      // Data
      Connection connection = dataSource.getConnection();
      String physicalName = message.getTableName() + '_' + message.getNewFragmentNr();
      new File(physicalName + ".exp").delete();
      PreparedStatement exportSt = connection.prepareStatement("CALL SYSCS_UTIL.SYSCS_EXPORT_TABLE (?,?,?,?,?,?)");
      exportSt.clearParameters();
      exportSt.setString(1, null);
      exportSt.setString(2, physicalName.toUpperCase());
      exportSt.setString(3, physicalName + ".exp");
      exportSt.setString(4, "%");
      exportSt.setString(5, null);
      exportSt.setString(6, null);
      exportSt.execute();
      exportSt.close();
      connection.close();

      InputStream is = new BufferedInputStream(new FileInputStream(physicalName + ".exp"));
      synchronized (buffer) {
        boolean isDone = false;
        while (!isDone) {
          int bytesRead = FragmentCopier.readBytes(is, buffer);
          if (bytesRead == buffer.length)
            networkLayer.sendMessage(new FragmentSplitResponseMsg(message, message.getTableName(),
                                                                  message.getNewFragmentNr(), definition, buffer, false),
                                     message.getReturnToSite());
          else if (bytesRead == -1) {
            networkLayer.sendMessage(new FragmentSplitResponseMsg(message, message.getTableName(),
                                                                  message.getNewFragmentNr(), definition, new byte[0], true),
                                     message.getReturnToSite());
            isDone = true;
          }
          else {
            byte b[] = new byte[bytesRead];
            System.arraycopy(buffer, 0, b, 0, bytesRead);
            networkLayer.sendMessage(new FragmentSplitResponseMsg(message, message.getTableName(),
                                                                  message.getNewFragmentNr(), definition, b, true),
                                     message.getReturnToSite());
            isDone = true;
          }
        }
      }
      is.close();
    } catch (SQLException e) {
      e.printStackTrace();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  // Executed at master
  String splitFragment(TableFragment tableFragment, int newFragmentNr, long minValue) throws SQLException
  {
    // Definition
    String oldFragment = tableFragment.getPhysicalTableName();
    String newFragment = tableFragment.getTableName() + '_' + newFragmentNr;
    String definition = FragmentSender.generateCreate(dataSource, oldFragment, newFragment);

    Connection connection = dataSource.getConnection();
    Statement statement = connection.createStatement();

    // Create new fragment
    statement.executeUpdate(definition);

    // Insert data into new fragment
    statement.executeUpdate("INSERT INTO " + newFragment +
                            " SELECT * FROM " + oldFragment +
                            " WHERE " + tableFragment.getPrimaryKeys()[0] + " >= " + minValue);

    // Remove data from old fragment
    statement.executeUpdate("DELETE FROM " + oldFragment +
                            " WHERE " + tableFragment.getPrimaryKeys()[0] + " >= " + minValue);

    statement.close();
    connection.close();

    return definition;
  }
}
