package dascosa.site.localdb;

import dascosa.Dascosa;
import dascosa.algebra.AlgebraNode;
import dascosa.algebra.scan.TableAccessNode;
import dascosa.data.FinalResult;
import dascosa.message.*;
import dascosa.message.request.*;
import dascosa.site.NetworkLayer;
import dascosa.site.SiteManager;
import dascosa.site.update.LockManager;
import dascosa.site.index.Index;
import dascosa.site.query.QueryListener;
import dascosa.site.query.optimizer.DataStatistics;
import dascosa.site.query.optimizer.Histogram;
import dascosa.util.FilteredOutput;
import rice.p2p.commonapi.Message;

import javax.sql.DataSource;
import javax.swing.Timer;
import java.awt.event.*;
import java.sql.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 28.aug.2007
 * Time: 21:14:23
 * To change this template use File | Settings | File Templates.
 */
public final class LocalFragmentHandler implements ActionListener, QueryListener
{
  private final Collection<DatabaseListener> databaseListeners = new ArrayList<DatabaseListener>();
  private final Collection<TableFragment> localReplicatesCopy = new ArrayList<TableFragment>();
  private final Map<String, List<TableFragment>> localReplicates = new HashMap<String, List<TableFragment>>();
  private final NetworkLayer networkLayer;
  private final SiteManager siteManager;
  private final Timer publishTimer;

  public boolean existsIn(String tableName) {
    for (String table : localReplicates.keySet()) {
      if (table.equals(tableName)) {
        return true;
      }
    }
    return false;
  }

  private static void readMetadata(TableFragment fragment, DataSource dataSource) throws SQLException
  {
    Connection connection = dataSource.getConnection();
    Statement statement = connection.createStatement();

    // Attribute names
    ResultSet rs = connection.getMetaData().getColumns(null, null, fragment.getPhysicalTableName().toUpperCase(), null);
    List<String> attributeNames = new ArrayList<String>();
    while (rs.next())
      attributeNames.add(rs.getString(4));
    rs.close();
    fragment.setAttributeNames(attributeNames.toArray(new String[attributeNames.size()]));

    // Primary keys
    rs = connection.getMetaData().getPrimaryKeys(null, null, fragment.getPhysicalTableName().toUpperCase());
    while (rs.next())
      fragment.setPrimaryKey(rs.getString(4), rs.getShort(5));
    rs.close();

    // Min-max values from primary keys
    for (String primaryKey : fragment.getPrimaryKeys()) {
      rs = statement.executeQuery("SELECT MIN(" + primaryKey + "), MAX(" + primaryKey + ") FROM " +
                                  fragment.getPhysicalTableName());
      if (rs.next())
        fragment.setPrimaryKeyRange(primaryKey, (Integer)rs.getObject(1), (Integer)rs.getObject(2));
      rs.close();
    }

    // Tuple count
    rs = statement.executeQuery("SELECT COUNT(*) FROM " + fragment.getPhysicalTableName());
    if (rs.next())
      fragment.setFragmentSize(rs.getInt(1));
    rs.close();
    /*
    for (String attribute: fragment.getPrimaryKeys()){
    	Histogram h = new Histogram(fragment, attribute, readConnection);
    	Dascosa.out.println(FilteredOutput.localdb, "readMetadata: Created histogram for: "+h.identifier);
    	fragment.setHistogram(attribute, h);
    	//Dascosa.out.println(FilteredOutput.localdb, h.toString());
    }
    */
    statement.close();
    connection.close();
  }

  public LocalFragmentHandler(SiteManager siteManager, NetworkLayer networkLayer,
                              DatabaseListener databaseListener, DataSource dataSource) throws SQLException
  {
    this.networkLayer = networkLayer;
    this.siteManager = siteManager;
    publishTimer = new Timer(Index.INDEX_PUBLISH_INTERVAL, this);
    publishTimer.setRepeats(true);
    if (databaseListener != null)
      addDatabaseListener(databaseListener);
    initializeTableFragments(dataSource);
  }

// --------------------- Begin Interface ActionListener ---------------------

  public void actionPerformed(ActionEvent e)
  {
    synchronized (localReplicates) {
      for (Collection<TableFragment> fragments: localReplicates.values()) {
        for (TableFragment fragment : fragments) {
          Message message = new PublishTableMsg(siteManager.getLocalSite(), fragment.getTableName(),
                                                fragment.getFragmentNr(), fragment.getPrimaryKeys()[0],
                                                fragment.getAttributeNames(), fragment.getMinUsed()[0],
                                                fragment.getMaxUsed()[0], fragment.getTupleCount(),
                                                fragment.getHistograms(),
                                                siteManager.getNCClient().getApplicationCoords());
          networkLayer.sendMessage(message, fragment.getTableName().toLowerCase());
        }
      }
    }
  }

// --------------------- End Interface ActionListener ---------------------


// --------------------- Begin Interface QueryListener ---------------------

  public void algebraNodeCompleted(AlgebraNode algebraNode)
  {
  }

  public void algebraNodePaused(AlgebraNode algebraNode)
  {
  }

  public void algebraNodeReceived(AlgebraNode algebraNode)
  {
    if (algebraNode instanceof TableAccessNode) {
      TableAccessNode tableAccessNode = (TableAccessNode)algebraNode;
      tableAccessNode.setTableFragment(getTableFragment(tableAccessNode.getTableName(),
                                                        tableAccessNode.getFragmentNr()));
    }
  }

  public void algebraNodeRemoved(AlgebraNode algebraNode)
  {
  }

  public void algebraNodeResumed(AlgebraNode algebraNode)
  {
  }

  public void algebraNodesCleared()
  {
  }

  public void queryCompleted(FinalResult result)
  {
  }

  public void queryStatsReceived(QueryStatsMsg queryStatsMsg)
  {
  }

// --------------------- End Interface QueryListener ---------------------

  public void addDatabaseListener(DatabaseListener databaseListener)
  {
    databaseListeners.add(databaseListener);
  }

  public TableFragment getTableFragment(String tableName, int fragmentNr)
  {
    synchronized (localReplicates) {
      for (TableFragment fragment : localReplicates.get(tableName.toLowerCase()))
        if (fragment.getFragmentNr() == fragmentNr)
          return fragment;
    }
    return null;
  }

  public Collection<TableFragment> getTableFragments()
  {
    localReplicatesCopy.clear();
    synchronized (localReplicates) {
      for (List<TableFragment> fragments: localReplicates.values())
        localReplicatesCopy.addAll(fragments);
    }
    return localReplicatesCopy;
  }

  public TableFragment initializeTableFragment(String tableName, int fragmentNr, DataSource dataSource) throws SQLException
  {
    TableFragment fragment = new TableFragment(tableName, fragmentNr);
    LocalFragmentHandler.readMetadata(fragment, dataSource);

    Message message = new PublishTableMsg(siteManager.getLocalSite(), fragment.getTableName(),
                                          fragment.getFragmentNr(), fragment.getPrimaryKeys()[0],
                                          fragment.getAttributeNames(), fragment.getMinUsed()[0],
                                          fragment.getMaxUsed()[0], fragment.getTupleCount(), fragment.getHistograms(),
                                          siteManager.getNCClient().getApplicationCoords());
    networkLayer.sendMessage(message, fragment.getTableName().toLowerCase());

    synchronized (localReplicates) {
      List<TableFragment> fragments = localReplicates.get(tableName.toLowerCase());
      if (fragments == null) {
        fragments = new ArrayList<TableFragment>();
        localReplicates.put(tableName.toLowerCase(), fragments);
      }
      else {
        for (Iterator<TableFragment> iterator = fragments.iterator(); iterator.hasNext();) {
          TableFragment oldFragment = iterator.next();
          if (oldFragment.getFragmentNr() == fragmentNr) {
            iterator.remove();
            for (DatabaseListener databaseListener : databaseListeners)
              databaseListener.localReplicateRemoved(oldFragment);
            if (oldFragment.isMaster())
              fragment.setMaster(true);
          }
        }
      }
      fragments.add(fragment);
    }
    for (DatabaseListener databaseListener : databaseListeners)
      databaseListener.localReplicateAdded(fragment);
    return fragment;
  }

  public void processDeletionMsg(DeleteReplicateRequestMsg message, DataSource dataSource)
  {
    try {
      deleteTableFragment(message.getTableName(), message.getFragmentNr(), dataSource);
      networkLayer.sendMessage(new DeleteReplicateResponseMsg(message, message.getTableName(),
                                                              message.getFragmentNr()), message.getReturnToSite());
    } catch (SQLException e) {
      e.printStackTrace();
      // TODO: Error handling
    }
  }

  public void processHistogramMsg(HistogramsRequestMsg msg, DataSource dataSource)
  {
    Histogram h = null;
    if (DataStatistics.getInstance().hasHistogram(msg.identifier))
      h = DataStatistics.getInstance().getHistogram(msg.identifier);
    else {
      try {
        h = new Histogram(getTableFragment(msg.identifier.tables.iterator().next(), msg.identifier.fragmentnr), msg.identifier.attribute, dataSource);
      } catch (SQLException e) {
        e.printStackTrace();
      }
    }
    HistogramResponseMessage response = new HistogramResponseMessage(msg, h);
    Dascosa.out.println(FilteredOutput.DATA_STATISTICS, "Created histogram with: "+h.identifier +" shipping it to: "//+msg.getReturnToSite() + " and "
                                                        +msg.originatingNode);
    networkLayer.sendMessage(response, msg.originatingNode);
    //networkLayer.sendMessage(response, msg.getReturnToSite());
  }

  public void processMasterMsg(MasterNotificationMsg message, LockManager lockManager, DataSource dataSource)
  {
    TableFragment fragment = getTableFragment(message.getTableName(), message.getFragmentNr());
    if (fragment != null) {
      fragment.setMaster(true);
      fragment.setAssignedInterval(message.getAssigned());
      if (Dascosa.isPropertySet(Dascosa.USE_DYNAMIC_REFRAGMENTATION))
        checkFragmentMerges(fragment, lockManager, dataSource);
    }
  }

  public void startPublish()
  {
    publishTimer.start();
  }

  public void stopPublish()
  {
    publishTimer.stop();
  }

  void deleteTableFragment(String tableName, int fragmentNr, DataSource dataSource) throws SQLException
  {
    Connection connection = dataSource.getConnection();
    Statement statement = connection.createStatement();
    statement.executeUpdate("DROP TABLE " + tableName.toUpperCase() + '_' + fragmentNr);
    statement.close();
    connection.close();
    synchronized (localReplicates) {
      List<TableFragment> fragments = localReplicates.get(tableName.toLowerCase());
      for (Iterator<TableFragment> iterator = fragments.iterator(); iterator.hasNext();) {
        TableFragment fragment = iterator.next();
        if (fragment.getFragmentNr() == fragmentNr) {
          iterator.remove();
          for (DatabaseListener databaseListener : databaseListeners)
            databaseListener.localReplicateRemoved(fragment);
          return;
        }
      }
    }
  }

  private void initializeTableFragments(DataSource dataSource) throws SQLException
  {
    Connection connection = dataSource.getConnection();

    ResultSet rs = connection.getMetaData().getTables(null, null, null, new String[] {"TABLE"});
    while (rs.next()) {
      TableFragment tableFragment = new TableFragment(rs.getString(3).toLowerCase());
      List<TableFragment> fragments = localReplicates.get(tableFragment.getTableName().toLowerCase());
      if (fragments == null) {
        fragments = new ArrayList<TableFragment>();
        localReplicates.put(tableFragment.getTableName().toLowerCase(), fragments);
      }
      fragments.add(tableFragment);
    }
    rs.close();
//    ResultSet rs2 = connection.getMetaData().getTables(null, null, null, new String[] { "VIEW" });
//    while (rs2.next()){
//    	System.out.println(rs2.getString(3).toLowerCase());
//      localReplicates.add(new TableFragment(rs2.getString(3).toLowerCase()));
//    }
//    rs2.close();
    connection.close();

    for (List<TableFragment> fragments: localReplicates.values()) {
      for (TableFragment fragment : fragments) {
        LocalFragmentHandler.readMetadata(fragment, dataSource);
        for (DatabaseListener databaseListener : databaseListeners)
          databaseListener.localReplicateAdded(fragment);
      }
    }
  }

  private void checkFragmentMerges(TableFragment changedFragment, LockManager lockManager,
                                   DataSource dataSource)
  {
    if (lockManager.isLocked(changedFragment.getTableName(), changedFragment.getFragmentNr()))
      return;
    synchronized (localReplicates) {
      List<TableFragment> fragments = localReplicates.get(changedFragment.getTableName().toLowerCase());
      boolean hasMerged = true;
      while (hasMerged) {
        hasMerged = false;
        for (int i = 0; !hasMerged && i < fragments.size() - 1; i++) {
          TableFragment fragment = fragments.get(i);
          if (fragment != changedFragment && fragment.isMaster() && changedFragment.isMaster() &&
              !lockManager.isLocked(fragment.getTableName(), fragment.getFragmentNr())) {
            long minMax = Math.min(fragment.getMaxAssigned(), changedFragment.getMaxAssigned());
            long maxMin = Math.max(fragment.getMinAssigned(), changedFragment.getMinAssigned());
            if (minMax + 1 == maxMin) {
              changedFragment = mergeFragments(changedFragment, fragment, lockManager, dataSource);
              if (changedFragment == null)
                return; // Error occured
              hasMerged = true;
            }
          }
        }
      }
    }
  }

  private TableFragment mergeFragments(TableFragment fragment1, TableFragment fragment2,
                                       LockManager lockManager, DataSource dataSource)
  { // TODO: Delete or update replicas!!!
    if (fragment1.getFragmentNr() > fragment2.getFragmentNr()) { // Swap
      TableFragment temp = fragment1;
      fragment1 = fragment2;
      fragment2 = temp;
    }

    // 1: Lock fragments
    boolean lock = lockManager.localLockRequested(fragment1.getTableName(), fragment1.getFragmentNr(),
                                                  true, siteManager.getLocalSite());
    if (!lock) {
      System.err.println("Unable to lock " + fragment1 + " - Merge aborted");
      return null;
    }
    lock = lockManager.localLockRequested(fragment2.getTableName(), fragment2.getFragmentNr(),
                                          true, siteManager.getLocalSite());
    if (!lock) {
      lockManager.localLockRequested(fragment1.getTableName(), fragment1.getFragmentNr(),
                                     false, siteManager.getLocalSite());
      System.err.println("Unable to lock " + fragment2 + " - Merge aborted");
      return null;
    }

    try {
      // 2: Insert data from fragment2 into fragment1
      Connection connection = dataSource.getConnection();
      Statement statement = connection.createStatement();

      statement.executeUpdate("INSERT INTO " + fragment1.getPhysicalTableName() +
                              " SELECT * FROM " + fragment2.getPhysicalTableName());
      

      // 3: Delete fragment2
      statement.executeUpdate("DROP TABLE " + fragment2.getPhysicalTableName());
      statement.close();
      connection.close();

      synchronized (localReplicates) {
        List<TableFragment> fragments = localReplicates.get(fragment2.getTableName().toLowerCase());
        fragments.remove(fragment2);
        for (DatabaseListener databaseListener : databaseListeners)
          databaseListener.localReplicateRemoved(fragment2);
      }

      // 4: Reinitialize new fragment
      return initializeTableFragment(fragment1.getTableName(), fragment1.getFragmentNr(), dataSource);

    } catch (SQLException e) {
      e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      return null;
    }
  }
}
