package dascosa.site.localdb;

import dascosa.message.request.*;
import dascosa.site.*;
import dascosa.site.dynfrag.FragmentationSuggestion;
import dascosa.site.update.*;

import javax.sql.DataSource;
import java.sql.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 14.aug.2008
 * Time: 14:35:45
 * To change this template use File | Settings | File Templates.
 */
public final class RefragmentHandler implements MessageClient, InvalidateReplicatesClient
{
  private final Collection<FragmentationSuggestion> activeSuggestions = new ArrayList<FragmentationSuggestion>();
  private final FragmentSender fragmentSender;
  private final InvalidateReplicatesHandler invalidateReplicatesHandler;
  private final LocalFragmentHandler localFragmentHandler;
  private final LockManager lockManager;
  private final RequestResponseManager requestResponseManager;
  private final SiteManager siteManager;

  private static int findFragmentNumber(int fragmentNumbers[], int alreadyUsed)
  {
    Arrays.sort(fragmentNumbers);
    for (int i = 1; i <= fragmentNumbers.length; i++) {
      if (i != alreadyUsed && Arrays.binarySearch(fragmentNumbers, i) < 0)
        return i;
    }
    if (alreadyUsed == fragmentNumbers.length + 1)
      return fragmentNumbers.length + 2;
    return fragmentNumbers.length + 1;
  }

  private static void splitFragment(DataSource dataSource, TableFragment tableFragment, int newFragmentNr, long minValue) throws SQLException
  {
    System.out.println("Splitting " + tableFragment);
    System.out.println("New fragment nr: " + newFragmentNr + " Split at " + minValue);
    // Definition
    String oldFragment = tableFragment.getPhysicalTableName();
    String newFragment = tableFragment.getTableName() + '_' + newFragmentNr;
    String definition = FragmentSender.generateCreate(dataSource, oldFragment, newFragment);

    Connection connection = dataSource.getConnection();
    Statement statement = connection.createStatement();

    // Create new fragment
    statement.executeUpdate(definition);

    // Insert data into new fragment
    statement.executeUpdate("INSERT INTO " + newFragment +
                            " SELECT * FROM " + oldFragment +
                            " WHERE " + tableFragment.getPrimaryKeys()[0] + " >= " + minValue);

    // Remove data from old fragment
    statement.executeUpdate("DELETE FROM " + oldFragment +
                            " WHERE " + tableFragment.getPrimaryKeys()[0] + " >= " + minValue);

    statement.close();
    connection.close();
  }

//  private final Statement readStatement2; // Not the same statement as MessageHandler

  public RefragmentHandler(RequestResponseManager requestResponseManager,
                           InvalidateReplicatesHandler invalidateReplicatesHandler,
                           LocalFragmentHandler localFragmentHandler, NetworkLayer networkLayer,
                           SiteManager siteManager, LockManager lockManager)
  {
    this.requestResponseManager = requestResponseManager;
    this.invalidateReplicatesHandler = invalidateReplicatesHandler;
    this.localFragmentHandler = localFragmentHandler;
    this.siteManager = siteManager;
    this.lockManager = lockManager;
    fragmentSender = new FragmentSender(localFragmentHandler, networkLayer, requestResponseManager, this);
  }

// --------------------- Begin Interface InvalidateReplicatesClient ---------------------

  public void replicatesInvalidated(String tableName, int fragmentNr, Object parameters[], DataSource dataSource)
  {
    executeRefragment((FragmentationSuggestion)parameters[0], (int[])parameters[1], dataSource);
  }

// --------------------- End Interface InvalidateReplicatesClient ---------------------


// --------------------- Begin Interface MessageClient ---------------------

  public void responseReceived(AbstractResponseMsg msg, Object parameters[], DataSource dataSource)
  {
    if (msg instanceof LookupResponseMsg) {
      LookupResponseMsg message = (LookupResponseMsg)msg;
      boolean granted = lockManager.localLockRequested(message.getTableName(), message.getFragmentNr(),
                                                       true, siteManager.getLocalSite());
      if (!granted) {
        System.err.println("RefragmentHandler.responseReceived: Lock not granted");
        // TODO: Implement wait/retry
      }
      else {
        if (message.getSites()[0].length > 1) {
          invalidateReplicatesHandler.invalidateReplicates(message.getTableName(), message.getFragmentNr(),
                                                           message.getSites()[0],
                                                           this, new Object[] {parameters[0],
                                                                               message.getFragmentNumbers()});
        }
        else
          executeRefragment((FragmentationSuggestion)parameters[0], message.getFragmentNumbers(), dataSource);
      }
    }
    else if (msg instanceof FragmentSendResponseMsg) {
      FragmentSendResponseMsg message = (FragmentSendResponseMsg)msg;
      // Deleting will also unlock
      try {
        localFragmentHandler.deleteTableFragment(message.getTableName(), message.getFragmentNr(), dataSource);
      } catch (SQLException e) {
        e.printStackTrace();
      }
      synchronized (activeSuggestions) {
        for (Iterator<FragmentationSuggestion> iterator = activeSuggestions.iterator(); iterator.hasNext();) {
          TableFragment activeFragment = iterator.next().getTableFragment();
          if (activeFragment.getTableName().equalsIgnoreCase(message.getTableName()) &&
              activeFragment.getFragmentNr() == message.getFragmentNr()) {
            iterator.remove();
          }
        }
      }
    }
  }

// --------------------- End Interface MessageClient ---------------------

  public FragmentSender getFragmentSender()
  {
    return fragmentSender;
  }

  // Entry point
  public void startRefragmentation(FragmentationSuggestion suggestion)
  {
    TableFragment suggestedFragment = suggestion.getTableFragment();
    synchronized (activeSuggestions) {
      for (FragmentationSuggestion activeSuggestion : activeSuggestions) {
        TableFragment activeFragment = activeSuggestion.getTableFragment();
        if (activeFragment.getTableName().equalsIgnoreCase(suggestedFragment.getTableName()) &&
            activeFragment.getFragmentNr() == suggestedFragment.getFragmentNr()) {
          System.out.println("RefragmentHandler.startRefragmentation: " + suggestedFragment + " already in processing");
          return;
        }
      }
      activeSuggestions.add(suggestion);
    }
    AbstractRequestMsg lookupMsg = new LookupRequestMsg(suggestedFragment.getTableName(),
                                                        suggestedFragment.getFragmentNr());
    requestResponseManager.sendRequest(lookupMsg, suggestedFragment.getTableName(), this,
                                       new Object[] {suggestion});
  }

  // Lookup, lock and invalidate have now been done
  private void executeRefragment(FragmentationSuggestion suggestion, int fragmentNumbers[], DataSource dataSource)
  {
    // 1: Check first that the fragment is still here
    TableFragment suggestedFragment = suggestion.getTableFragment();
    TableFragment actualFragment = localFragmentHandler.getTableFragment(suggestedFragment.getTableName(),
                                                                         suggestedFragment.getFragmentNr());
    if (!suggestedFragment.equals(actualFragment)) {
      System.out.println("RefragmentHandler.executeRefragment: Fragment changed - no refragment");
      return;
    }

    System.out.println(suggestion);

    try {
      TableFragment migrateFragment;
      if (suggestion.getSplit2() != -1) { // Middle fragment
        int newFragmentNr1 = RefragmentHandler.findFragmentNumber(fragmentNumbers, -1);
        RefragmentHandler.splitFragment(dataSource, suggestion.getTableFragment(), newFragmentNr1,
                                        suggestion.getSplit2());
        localFragmentHandler.initializeTableFragment(suggestion.getTableFragment().getTableName(),
                                                     newFragmentNr1, dataSource);
        int newFragmentNr2 = RefragmentHandler.findFragmentNumber(fragmentNumbers, newFragmentNr1);
        RefragmentHandler.splitFragment(dataSource, suggestion.getTableFragment(), newFragmentNr2,
                                        suggestion.getSplit1());
        migrateFragment = localFragmentHandler.initializeTableFragment(suggestion.getTableFragment().getTableName(),
                                                                       newFragmentNr2, dataSource);
        localFragmentHandler.initializeTableFragment(suggestion.getTableFragment().getTableName(),
                                                     suggestion.getTableFragment().getFragmentNr(),
                                                     dataSource);
        lockManager.forceLocalLock(migrateFragment.getTableName(), migrateFragment.getFragmentNr(),
                                   siteManager.getLocalSite());
      }
      else if (suggestion.getSplit1() != -1 && suggestion.migrateLower()) {
        int newFragmentNr = RefragmentHandler.findFragmentNumber(fragmentNumbers, -1);
        RefragmentHandler.splitFragment(dataSource, suggestion.getTableFragment(), newFragmentNr,
                                        suggestion.getSplit1());
        migrateFragment = localFragmentHandler.initializeTableFragment(suggestion.getTableFragment().getTableName(),
                                                                       suggestion.getTableFragment().getFragmentNr(),
                                                                       dataSource);
        localFragmentHandler.initializeTableFragment(suggestion.getTableFragment().getTableName(),
                                                     newFragmentNr, dataSource);
        lockManager.forceLocalLock(migrateFragment.getTableName(), migrateFragment.getFragmentNr(),
                                   siteManager.getLocalSite());
      }
      else if (suggestion.getSplit1() != -1 && !suggestion.migrateLower()) {
        System.out.println("Fragment Nunbers: " + Arrays.toString(fragmentNumbers));
        int newFragmentNr = RefragmentHandler.findFragmentNumber(fragmentNumbers, -1);
        System.out.println("New number: " + newFragmentNr);
        RefragmentHandler.splitFragment(dataSource, suggestion.getTableFragment(), newFragmentNr,
                                        suggestion.getSplit1());
        localFragmentHandler.initializeTableFragment(suggestion.getTableFragment().getTableName(),
                                                     suggestion.getTableFragment().getFragmentNr(), dataSource);
        migrateFragment = localFragmentHandler.initializeTableFragment(suggestion.getTableFragment().getTableName(),
                                                                       newFragmentNr, dataSource);
        lockManager.forceLocalLock(migrateFragment.getTableName(), migrateFragment.getFragmentNr(),
                                   siteManager.getLocalSite());
      }
      else // Whole fragment
        migrateFragment = suggestion.getTableFragment();

      fragmentSender.sendCopy(migrateFragment, suggestion.getNewSite(), dataSource);
//      localFragmentHandler.deleteTableFragment(migrateFragment.getTableName(), migrateFragment.getFragmentNr(),
//                                               dataSource);
    } catch (SQLException e) {
      e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
    }
  }
}
