package dascosa.site.query;

import dascosa.Dascosa;
import dascosa.algebra.AlgebraNode;
import dascosa.algebra.CacheNode;
import dascosa.algebra.CrashNode;
import dascosa.algebra.DisplayNode;
import dascosa.algebra.scan.LogicalTableAccessNode;
import dascosa.algebra.scan.TableAccessNode;
import dascosa.data.*;
import dascosa.message.PublishResultSetUsageMsg;
import dascosa.message.QueryCompletedMsg;
import dascosa.message.QueryStatsMsg;
import dascosa.message.request.*;
import dascosa.message.restart.RestartSiteMsg;
import dascosa.site.*;
import dascosa.site.cache.ResultCache;
import dascosa.site.fault.FaultDetector;
import dascosa.site.fault.RestartCache;
import dascosa.site.query.optimizer.JoinNode;
import dascosa.site.query.optimizer.LinkStatistics;
import dascosa.site.query.optimizer.ResultSetIdentifier;
import dascosa.util.FilteredOutput;
import edu.harvard.syrah.pyxida.nc.lib.Coordinate;
import rice.p2p.commonapi.NodeHandle;
import rice.p2p.commonapi.Message;

import javax.sql.DataSource;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 17.aug.2007
 * Time: 10:50:06
 * To change this template use File | Settings | File Templates.
 */
public final class QueryHandler implements MessageClient
{
  private final Collection<QueryListener> queryListeners = new ArrayList<QueryListener>();
  private final Executor executor;
  private final FaultDetector faultDetector;
  private final List<Packet> waitingOperandPackets = new LinkedList<Packet>();
  private final List<PacketHeader> waitingOperandHeaders = new LinkedList<PacketHeader>();
  private final Map<Long, AlgebraNode> executingNodes = new HashMap<Long, AlgebraNode>();
  private final Map<Long, AlgebraNode> completedNodes = new HashMap<Long, AlgebraNode>();
  private final Map<Long, AlgebraNode> pausedNodes = new HashMap<Long, AlgebraNode>();
  private final NetworkLayer networkLayer;
  private final RestartCache restartCache;
  private final QueryStats queryStats;
  private final RequestResponseManager requestResponseManager;
  private final ResultCache resultCache;
  private final SiteManager siteManager;
  public static long lastQueryStartedNano;
  private final ConcurrentHashMap<Long,Map<ResultSetIdentifier,Set<NodeHandle>>> cacheSites;


  public QueryHandler(SiteManager siteManager, NetworkLayer networkLayer,
                      RequestResponseManager requestResponseManager, RestartCache restartCache, ResultCache resultCache)
  {
    this.siteManager = siteManager;
    this.networkLayer = networkLayer;
    this.requestResponseManager = requestResponseManager;
    this.restartCache = restartCache;
    this.resultCache = resultCache;
    cacheSites = new ConcurrentHashMap<Long, Map<ResultSetIdentifier, Set<NodeHandle>>>();
    faultDetector = new FaultDetector(requestResponseManager, networkLayer, this);
    queryStats = new QueryStats();
    executor = new Executor(siteManager, networkLayer, restartCache, queryStats, this, resultCache);
  }

//--------------------- Begin Interface MessageClient ---------------------

  public void responseReceived(AbstractResponseMsg msg, Object parameters[], DataSource dataSource)
  {
    LookupResponseMsg message = (LookupResponseMsg)msg;
    AlgebraNode topNode = (AlgebraNode)parameters[0];
    TableAccessNode tableNode = (TableAccessNode)parameters[1];
    tableNode.setAttributeNames(message.getAttributeNames());
    synchronized (topNode) {
      if (message.getSites() == null)
        siteManager.fireDebugErrorEvent("Unknown table: " + message.getTableName() + ".\n");
      else if (tableNode instanceof LogicalTableAccessNode) {
        if (Dascosa.isPropertySet(Dascosa.USE_SEMANTIC_CACHE))
          processCacheResponse(message, topNode);
        QueryPlanner.localize(topNode, (LogicalTableAccessNode) tableNode, message.getSites(), message.getFragmentTupleCounts(), message.getCoordinates(), message.getFragmentNumbers(), message.getAttributeNames());
      } else {
        if (Dascosa.isPropertySet(Dascosa.USE_SEMANTIC_CACHE))
          processCacheResponse(message, topNode);
        if (tableNode.getAssignedSite() == null)
          tableNode.setAssignedSite(message.getSites()[0][0]);
        if (Dascosa.isPropertySet(Dascosa.USE_OPTIMIZER)){
          assert message.getSites().length == 1 : "Only LogicalTableAccessNode should be fragmented";
          tableNode.setOptionalSites(message.getSites()[0]);
          tableNode.setCoordinates(message.getCoordinates()[0]);
          tableNode.setExpectedCardinallity(message.getFragmentTupleCounts()[0]);
          tableNode.setAttributeNames(message.getAttributeNames());
        }
        tableNode.setNumberOfTuples(message.getFragmentTupleCounts()[message.getFragmentIndex(message.getFragmentNr())]);
      }
      LinkStatistics.getInstance(siteManager).setCoordinates(message.getSites(), message.getCoordinates());
      LinkStatistics.getInstance(siteManager).setCoordinates(message.getCacheCoordinateSites(), message.getCacheCoordinates());
      if (topNode.allTableNodesAssigned()) {
        try {
          lastQueryStartedNano = System.nanoTime();
          if (Dascosa.isPropertySet(Dascosa.USE_OPTIMIZER))
            QueryPlanner.plan(topNode, siteManager, cacheSites.remove(topNode.getNodeID()));
          else
            QueryPlanner.planOld(topNode);
          Dascosa.out.println(FilteredOutput.PLANNING_TIME, Long.toString(System.nanoTime()-lastQueryStartedNano));
          Dascosa.out.println('\n' + QueryPlanner.toString(topNode));
          networkLayer.sendMessage(topNode, topNode.getAssignedSite());
        } catch (RuntimeException e) {
          siteManager.fireDebugErrorEvent(e.getMessage());
          e.printStackTrace();
          //Dascosa.out.println(FilteredOutput.UNSPECIFIED, e.getStackTrace());
          throw e;
        }
      }
    }
  }

//--------------------- End Interface MessageClient ---------------------

  public void addQueryListener(QueryListener queryListener)
  {
    queryListeners.add(queryListener);
  }

  public void algebraNodeReceived(AlgebraNode algebraNode, DataSource dataSource)
  {
    synchronized (this) {
      executingNodes.put(algebraNode.getNodeID(), algebraNode);
      for (QueryListener queryListener : queryListeners)
        queryListener.algebraNodeReceived(algebraNode);
      for (Iterator<PacketHeader> iterator = waitingOperandHeaders.iterator(); iterator.hasNext();) {
        PacketHeader part = iterator.next();
        if (part.getTargetAlgebraNodeID() == algebraNode.getNodeID()) {
          algebraNode.deliverOperandHeader(part);
          iterator.remove();
        }
      }
      for (Iterator<Packet> iterator = waitingOperandPackets.iterator(); iterator.hasNext();) {
        Packet packet = iterator.next();
        if (packet.getTargetAlgebraNodeID() == algebraNode.getNodeID()) {
          algebraNode.deliverOperandPacket(packet);
          iterator.remove();
        }
      }
    }
    executor.tryToExecute(algebraNode, dataSource);
  }

  public synchronized void crashSite()
  {
    restartCache.clearCache();
    faultDetector.clearWatchList();
    executingNodes.clear();
    pausedNodes.clear();
    completedNodes.clear();
    for (QueryListener queryListener : queryListeners)
      queryListener.algebraNodesCleared();
    waitingOperandPackets.clear();
    waitingOperandHeaders.clear();
  }

  // Entry point
  public void executeQuery(AlgebraNode topNode)
  {
    topNode.setAssignedSite(siteManager.getLocalSite());
    List<ResultSetIdentifier> rsis = Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER) ? topNode.buildResultSetIdentifier(): null;
    for (TableAccessNode tableNode : topNode.getUnassignedTableNodes()) {
      AbstractRequestMsg message = new LookupRequestMsg(tableNode.getTableName(), tableNode.getFragmentNr(), topNode.canonicalForm(), Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER) ? rsis.toArray(new ResultSetIdentifier[rsis.size()]):null);
      requestResponseManager.sendRequest(message, tableNode.getTableName(), this,
          new Object[] {topNode, tableNode});
    }
  }

  public synchronized CrashNode getCrashNode()
  {
    for (AlgebraNode algebraNode : executingNodes.values())
      if (algebraNode instanceof CrashNode)
        return (CrashNode)algebraNode;
    for (AlgebraNode algebraNode : pausedNodes.values())
      if (algebraNode instanceof CrashNode)
        return (CrashNode)algebraNode;
    return null;
  }

  public FaultDetector getFaultDetector()
  {
    return faultDetector;
  }

  public synchronized AlgebraNode getLocalAlgebraNode(long algebraNodeID)
  {
    if (executingNodes.containsKey(algebraNodeID))
      return executingNodes.get(algebraNodeID);
    else if (pausedNodes.containsKey(algebraNodeID))
      return pausedNodes.get(algebraNodeID);
    return completedNodes.get(algebraNodeID);
  }

  public synchronized AlgebraNode getOperandAlgebraNode(long algebraNodeID)
  {
    for (AlgebraNode algebraNode : executingNodes.values())
      for (AlgebraNode operandNode : algebraNode.getOperandNodes())
        if (operandNode.getNodeID() == algebraNodeID)
          return operandNode;
    for (AlgebraNode algebraNode : pausedNodes.values())
      for (AlgebraNode operandNode : algebraNode.getOperandNodes())
        if (operandNode.getNodeID() == algebraNodeID)
          return operandNode;
    for (AlgebraNode algebraNode : completedNodes.values())
      for (AlgebraNode operandNode : algebraNode.getOperandNodes())
        if (operandNode.getNodeID() == algebraNodeID)
          return operandNode;
    return null;
  }

  public RestartCache getPacketCache()
  {
    return restartCache;
  }

  public synchronized void operandHeaderReceived(PacketHeader operandHeader)
  {
    if (executingNodes.containsKey(operandHeader.getTargetAlgebraNodeID()))
      executingNodes.get(operandHeader.getTargetAlgebraNodeID()).deliverOperandHeader(operandHeader);
    else if (pausedNodes.containsKey(operandHeader.getTargetAlgebraNodeID()))
      pausedNodes.get(operandHeader.getTargetAlgebraNodeID()).deliverOperandHeader(operandHeader);
    else
      waitingOperandHeaders.add(operandHeader);
  }

  public synchronized void operandPacketReceived(Packet operandPacket, DataSource dataSource)
  {
    if (executingNodes.containsKey(operandPacket.getTargetAlgebraNodeID())) {
      AlgebraNode algebraNode = executingNodes.get(operandPacket.getTargetAlgebraNodeID());
      algebraNode.deliverOperandPacket(operandPacket);
      executor.tryToExecute(algebraNode, dataSource);
    }
    else if (pausedNodes.containsKey(operandPacket.getTargetAlgebraNodeID()))
      pausedNodes.get(operandPacket.getTargetAlgebraNodeID()).deliverOperandPacket(operandPacket);
    else
      waitingOperandPackets.add(operandPacket);
  }

  public void queryCleanup(QueryCompletedMsg msg)
  {
    AlgebraNode algebraNode;
    synchronized (this) {
      algebraNode = completedNodes.remove(msg.getAlgebraNodeID());
    }
    if (algebraNode != null) {
      //System.err.println("DEBUG: Query cleanup");
      AlgebraNode cacheNode = resultCache.removeCacheNode(algebraNode);
      if (cacheNode != null) {
        algebraNode = cacheNode;
        System.err.println("DEBUG: Query used cache " + algebraNode.canonicalForm());
      }

      for (QueryListener queryListener : queryListeners)
        queryListener.algebraNodeRemoved(algebraNode);
      restartCache.removeFrom(algebraNode);
      networkLayer.sendMessage(queryStats.getQueryStats(algebraNode), msg.getReturnStatsTo());
      for (AlgebraNode operandNode : algebraNode.getOperandNodes())
        networkLayer.sendMessage(new QueryCompletedMsg(operandNode.getNodeID(), msg.getReturnStatsTo()),
            operandNode.getAssignedSite());
      algebraNode.clearResult();
    }
    for (NodeHandle failedSite : faultDetector.clearFailedSites())
      networkLayer.sendMessage(new RestartSiteMsg(), failedSite);
  }

  public void queryStatsReceived(QueryStatsMsg queryStatsMsg)
  {
    for (QueryListener queryListener : queryListeners)
      queryListener.queryStatsReceived(queryStatsMsg);
  }

  public synchronized void receiptReceived(long sourceAlgebraNodeID, DataSource dataSource)
  {
    if (pausedNodes.containsKey(sourceAlgebraNodeID)) {
      AlgebraNode algebraNode = pausedNodes.remove(sourceAlgebraNodeID);
      executingNodes.put(sourceAlgebraNodeID, algebraNode);
      for (QueryListener queryListener : queryListeners)
        queryListener.algebraNodeResumed(algebraNode);
//    System.out.println("QueryHandler.receiptReceived: " + algebraNode + " " + algebraNode.getNodeID());
      executor.tryToExecute(algebraNode, dataSource);
    }
    else {
      System.err.println("QueryHandler.receiptReceived: Unknown reciept: " + sourceAlgebraNodeID);
      System.err.println("Executing: " + executingNodes.keySet() + "   " + executingNodes.values());
      System.err.println("Completed: " + completedNodes.keySet());
    }
  }

  synchronized void nodeCompleted(AlgebraNode algebraNode, Coordinate coordinate)
  {
    AlgebraNode node = executingNodes.remove(algebraNode.getNodeID());

    if (Dascosa.isPropertySet(Dascosa.CACHEINVESTMENT) && node != null && !(algebraNode instanceof DisplayNode) && (node.isPJS()))// && !(node instanceof TableAccessNode)) //logger IKKE tabell aksesser
    {
      ResultSetIdentifier rsid = algebraNode.buildResultSetIdentifier().get(0);
      if (rsid.getTables().size() > 0) {
        rsid.setTupleCount(node.getTuplesCreated());
        Message message = new PublishResultSetUsageMsg(coordinate, rsid, algebraNode.getResultSite(0), node.getPlannedCost(), node instanceof TableAccessNode || node instanceof JoinNode ? (int)node.getSizeEstimate() : -1);
        networkLayer.sendMessage(message, rsid.getTables().first());
      } else {
        Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: NodeCompleted with resultSetIdentifier = null");
      }
    }

    if (node == null)
      System.err.println("QueryHandler.nodeCompleted: Could not find completed node " + algebraNode);
    for (QueryListener queryListener : queryListeners)
      queryListener.algebraNodeCompleted(algebraNode);
    completedNodes.put(algebraNode.getNodeID(), algebraNode);
  }

  synchronized void pauseNode(AlgebraNode algebraNode)
  {
//  System.out.println("QueryHandler.pauseNode: " + algebraNode + " " + algebraNode.getNodeID());
    for (QueryListener queryListener : queryListeners)
      queryListener.algebraNodePaused(algebraNode);
    executingNodes.remove(algebraNode.getNodeID());
    pausedNodes.put(algebraNode.getNodeID(), algebraNode);
  }

  void queryCompleted(AlgebraNode algebraNode, FinalResult result)
  {
    Dascosa.out.println("Query completed: " + System.currentTimeMillis());
    if (Dascosa.isPropertySet(Dascosa.PRINT_STATISTICS))
      System.out.println("Query completed: " + System.currentTimeMillis());
    for (QueryListener queryListener : queryListeners)
      queryListener.queryCompleted(result);
    networkLayer.sendMessage(new QueryCompletedMsg(algebraNode.getNodeID(), algebraNode.getAssignedSite()),
        algebraNode.getAssignedSite());
    siteManager.queryCompleted();
  }

  private void processCacheResponse(LookupResponseMsg message, AlgebraNode topNode) throws Error {
    if (message.getCachesByRSI() != null) {
      Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: ResultSetIdentifiers " + message.getCachesByRSI().toString());
      Map<ResultSetIdentifier, Set<NodeHandle>> cacheSitesForQuery = cacheSites.get(topNode.getNodeID());
      if (cacheSitesForQuery == null) {
        cacheSitesForQuery = message.getCachesByRSI();
        cacheSites.put(topNode.getNodeID(), cacheSitesForQuery);
      }
      else
        for (ResultSetIdentifier rsi : message.getCachesByRSI().keySet()) {
          Set<NodeHandle> sites = cacheSitesForQuery.get(rsi);
          if (sites == null) {
            sites = message.getCachesByRSI().get(rsi);
            cacheSitesForQuery.put(rsi, sites);
          }
          else
            sites.addAll(message.getCachesByRSI().get(rsi)); // does not check for duplicates, but this is not supposed to happen
        }
    }
    if (message.getCachesByRSI() == null && message.getCacheSites() != null)
      throw new Error("CacheSites mismatch");
    QueryPlanner.assignCachedNodes(topNode, message.getCacheSites());
  }
}
