package dascosa.site.query;

import dascosa.Dascosa;
import dascosa.algebra.*;
import dascosa.algebra.constraints.Constraint;
import dascosa.algebra.constraints.SimpleConstraint;
import dascosa.algebra.scan.*;
import dascosa.data.PacketHeader;
import dascosa.data.PartialResult;
import dascosa.site.SiteManager;
import dascosa.site.query.optimizer.*;
import dascosa.util.FilteredOutput;
import edu.harvard.syrah.pyxida.nc.lib.Coordinate;
import rice.p2p.commonapi.NodeHandle;

import java.util.*;
import java.util.Map.Entry;

/**
 * Created by IntelliJ IDEA. User: joh Date: 26.apr.2007 Time: 14:02:13 To
 * change this template use File | Settings | File Templates.
 */
public final class QueryPlanner {

  private static void addAlgebraNodes(AlgebraNode parentNode, List<AlgebraNode> algebraNodes) {
    algebraNodes.add(parentNode);
    for (AlgebraNode algebraNode : parentNode.getOperandNodes())
      addAlgebraNodes(algebraNode, algebraNodes);
  }

  /**
   * Recursively assign algebra nodes for cached (sub)queries to sites where they are cached.
   * If a node is already assigned to a site, it is not reassigned.
   *
   * This is called for each index lookup response with caching information from the responding index site.
   *
   * @param parentNode The algebra node under consideration
   * @param cacheSites List of known caches, canonical form => node handle
   */
  public static void assignCachedNodes(AlgebraNode parentNode, Map<String, Set<NodeHandle>> cacheSites) {
    //System.err.println("DEBUG: assignCachedNodes(" + parentNode.canonicalForm() + ")");
    if (cacheSites == null) {
      //System.err.println("DEBUG: cacheSites==null");
      return;
    }
    //for (String query : cacheSites.keySet()) System.err.println("DEBUG:   " + query);
    for (AlgebraNode algebraNode : parentNode.getOperandNodes())
      assignCachedNodes(algebraNode, cacheSites);
    if (parentNode.getAssignedSite() == null && cacheSites.containsKey(parentNode.canonicalForm())) {
      parentNode.setAssignedSite(cacheSites.get(parentNode.canonicalForm()).iterator().next());
      //System.err.println("DEBUG: Assigned node " + parentNode.canonicalForm() + " to site " + parentNode.getAssignedSite() + " where it's cached.");
    } else if (parentNode.getAssignedSite() != null && cacheSites.containsKey(parentNode.canonicalForm()))
      ;//System.err.println("DEBUG: Node " + parentNode.canonicalForm() + " already assigned to site " + parentNode.getAssignedSite());
  }

  public static void assignNodes(AlgebraNode parentNode) {
    for (AlgebraNode algebraNode : parentNode.getOperandNodes())
      assignNodes(algebraNode);
    if (parentNode.getAssignedSite() == null) {
      int child = 0;//(int)(parentNode.getOperandNodes().size() * Math.random());
      parentNode.setAssignedSite(parentNode.getOperandNodes().get(child).getAssignedSite());
      if (parentNode.getOperandNodes().get(child).getAssignedSite() == null)
        System.err.println("DEBUG: WARNING: Assigning node to site 'null'");
    }
  }

  static void localize(AlgebraNode topNode, LogicalTableAccessNode logicalNode, NodeHandle[][] sites,
          int[] fragmentCardinallity, Coordinate[][] coordinates, int[] fragmentNumbers, String[] attributes) {

    AlgebraNode targetNode = logicalNode.getResultNode(0);
    int index = targetNode.getOperandNodes().indexOf(logicalNode);
    for (int i = 0; i < logicalNode.getResultNodeCount(); i++)
      logicalNode.getResultNode(i).removeOperandNode(logicalNode);



    for (int fragmentIndex = 0; fragmentIndex < fragmentNumbers.length; fragmentIndex++) {
      int fragmentNr = fragmentNumbers[fragmentIndex];
      SimpleTableAccessNode node = new SimpleTableAccessNode(index++, targetNode, logicalNode, fragmentNr);
      node.setAssignedSite(sites[fragmentIndex][0]); // TODO: Consider removing this line...
      node.setOptionalSites(sites[fragmentIndex]);
      node.setCoordinates(coordinates[fragmentIndex]);
      node.setExpectedCardinallity(fragmentCardinallity[fragmentIndex]);
      node.setAttributeNames(attributes);
    }
  }

  private static List<HashMap<Set<String>, List<Option>[]>> buildPlans(ResultSetIdentifier topRsid,
          Map<String, SortedMap<Integer, ArrayList<Option>>> fragmentsForTables,
          Map<ResultSetIdentifier, Set<NodeHandle>> cacheSites,
          Set<SimpleConstraint> queryConstraints, Map<String, HashSet<String>> connected,
          HashMap<String, HashMap<String, String[]>> joinAttributes,
          SiteManager siteManager,
          NodeHandle targetSite) {

    //Populerer plans med tomme hashmaps
    List<HashMap<Set<String>, List<Option>[]>> plans = new ArrayList<HashMap<Set<String>, List<Option>[]>>();
    for (int i = 0; i < fragmentsForTables.size(); i++)
      plans.add(i, new HashMap<Set<String>, List<Option>[]>());
   
    //Første runde i DP algoritmen håndteres spesielt på grunn av fragmenter
    for (String key : fragmentsForTables.keySet()) {
      Set<String> relationSet = new HashSet<String>();
      relationSet.add(key);
      Collection<ArrayList<Option>> fragmentReplicateMatrix = new ArrayList<ArrayList<Option>>();
      for (ArrayList<Option> list: fragmentsForTables.get(key).values())
        fragmentReplicateMatrix.add((ArrayList<Option>) list.clone());
      plans.get(0).put(relationSet, fragmentReplicateMatrix.toArray(new ArrayList[fragmentReplicateMatrix.size()]));
    }

    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Første runde planning done");

    //Setter inn opplysninger om relevant cache som delplaner i plans
    Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: RSIDs for CacheOptions: " + cacheSites.keySet());
    for (ResultSetIdentifier rsi : cacheSites.keySet())
      if (topRsid.contains(rsi))
        for (NodeHandle site : cacheSites.get(rsi)) {
          CacheOption cacheOption = new CacheOption(site, 0, rsi, queryConstraints);
          List<Option>[] optionListArray = plans.get(rsi.getTables().size() - 1).get(rsi.getTables());
          if (optionListArray == null) {
            optionListArray = new List[1];
            optionListArray[0] = new ArrayList<Option>();
            plans.get(rsi.getTables().size() - 1).put(rsi.getTables(), optionListArray);
          }
          optionListArray[0].add(cacheOption);
        }
      
    //Resten av DP algorimten (delplaner større enn 1)
    for (int currentCardinality = 2; currentCardinality <= fragmentsForTables.size(); currentCardinality++) {
      Dascosa.out.println(FilteredOutput.OPTIMIZER, "Planning runde: " + currentCardinality);
      for (int smallOperand = 1; smallOperand < currentCardinality; smallOperand++) {
        int largeOperand = currentCardinality - smallOperand;
        Set<Set<Set<String>>> evaluated = new HashSet<Set<Set<String>>>();
        for (Set<String> large : plans.get(largeOperand - 1).keySet())
          for (Set<String> small : plans.get(smallOperand - 1).keySet()) {
            Set<Set<String>> operandSet = new HashSet<Set<String>>();
            operandSet.add(large);
            operandSet.add(small);
            if (Collections.disjoint(large, small) && connected(large, small, connected) && !evaluated.contains(operandSet))
              evaluate(large, small, plans, joinAttributes, siteManager, targetSite, queryConstraints);
            else
              Dascosa.out.println(FilteredOutput.OPTIMIZER|FilteredOutput.DEBUG, "Not evaluated: " + large + ", " + small);
          }
      }
    }

    return plans;
  }

  private static void makeLocalDBConstraint(ConstraintNode constraintNode) {
    try {
      String sqlConstraint = constraintNode.getSQLConstraint();

      AlgebraNode resultNode = constraintNode.getResultNode(0);
      resultNode.removeOperandNode(constraintNode);

      for (AlgebraNode operandNode : constraintNode.getOperandNodes()) {
        AlgebraNode newOperandNode = new SelProjTableAccessNode(
                resultNode, (TableAccessNode) operandNode,
                sqlConstraint);
        newOperandNode.setOperationID(constraintNode.getOperationID());
        newOperandNode.setAssignedSite(operandNode.getAssignedSite());
      }
    } catch (RuntimeException e) {
      System.out.println("QueryPlanner.makeLocalDBConstraint: " + e.getMessage());
    }
  }

  private static void evaluate(Set<String> large, Set<String> small,
          List<HashMap<Set<String>, List<Option>[]>> plans,
          HashMap<String, HashMap<String, String[]>> joinAttributes,
          SiteManager siteManager, NodeHandle targetSite,
          Set<SimpleConstraint> queryConstraints) {
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Evaluate: " + large + ", " + small);

    HashSet<String> resultSet = new HashSet<String>();
    resultSet.addAll(large);
    resultSet.addAll(small);

    List<Option>[] resultPlans = plans.get(resultSet.size() - 1).get(resultSet);

    //Dascosa.out.println(Dascosa.FilteredOutput.optimizer, resultPlans);
    //System.out.print(resultPlans);

    List<Option>[] smallPlans = plans.get(small.size() - 1).get(small);
    List<Option>[] largePlans = plans.get(large.size() - 1).get(large);

    Collection<NodeHandle> destinations = new HashSet<NodeHandle>();
    for (List<Option> smallPlan : smallPlans)
      for (Option n : smallPlan)
        destinations.add(n.site);
    for (List<Option> largePlan : largePlans)
      for (Option n : largePlan)
        destinations.add(n.site);
    destinations.add(targetSite);
    Set<CompositeOption> destinationOptions = new HashSet<CompositeOption>();
    for (NodeHandle n : destinations) {
      String[] attributePair = findJoinAttributes(small, large, joinAttributes);
      destinationOptions.add(new CompositeOption(n, resultSet, attributePair[0], attributePair[1]));
    }
    if (resultPlans == null) {
      resultPlans = new List[1];
      resultPlans[0] = new LinkedList<Option>();
    }
    resultPlans[0].addAll(CostModel.getTotalCost(largePlans, smallPlans,
            destinationOptions, LinkStatistics.getInstance(siteManager)));
    for (CompositeOption co : destinationOptions)
      co.addConstraints(queryConstraints);

    plans.get(resultSet.size() - 1).put(resultSet, resultPlans);
  }

  private static void processAlgebraTree(AlgebraNode topNode, HashMap<String, SortedMap<Integer, ArrayList<Option>>> fragmentsForTables, HashMap<String, HashMap<String, String[]>> joinAttributes, Set<SimpleConstraint> queryConstraints, Set<NodeHandle> sites, HashMap<String, HashSet<String>> connected) {
    List<AlgebraNode>algebraNodes = new LinkedList<AlgebraNode>();
    QueryPlanner.addAlgebraNodes(topNode, algebraNodes);
    for (AlgebraNode algebraNode : algebraNodes)
      if (algebraNode.isPJS())
        if (algebraNode instanceof TableAccessNode) {
          TableAccessNode tableNode = (TableAccessNode) algebraNode;
          SortedMap<Integer, ArrayList<Option>> fragmentMap = fragmentsForTables.get(tableNode.getTableName());
          long operationID = tableNode.getOperationID();
          if (fragmentMap == null) {
            fragmentMap = new TreeMap<Integer, ArrayList<Option>>();
            fragmentsForTables.put(tableNode.getTableName(), fragmentMap);
          }
          ArrayList<Option> replics = fragmentMap.get(tableNode.getFragmentNr());
          if (replics == null) {
            replics = new ArrayList<Option>();
            fragmentMap.put(tableNode.getFragmentNr(), replics);
          }
          if (tableNode instanceof SelProjTableAccessNode)
            throw new Error("Unable to parse sqlConstraint into SimpleConstraint. SelProjTableAccessNode not supported as input to optimizer");

          for (NodeHandle site : tableNode.getOptionalSites()) {
            HashSet<String> tableSet = new HashSet<String>();
            tableSet.add(tableNode.getTableName());
            Set<String> relatedAttributes = new HashSet<String>();
            Map<String, String[]> thisTablesJoinAttributes = joinAttributes.get(tableNode.getTableName());
            if (thisTablesJoinAttributes != null)
              for (String[] attributes : thisTablesJoinAttributes.values())
                if (attributes != null)
                  relatedAttributes.add(attributes[0]);
            BaseAccessOption option = new BaseAccessOption(site, tableSet, tableNode.getExpectedCardinality(), tableNode.getFragmentNr(), tableNode.getOperationID(), tableNode.getAttributeNames(), relatedAttributes, queryConstraints);
            replics.add(option);
          }

          sites.addAll(Arrays.asList(tableNode.getOptionalSites()));
          Dascosa.out.println(FilteredOutput.OPTIMIZER, tableNode.getTableName() + '_' + tableNode.getFragmentNr() + " has " + tableNode.getExpectedCardinality() + " tuples");
        } else if (algebraNode instanceof JoinNode) {

          JoinNode joinNode = (JoinNode) algebraNode;

          String tableA = joinNode.getInnerRelationName()[0];
          String attributeA = joinNode.getInnerRelationName()[1];
          String tableB = joinNode.getOuterRelationName()[0];
          String attributeB = joinNode.getOuterRelationName()[1];

          HashSet<String> a = connected.get(tableA);
          if (a == null)
            a = new HashSet<String>();
          connected.put(tableA, a);
          a.add(tableB);

          HashSet<String> b = connected.get(tableB);
          if (b == null)
            b = new HashSet<String>();
          connected.put(tableB, b);
          b.add(tableA);

          HashMap<String, String[]> mapA = joinAttributes.get(tableA);
          if (mapA == null) {
            mapA = new HashMap<String, String[]>();
            joinAttributes.put(tableA, mapA);
          }
          String[] arrayA = {attributeA, attributeB};
          mapA.put(tableB, arrayA);

          HashMap<String, String[]> mapB = joinAttributes.get(tableB);
          if (mapB == null) {
            mapB = new HashMap<String, String[]>();
            joinAttributes.put(tableB, mapB);
          }
          String[] arrayB = {attributeB, attributeA};
          mapB.put(tableA, arrayB);

          // Dascosa.out.println(Dascosa.FilteredOutput.optimizer,
          // operands[0] + "§" + operands[1]);
        } else if (algebraNode instanceof ConstraintNode) {
          boolean onlySimpleTableAccess = true;

          ConstraintNode constraintNode = (ConstraintNode) algebraNode;

          queryConstraints.addAll(constraintNode.getConjunctiveSimpleConstraintsSet());
        }

  //Ferdig å parse relalg-treet for info
  }

  private static void retrieveHistograms(Map<String, HashMap<String, String[]>> joinAttributes, HashMap<String, SortedMap<Integer, ArrayList<Option>>> fragmentsForTables, DataStatistics dataStatisticsInstance) {
    for (String table : joinAttributes.keySet()) {
      SortedMap<Integer, ArrayList<Option>> fragmentMap = fragmentsForTables.get(table);
      for (int fragmentnr : fragmentMap.keySet())
        for (String[] attribute : joinAttributes.get(table).values())
          dataStatisticsInstance.fetchHistogram(table, fragmentnr, attribute[0]);
    }
  }

  private static List<Option> selectFinalPlan(List<HashMap<Set<String>, List<Option>[]>> plans, SiteManager siteManager, NodeHandle targetSite) {

    double minNoCacheCost = Double.MAX_VALUE;
    double minCost = Double.MAX_VALUE;
    Set<String> tableSet = plans.get(plans.size() - 1).keySet().iterator().next();
    final List<Option>[] fragments = plans.get(plans.size() - 1).get(tableSet);
    //Option[] cheapestOption = new Option[fragments.length];
    //Option[] cheapestNoCacheOption = new Option[fragments.length];
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Selecting final plan");
    assert plans.get(plans.size() - 1).keySet().size() == 1: "There can only be one set of the largest size";

    List<Option> plan = CostModel.selectPlanForSite(fragments, LinkStatistics.getInstance(siteManager), targetSite);

    /*
    for (int fragmentNumber = 0; fragmentNumber < fragments.length; fragmentNumber++) {
      //for (List<Option> replicatesForFragment : plans.get(plans.size() - 1).get(tableSet)){
      List<Option> replicatesForFragment = fragments[fragmentNumber];
      minNoCacheCost = Double.MAX_VALUE;
      minCost = Double.MAX_VALUE;
      for (Option replicate : replicatesForFragment) {
        double currentCost = replicate.getPriorCost() + (targetSite != null ? LinkStatistics.getInstance(siteManager).getLinkCost(replicate.site, targetSite) : 0) * replicate.getTupleCount();
        if (currentCost < minCost) {
          minCost = currentCost;
          cheapestOption[fragmentNumber] = replicate;
        }
        if (currentCost < minNoCacheCost && !(replicate instanceof CacheOption)) {
          minNoCacheCost = currentCost;
          cheapestNoCacheOption[fragmentNumber] = replicate;
        }

        Dascosa.out.println(FilteredOutput.OPTIMIZER, "Alternativ:\n" + replicate.toString());
      }
    }
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Plan selection done");
    if (cheapestOption[0] instanceof CacheOption) {
      List<Option> alternativePlan = new ArrayList<Option>(fragments.length);
      alternativePlan.add(cheapestNoCacheOption);
      ((CacheOption) cheapestOption).addAlternativePlan(alternativePlan);
    }

    if (cheapestOption.containsCacheOption())
      Dascosa.out.println(FilteredOutput.OPTIMIZER, "CACHE: Cache option won");
    */
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Vinner:\n" + plan.toString());
    Dascosa.out.println(FilteredOutput.OPTIMIZER_TESTLOG, "Estimate: " + minCost);
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Optimize done");

    return plan;
  }

  public static void buildAlternatePlan(CacheOption cacheOption,
                Map<String, SortedMap<Integer, ArrayList<Option>>> fragmentsForTables,
                Map<String, HashSet<String>> connected,
                HashMap<String, HashMap<String, String[]>> joinAttributes,
                SiteManager siteManager) {

    ResultSetIdentifier rsi = cacheOption.getResultSetIdentifier();
    Map<ResultSetIdentifier, Set<NodeHandle>> cacheSites = new HashMap<ResultSetIdentifier, Set<NodeHandle>>();
    Set<SimpleConstraint> queryConstraints = cacheOption.getConstraints();

    Map<String, SortedMap<Integer, ArrayList<Option>>> filteredTables = new HashMap<String, SortedMap<Integer, ArrayList<Option>>>();
    Map<String, HashSet<String>> filteredConnected = new HashMap<String, HashSet<String>>();
    HashMap<String, HashMap<String, String[]>> filteredJoinAttributes = new HashMap<String, HashMap<String, String[]>>();
    for (String table : rsi.getTables()) {
      filteredTables.put(table, fragmentsForTables.get(table));
      filteredConnected.put(table, connected.get(table));
      filteredJoinAttributes.put(table, joinAttributes.get(table));
    }

    List<HashMap<Set<String>, List<Option>[]>> alternatePlan = buildPlans(rsi, filteredTables, cacheSites, rsi.getConstraints(), filteredConnected, filteredJoinAttributes, siteManager, cacheOption.site);

    cacheOption.addAlternativePlan(selectFinalPlan(alternatePlan, siteManager, cacheOption.site));

  }

  /**
   * Executes the query without actually using tuples, just doing operations on table columns.
   * For now, the only point of this is to find join attribute for natural join which is
   * needed to make the join distributed. Also, error checks are made.
   */
  private static void simulate(AlgebraNode topNode) {
    List<AlgebraNode> algebraNodes = new ArrayList<AlgebraNode>();
    addAlgebraNodes(topNode, algebraNodes);
    while (!algebraNodes.isEmpty())
      for (Iterator<AlgebraNode> iterator = algebraNodes.iterator(); iterator.hasNext();) {
        AlgebraNode algebraNode = iterator.next();
        PartialResult simulation = algebraNode.simulate();
        if (simulation != null) { // Not stream execution during simulation
          iterator.remove();
          if (!(algebraNode instanceof DisplayNode))
            for (int i = 0; i < simulation.getPacketHeaderCount(); i++) {
              PacketHeader resultPart = simulation.getPacketHeader(i);
              for (AlgebraNode node : algebraNodes)
                if (resultPart.getTargetAlgebraNodeID() == node.getNodeID()) {
                  node.deliverOperandHeader(resultPart);
                  for (int j = 0; j < resultPart.getPacketCount(); j++)
                    node.deliverOperandPacket(resultPart.getPacket(j));
                }
            }
        }
      }
    System.out.println("Query simulated");
  }

  private static String[] findJoinAttributes(Set<String> a, Set<String> b,
          HashMap<String, HashMap<String, String[]>> joinAttributes) {
    for (String tableA : a) {
      HashMap<String, String[]> mapA = joinAttributes.get(tableA);
      for (String tableB : b) {
        String[] attributes = mapA.get(tableB);
        if (attributes != null)
          return attributes;
      }
    }

    throw new RuntimeException("Join attributes not found");
  }

  private static void optimizeQueryOld(AlgebraNode topNode) {
    List<AlgebraNode> algebraNodes = new ArrayList<AlgebraNode>();
    QueryPlanner.addAlgebraNodes(topNode, algebraNodes);

//    for (AlgebraNode algebraNode: algebraNodes)
//      if (algebraNode instanceof NaturalJoinNode) {
//        distributeJoin((NaturalJoinNode)algebraNode);
//        break;
//      }

//    for (AlgebraNode algebraNode: algebraNodes)
//      if (algebraNode instanceof AggregateNode)
//        ((AggregateNode)algebraNode).distribute();

    for (AlgebraNode algebraNode : algebraNodes)
      if (algebraNode instanceof ConstraintNode && algebraNode.getNumberOfOperands() == 1) {
        boolean onlySimpleTableAccess = true;
        for (AlgebraNode operandNode : algebraNode.getOperandNodes())
          if (!(operandNode instanceof SimpleTableAccessNode))
            onlySimpleTableAccess = false;
        if (onlySimpleTableAccess)
          makeLocalDBConstraint((ConstraintNode) algebraNode);
      }
  }

  private static List<Option> optimizeQuery(AlgebraNode topNode, SiteManager siteManager, Map<ResultSetIdentifier, Set<NodeHandle>> cacheSites) {

    Set<SimpleConstraint> queryConstraints = new HashSet<SimpleConstraint>();

    HashMap<String, SortedMap<Integer, ArrayList<Option>>> fragmentsForTables = new HashMap<String, SortedMap<Integer, ArrayList<Option>>>();
    HashMap<String, HashSet<String>> connected = new HashMap<String, HashSet<String>>();

    Set<NodeHandle> sites = new HashSet<NodeHandle>();
    sites.add(siteManager.getLocalSite());

    HashMap<String, HashMap<String, String[]>> joinAttributes = new HashMap<String, HashMap<String, String[]>>();
    DataStatistics dataStatisticsInstance = DataStatistics.getInstance();

    processAlgebraTree(topNode, fragmentsForTables, joinAttributes, queryConstraints, sites, connected);

    if (Dascosa.isPropertySet(Dascosa.USE_HISTOGRAMS))
      retrieveHistograms(joinAttributes, fragmentsForTables, dataStatisticsInstance);

    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Tabeller: " + fragmentsForTables.toString());
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Joins: " + connected);
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Sites: " + sites.toString());
    
    List<HashMap<Set<String>, List<Option>[]>> plans = buildPlans(topNode.buildResultSetIdentifier().get(0), fragmentsForTables, cacheSites, queryConstraints, connected, joinAttributes, siteManager, siteManager.getLocalSite());

    List<Option> finalPlan = selectFinalPlan(plans, siteManager, topNode.getAssignedSite());

    addAlternatePlansForCache(finalPlan.iterator(), fragmentsForTables, connected, joinAttributes, siteManager);

    return finalPlan;
  }

  private static void insertPlan(Option plan, AlgebraNode parent, int index, LinkStatistics linkStatistics) {
    AlgebraNode result;

    if (!plan.getConstraints().isEmpty() && !(plan instanceof BaseAccessOption)) {

      Constraint[] andBlock = new Constraint[plan.getConstraints().size()];
      plan.getConstraints().toArray(andBlock);
      parent = new ConstraintNode(index, parent, andBlock);
      parent.setAssignedSite(plan.site);
      index = 0;
    }

    if (plan instanceof CompositeOption) {
      CompositeOption composite = (CompositeOption) plan;
      result = new NaturalJoinNode(index, parent);
      result.setAssignedSite(plan.site);
      Iterator<Option> subOptions = composite.getSubOptions();
      int i = 0;
      while (subOptions.hasNext())
        insertPlan(subOptions.next(), result, i++, linkStatistics);
    } else if (plan instanceof BaseAccessOption) {
      BaseAccessOption fragment = (BaseAccessOption) plan;
      if (!plan.getConstraints().isEmpty()) {
        //AlgebraNode resultNode, String tableName, int fragmentNr, String projectNames[], String sqlConstraint)
        Constraint[][] constraints = new Constraint[1][plan.getConstraints().size()];
        plan.getConstraints().toArray(constraints[0]);
        String sqlConstraint = ConstraintNode.disjunctiveSQLBuilder(constraints);
        result = new SelProjTableAccessNode(index, parent, fragment.tables.iterator().next(), fragment.getFragmentNr(), new String[0], sqlConstraint, plan.getConstraints());
      } else
        result = new SimpleTableAccessNode(index, parent, fragment.tables.iterator().next(), fragment.getFragmentNr());
      result.setAssignedSite(plan.site);
      result.setOperationID(fragment.operationID);
    } else if (plan instanceof CacheOption) {
      CacheOption cacheOption = (CacheOption) plan;
      ResultSetIdentifier rsid = cacheOption.getResultSetIdentifier();
      result = new CacheNode(index, parent, rsid);
      result.setAssignedSite(plan.site);
      int i = 0;
      for (Option o : cacheOption.getAlternativePlan())
        insertPlan(o, result, i++, linkStatistics);
      if (rsid.getCanonicalForm() == null)
        rsid.setCanonicalForm(result.getOperandNodes().get(0).canonicalForm());
    } else
      //Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: Option " + plan.toString());
      throw new IllegalArgumentException("Option sub-class not supported: " + plan.getClass().toString());
    
    double cost = (parent.getAssignedSite() != null? linkStatistics.getLinkCost(plan.site, parent.getAssignedSite()):0) * plan.getTupleCount() + plan.getPriorCost();//TODO: Set this cost when result site is guaranteed to be known
    Dascosa.out.println(FilteredOutput.DEBUG, "DEBUG: Cost " + result.toString() + " = " + cost);
    result.setPlannedCost(cost);
  }

  private static boolean connected(Set<String> large, Set<String> small,
          Map<String, HashSet<String>> connected) {
    for (String a : large)
      for (String b : small) {
        if (!connected.containsKey(a) || !connected.containsKey(b))
          return false;
        if (connected.get(a).contains(b))
          return true;
      }
    return false;
  }

  static void plan(AlgebraNode topNode, SiteManager siteManager, Map<ResultSetIdentifier, Set<NodeHandle>> cacheSites) {
    Dascosa.out.counter++;
    Dascosa.out.println(FilteredOutput.OPTIMIZER|FilteredOutput.DEBUG, "++Optimizer input: ++");
    Dascosa.out.println(FilteredOutput.OPTIMIZER|FilteredOutput.DEBUG, toString(topNode));
    simulate(topNode);

    if (cacheSites == null || !Dascosa.isPropertySet(Dascosa.USE_CACHE_AWARE_OPTIMIZER))
      cacheSites = new HashMap<ResultSetIdentifier, Set<NodeHandle>>();

    planPJS(topNode, siteManager, cacheSites);
    assignNodes(topNode);
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "++Optimizer output: ++");
    Dascosa.out.println(FilteredOutput.OPTIMIZER, toString(topNode));
    Dascosa.out.println(FilteredOutput.OPTIMIZER_TESTLOG, toString(topNode));
    Dascosa.out.println(FilteredOutput.OPTIMIZER, "Query planned");
  }

  private static void planPJS(AlgebraNode currentNode, SiteManager siteManager, Map<ResultSetIdentifier, Set<NodeHandle>> cacheSites)
  {
    LinkStatistics linkStatistics = LinkStatistics.getInstance(siteManager);
    if (currentNode.isPJS()){
      List<Option> plans = optimizeQuery(currentNode, siteManager, cacheSites);
      AlgebraNode resultNode = currentNode.getResultNode(0);
      int operandIndex = resultNode.getOperandNodes().indexOf(currentNode);
      resultNode.removeOperandNode(currentNode);
      for (Option plan : plans)
        insertPlan(plan, resultNode, operandIndex++,  linkStatistics);
    }else{
      Dascosa.out.println(FilteredOutput.DEBUG, "PJS: processing children of "+currentNode.toString());
      HashMap<Long, List<AlgebraNode>> operationMap = new HashMap<Long, List<AlgebraNode>>();
      for (AlgebraNode child: currentNode.getOperandNodes()){
        if (child.isPJS()){
          List<AlgebraNode> fragments = operationMap.get(child.getOperationID());
          if (fragments == null){
            fragments = new LinkedList<AlgebraNode>();
            operationMap.put(child.getOperationID(), fragments);
          }
          fragments.add(child);
        }
        else
          planPJS(child, siteManager, cacheSites);

      }


      for (List<AlgebraNode> fragments : operationMap.values()){
        List<Option> plans;
        if (fragments.size()>1)
          plans = optimizeTableScan(fragments, siteManager, cacheSites, currentNode.getAssignedSite());
        else
          plans = optimizeQuery(fragments.get(0), siteManager, cacheSites);

        int index = currentNode.getOperandNodes().indexOf(fragments.get(0));
        for (AlgebraNode fragment: fragments){
          currentNode.removeOperandNode(fragment);
        }

        //if (plans.size() != indexes.length)
        //  Dascosa.out.println(FilteredOutput.OPTIMIZER|FilteredOutput.DEBUG, "plans.size(): "+plans.size()+ " indexes.length: "+indexes.length);
        for (Option plan : plans)
          insertPlan(plan, currentNode, index++,  linkStatistics);
      }
    }
  }



  private static List<Option> optimizeTableScan(List<AlgebraNode> fragments, SiteManager siteManager, Map<ResultSetIdentifier, Set<NodeHandle>> cacheSites, NodeHandle targetSite)
  {
    ResultSetIdentifier rsi = new ResultSetIdentifier();
    rsi.addTable(((TableAccessNode)fragments.get(0)).getTableName());

    ArrayList<Option>[] plans = new ArrayList[fragments.size()];
    int i = 0;
    for (AlgebraNode fragment: fragments){
      TableAccessNode tableNode = (TableAccessNode)fragment;
      plans[i] = new ArrayList<Option>();

      for (NodeHandle site: tableNode.getOptionalSites())
        plans[i].add(new BaseAccessOption(site, rsi.getTables(), fragment.getExpectedCardinality(), tableNode.getFragmentNr(), fragment.getOperationID(), tableNode.getAttributeNames(), new HashSet<String>(), new HashSet<SimpleConstraint>()));
      i++;
    }
    ArrayList<Option> noCache = (ArrayList<Option>) plans[0].clone();
    if (cacheSites.containsKey(rsi))
      for (NodeHandle site : cacheSites.get(rsi))
        plans[0].add(new CacheOption(site, 0, rsi, new HashSet<SimpleConstraint>()));


    List<Option> plan = CostModel.selectPlanForSite(plans, LinkStatistics.getInstance(siteManager), targetSite);
    if (plan.get(0) instanceof CacheOption){
      CacheOption cacheOption = (CacheOption)plan.get(0);
      plans[0] = noCache;
      cacheOption.addAlternativePlan(CostModel.selectPlanForSite(plans, LinkStatistics.getInstance(siteManager), cacheOption.site));
    }
    return plan;
  }

  private static void addAlternatePlansForCache(Iterator<Option> plan,
                Map<String, SortedMap<Integer, ArrayList<Option>>> fragmentsForTables,
                Map<String, HashSet<String>> connected,
                HashMap<String, HashMap<String, String[]>> joinAttributes,
                SiteManager siteManager) {

    while (plan.hasNext()) {
      Option option = plan.next();
      if (option instanceof CacheOption) {
        CacheOption cacheOption = (CacheOption)option;
        buildAlternatePlan(cacheOption, fragmentsForTables, connected, joinAttributes, siteManager);
      } else if (option instanceof CompositeOption) {
        CompositeOption compositeOption = (CompositeOption)option;
          addAlternatePlansForCache(compositeOption.getSubOptions(), fragmentsForTables, connected, joinAttributes, siteManager);
      }
    }
  }

  static void planOld(AlgebraNode topNode) {
    Dascosa.out.counter++;
    simulate(topNode);

    optimizeQueryOld(topNode);

    assignNodes(topNode);

    System.out.println("Query planned");
  }

  private static void treeToString(AlgebraNode parentNode, StringBuilder sb, String indent) {
    sb.append(indent).append(parentNode).append(" at site: ").append(parentNode.getAssignedSite()).append('\n');
    indent += "  ";
    for (AlgebraNode algebraNode : parentNode.getOperandNodes())
      treeToString(algebraNode, sb, indent);
  }

  static String toString(AlgebraNode topNode) {
    StringBuilder sb = new StringBuilder();
    treeToString(topNode, sb, "");
    return sb.toString();
  }

	private QueryPlanner()
  {
	}

	// private void distributeAggregation(AggregateNode aggregateNode)
	// {
	// if (aggregateNode.hasPartialInput() || aggregateNode.hasPartialOutput())
	// return;
	// AlgebraNode targetNode = aggregateNode.getResultNode(0);
	//
	// // 1: Remove old site
	// targetNode.removeOperandNode(aggregateNode);
	//
	// // 2: Make new topnode
	// AlgebraNode topAggregationNode =
	// AggregateNode.createSimilar(aggregateNode, targetNode, true, false);
	// topAggregationNode.setOperationID(aggregateNode.getOperationID());
	//
	// // 3: Make partial aggreation nodes
	// AlgebraNode partialNodes[] = new
	// AlgebraNode[aggregateNode.getOperandNodes().size()];
	// for (int i = 0; i < partialNodes.length; i++) {
	// partialNodes[i] = AggregateNode.createSimilar(aggregateNode,
	// topAggregationNode, false, true);
	// partialNodes[i].setAssignedSite(aggregateNode.getOperandNodes().get(i).getAssignedSite());
	// // Assign to same site
	// partialNodes[i].setOperationID(aggregateNode.getOperationID());
	// }
	//
	// // 4: Update source nodes
	// for (int i = 0; i < aggregateNode.getOperandNodes().size(); i++)
	// aggregateNode.getOperandNodes().get(i).setResultNode(0, partialNodes[i]);
	//
	// Dascosa.out.println("\nDistributed aggregation:");
	// Dascosa.out.println(this);
	// }

	// private void distributeJoin(NaturalJoinNode joinNode)
	// {
	// AlgebraNode targetNode = joinNode.getResultNode(0);
	//
	// // 1: Remove old site
	// targetNode.removeOperandNode(joinNode);
	//
	// // 2: Find source nodes and number of sources per data
	// Map<Long, Integer> partsPerOperand = new HashMap<Long, Integer>();
	// for (AlgebraNode algebraNode: joinNode.getOperandNodes()) {
	// Integer operandParts = partsPerOperand.get(algebraNode.getOperationID());
	// if (operandParts == null)
	// partsPerOperand.put(algebraNode.getOperationID(), 1);
	// else
	// partsPerOperand.put(algebraNode.getOperationID(), operandParts + 1);
	// }
	//
	// // 3: Make new join nodes
	// AlgebraNode joinNodes[] = new
	// AlgebraNode[joinNode.getOperandNodes().size()];
	// for (int i = 0; i < joinNodes.length; i++) {
	// joinNodes[i] = new NaturalJoinNode(targetNode);
	// joinNodes[i].setOperationID(joinNode.getNodeID());
	// }
	//
	// // 4: Make redistri-nodes
	// AlgebraNode redistriNodes[] = new AlgebraNode[joinNodes.length];
	// for (int i = 0; i < redistriNodes.length; i++) {
	// redistriNodes[i] = new HashRedistriNode(joinNodes,
	// joinNode.getJoinAttribute());
	// redistriNodes[i].setOperationID(joinNode.getOperandNodes().get(i).getOperationID());
	// redistriNodes[i].getOperandNodes().add(joinNode.getOperandNodes().get(i));
	// }
	//
	// // 5: Update source nodes
	// for (int i = 0; i < joinNodes.length; i++)
	// joinNode.getOperandNodes().get(i).setResultNode(0, redistriNodes[i]);
	//
	// // 6: Assign sites
	// for (int i = 0; i < joinNodes.length; i++) {
	// joinNodes[i].setAssignedSite(joinNode.getOperandNodes().get(i).getAssignedSite());
	// redistriNodes[i].setAssignedSite(joinNodes[i].getAssignedSite());
	// }
	//
	// Dascosa.out.println("\nDistributed join:");
	// Dascosa.out.println(this);
	// }
}
