package dascosa.site.query.optimizer;

import java.util.*;

import dascosa.Dascosa;
import dascosa.algebra.constraints.AttributeConstraint;
import dascosa.algebra.constraints.SimpleConstraint;
import dascosa.util.FilteredOutput;

import rice.p2p.commonapi.NodeHandle;

public final class CompositeOption extends Option {

	private final List<Option> inner = new ArrayList<Option>();

	private final List<Option> outer = new ArrayList<Option>();

	private final String innerAttribute;

	private final String outerAttribute;
	
	private int tuplecount;

	private int operandCost;

	private final Comparator<Histogram> c = new Comparator<Histogram>(){
		public int compare(Histogram a, Histogram b) {
			return a.identifier.fragmentnr-b.identifier.fragmentnr;
		}
	};

	public CompositeOption(NodeHandle site, Set<String> tables,
			String innerAttribute, String outerAttribute) {
		super(site, tables);
		this.innerAttribute = innerAttribute;
		this.outerAttribute = outerAttribute;
		tuplecount = -1;
	}
	
	public String getInner() {
		return null;
	}

  private boolean isContained(AttributeConstraint constraint)
  {
    boolean contained = false;
    Iterator<Option> it = getSubOptions();
    while (it.hasNext()) {
      Option o = it.next();
      if (o instanceof BaseAccessOption)
        if (o.constraints.contains(constraint)) {
          contained = true;
          break;
        } else if (o instanceof CompositeOption) {
          if (((CompositeOption) o).isContained(constraint)) {
            contained = true;
            break;
          }
        }

    }
    return contained;
  }

	@Override
  public void addConstraints(Set<SimpleConstraint> queryConstraints)
  {
    Set<SimpleConstraint> filteredConstraints = new HashSet<SimpleConstraint>();
    for (SimpleConstraint sc : queryConstraints)
      if (sc instanceof AttributeConstraint)
        if (!isContained((AttributeConstraint) sc))
          filteredConstraints.add(sc);
    super.addConstraints(filteredConstraints);
  }

  @Override
	public int getPriorCost() {
		return operandCost;
	}

	@Override
	public int getTupleCount() {
		if (tuplecount != -1)
			return tuplecount;
		else if (Dascosa.isPropertySet(Dascosa.USE_HISTOGRAMS) && inner.get(0) instanceof BaseAccessOption && outer.get(0) instanceof BaseAccessOption){
			Dascosa.out.println(FilteredOutput.OPTIMIZER, "Get tuplecount");
			List<Histogram> innerHist = getHistogramsForBaseOptions(inner);
			List<Histogram> outerHist = getHistogramsForBaseOptions(outer);
			
			Double factor = DataStatistics.getInstance().reductionFactors.get(tables);
			int sumInner = 0;
			for (Histogram h : innerHist)
				sumInner += h.tupleCount;
			int sumOuter = 0;
			for (Histogram h : outerHist)
				sumOuter += h.tupleCount;
			
			if (factor == null){
				Dascosa.out.println(FilteredOutput.OPTIMIZER, "inner: "+sumInner +", outer: "+sumOuter);
				factor = DataStatistics.estimateTupleSize(innerHist, outerHist) / (sumInner * sumOuter);
				
				DataStatistics.getInstance().reductionFactors.put(tables, factor);
			}
			Dascosa.out.println(FilteredOutput.OPTIMIZER, "Factor: "+factor.toString()+" for tables: "+tables.toString());
			tuplecount = (int) (factor*(double)sumInner*(double)sumOuter * getConstraintReductionFactor());
			//return tuplecount;
		} else if (Dascosa.isPropertySet(Dascosa.USE_HISTOGRAMS) && inner.get(0) instanceof BaseAccessOption && outer.get(0) instanceof CompositeOption){
			estimateBasicAndComposite(inner, (CompositeOption) outer.get(0), outerAttribute);
			//return tuplecount;
		} else if (Dascosa.isPropertySet(Dascosa.USE_HISTOGRAMS) && outer.get(0) instanceof BaseAccessOption && inner.get(0) instanceof CompositeOption){
			estimateBasicAndComposite(outer, (CompositeOption) inner.get(0), innerAttribute);
			//return tuplecount;
		} else {
			System.out.println("No histogram used");
			int innerSum = 0;
			for (Option o : inner) {
				innerSum += o.getTupleCount();
			}
			int outerSum = 0;
			for (Option o : outer) {
				outerSum += o.getTupleCount();
			}
			tuplecount = (int)(Math.max(innerSum, outerSum) * getConstraintReductionFactor());
			//return tuplecount;
		}
		Dascosa.out.println(FilteredOutput.DATA_STATISTICS, tables.toString()+" estimeres til: "+tuplecount+"tupler");
		return tuplecount;
		
	}

	private void estimateBasicAndComposite(List<Option> basic, CompositeOption compositeOption, String attributefromComposite) throws Error {
		List<Histogram> outerHist = getHistogramsForBaseOptions(basic);
		int sumOuter = 0;
		for (Histogram h : outerHist)
			sumOuter += h.tupleCount;
		int sumInner = compositeOption.getTupleCount();
		Set<String> tables = new HashSet<String>();
		String compositeTableName = compositeOption.getTableNameForAttribute(attributefromComposite);
		tables.add(compositeTableName);
		tables.addAll(basic.get(0).tables);
		Double factor = DataStatistics.getInstance().reductionFactors.get(tables);
		if (factor == null){
			List<Histogram> innerHist = compositeOption.getHistograms(compositeTableName, attributefromComposite);//DataStatistics.getInstance().getHistogram(new HistogramIdentifier())
			factor = DataStatistics.estimateTupleSize(outerHist, innerHist) / (sumInner * sumOuter);
			DataStatistics.getInstance().reductionFactors.put(tables, factor);
		}
		tuplecount = (int) (factor*sumInner*sumOuter * getConstraintReductionFactor());
		Dascosa.out.println(FilteredOutput.OPTIMIZER, "Factor: "+factor.toString()+" for tables: "+tables.toString());
	}

	private List<Histogram> getHistogramsForBaseOptions(Iterable<Option> operandOptions) throws Error {
		List<Histogram> result = new ArrayList<Histogram>();
		for (Option o : operandOptions){
			if (o instanceof BaseAccessOption){
				BaseAccessOption option = (BaseAccessOption) o;
				HistogramIdentifier h = new HistogramIdentifier(option.tables.iterator().next(), option.getFragmentNr(), innerAttribute);
				result.add(DataStatistics.getInstance().getHistogram(h));
			}else
				throw new Error("Mixed suboptions not supported");
		}
		
		Collections.sort(result, c);
		return result;
	}

	public void addInnerOperand(Option prefReplicate, double cost) {
		operandCost += cost;
		inner.add(prefReplicate);
	}

	public void addOuterOperand(Option prefReplicate, double cost) {
		operandCost += cost;
		outer.add(prefReplicate);
	}

	@Override
	public Iterator<Option> getSubOptions() {
    return new Iterator<Option>() {
      private final Iterator<Option> a = inner.iterator();

      private final Iterator<Option> b = outer.iterator();

      public boolean hasNext() {
        return a.hasNext() || b.hasNext();
      }

      public Option next() {
        if (a.hasNext())
          return a.next();
        else
          return b.next();
      }

      public void remove() {
        throw new UnsupportedOperationException();
      }
    };
	}

	@Override
  protected String getTableNameForAttribute(String attribute) {
		String table = inner.get(0).getTableNameForAttribute(attribute);
		if (table == null)
			table = outer.get(0).getTableNameForAttribute(attribute);
		Dascosa.out.println(FilteredOutput.OPTIMIZER, "TEST: "+attribute);
		return table;
	}

	@Override
  protected List<Histogram> getHistograms(String table, String attribute) {
		List<Histogram> histograms = new ArrayList<Histogram>();
		for (Option o : inner)
			histograms.addAll(o.getHistograms(table, attribute));
		if (histograms.isEmpty())
			for (Option o : outer)
				histograms.addAll(o.getHistograms(table, attribute));
		Collections.sort(histograms, c);
		return histograms;
	}

  @Override
  protected void attributesContained(Map<String, Boolean> contained) {
    inner.get(0).attributesContained(contained);
    outer.get(0).attributesContained(contained);
  }
}
