package dascosa.site.update;

import dascosa.message.request.*;
import dascosa.site.*;
import rice.p2p.commonapi.NodeHandle;

import javax.sql.DataSource;
import java.sql.*;
import java.util.HashMap;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 06.nov.2007
 * Time: 14:59:04
 * To change this template use File | Settings | File Templates.
 */
public final class CommitManager implements MessageClient, InvalidateReplicatesClient
{
  private final DataSource dataSource;
  private final InvalidateReplicatesHandler invalidateReplicatesHandler;
  private final Map<CommitStatement, Connection> activeStatements = new HashMap<CommitStatement, Connection>();
  private final NetworkLayer networkLayer;
  private final RequestResponseManager requestResponseManager;
  private final SiteManager siteManager;

  public CommitManager(SiteManager siteManager, NetworkLayer networkLayer, RequestResponseManager requestResponseManager,
                       InvalidateReplicatesHandler invalidateReplicatesHandler,
                       DataSource dataSource)
  {
    this.siteManager = siteManager;
    this.requestResponseManager = requestResponseManager;
    this.networkLayer = networkLayer;
    this.invalidateReplicatesHandler = invalidateReplicatesHandler;
    this.dataSource = dataSource;
  }

// --------------------- Begin Interface InvalidateReplicatesClient ---------------------

  public void replicatesInvalidated(String tableName, int fragmentNr, Object parameters[], DataSource dataSource)
  {
    StatementRequestMsg request = (StatementRequestMsg)parameters[0];
    executeStatement(request, this.dataSource);
  }

// --------------------- End Interface InvalidateReplicatesClient ---------------------


// --------------------- Begin Interface MessageClient ---------------------

  public void responseReceived(AbstractResponseMsg msg, Object parameters[], DataSource dataSource)
  {
    if (msg instanceof LockResponseMsg) {
      LockResponseMsg message = (LockResponseMsg)msg;
      Transaction transaction = (Transaction)parameters[0];
      transaction.gotLockResult(message.getFragmentNr(), message.isGranted());
      if (transaction.hasAllLockResults()) {
        if (transaction.hasAllLocks()) {
          for (int i = 0; i < transaction.cohorts.length; i++)
            requestResponseManager.sendRequest(new StatementRequestMsg(transaction.statement,
                                                                       message.getTableName(),
                                                                       i + 1, transaction.cohorts[i]),
                                               transaction.getMaster(i), this, new Object[] {transaction});
        }
        else {
          releaseLocks(transaction);
          transaction.client.statementCompleted(transaction.statement, false);
        }
      }
    }
    else if (msg instanceof StatementResponseMsg) {
      StatementResponseMsg message = (StatementResponseMsg)msg;
      Transaction transaction = (Transaction)parameters[0];
      transaction.gotResult(message.getSite(), message.getResult());
      if (transaction.hasAllResults()) {
        boolean isCommit = transaction.isCommit();
        for (NodeHandle[] replicates : transaction.cohorts)
          requestResponseManager.sendRequest(new CommitRequestMsg(transaction.statement, isCommit), replicates[0],
                                             this, new Object[] {transaction});
      }
    }
    else if (msg instanceof CommitResponseMsg) {
      CommitResponseMsg message = (CommitResponseMsg)msg;
      Transaction transaction = (Transaction)parameters[0];
      transaction.gotCommit(message.getSite());
      if (transaction.hasAllCommits()) {
        releaseLocks(transaction);
        transaction.client.statementCompleted(transaction.statement, transaction.isCommit());
      }
    }
  }

// --------------------- End Interface MessageClient ---------------------

  public void executeCommit(CommitRequestMsg message)
  {
    try {
      Connection connection = activeStatements.remove(message.getCommitStatement());
      if (message.isCommit())
        connection.commit();
      else
        connection.rollback();
      connection.close();
    } catch (SQLException e) {
      e.printStackTrace();
    }
    networkLayer.sendMessage(new CommitResponseMsg(message, siteManager.getLocalSite()),
                             message.getReturnToSite());
  }

  // Entry point
  public void startCommitProcess(CommitClient client, CommitStatement statement,
                                 NodeHandle sites[][], int masterIndex[],
                                 int fragmentNumbers[], String tableName)
  {
    Transaction transaction = new Transaction(client, statement, sites, masterIndex, fragmentNumbers, tableName);
    for (int i = 0; i < sites.length; i++) {
      requestResponseManager.sendRequest(new LockRequestMsg(tableName, i + 1, true), transaction.getMaster(i),
                                         this, new Object[] {transaction});
    }
  }

  // Executed at master site
  public void statementRequested(StatementRequestMsg message, DataSource dataSource)
  {
    // TODO: Invalidate is not needed in case of no changes...
    if (message.getReplicates().length > 1)  // Must invalidate
      invalidateReplicatesHandler.invalidateReplicates(message.getTableName(),
                                                       message.getFragmentNr(),
                                                       message.getReplicates(),
                                                       this, new Object[] {message});
    else
      executeStatement(message, dataSource);
  }

  // Executed at master site
  private void executeStatement(StatementRequestMsg message, DataSource dataSource)
  {
    try {
      CommitStatement commitStatement = message.getCommitStatement();
      Connection connection = dataSource.getConnection();
      Statement statement = connection.createStatement();
      activeStatements.put(commitStatement, connection);
      int result = commitStatement.executeStatement(message.getTableName(),
                                                    message.getFragmentNr(), statement);
      statement.close();
      networkLayer.sendMessage(new StatementResponseMsg(message, siteManager.getLocalSite(), result),
                               message.getReturnToSite());
    } catch (SQLException e) {
      e.printStackTrace();
      networkLayer.sendMessage(new StatementResponseMsg(message, siteManager.getLocalSite(), CommitStatement.STATEMENT_FAILED),
                               message.getReturnToSite());
    }
  }

  private void releaseLocks(Transaction transaction)
  {
    for (int i = 0; i < transaction.cohorts.length; i++)
      if (transaction.hasLock[i]) {
        AbstractRequestMsg unlockMsg = new LockRequestMsg(transaction.tableName, i + 1, false);
        unlockMsg.setReturnToSite(siteManager.getLocalSite());
        networkLayer.sendMessage(unlockMsg, transaction.tableName);
      }
  }

// -------------------------- INNER CLASSES --------------------------

  private static final class Transaction
  {
    private final CommitClient client;
    private final CommitStatement statement;
    private final NodeHandle cohorts[][];
    private final int results[];
    private final boolean committed[];
    private final boolean hasLock[];
    private final boolean hasLockResult[];
    private final String tableName;
    private final int masterIndex[];
    private final int fragmentNumbers[];

    private Transaction(CommitClient client, CommitStatement statement, NodeHandle[][] cohorts,
                        int[] masterIndex, int fragmentNumbers[], String tableName)
    {
      this.client = client;
      this.statement = statement;
      this.cohorts = cohorts;
      this.masterIndex = masterIndex;
      this.fragmentNumbers = fragmentNumbers;
      this.tableName = tableName;
      results = new int[cohorts.length];
      committed = new boolean[cohorts.length];
      hasLock = new boolean[cohorts.length];
      hasLockResult = new boolean[cohorts.length];
    }

    private void gotResult(NodeHandle site, int result)
    {
      for (int i = 0; i < cohorts.length; i++)
        if (cohorts[i][0].equals(site))
          results[i] = result;
    }

    private boolean hasAllResults()
    {
      for (int result : results)
        if (result == 0)
          return false;
      return true;
    }

    private boolean isCommit()
    {
      for (int result : results)
        if (result == CommitStatement.STATEMENT_FAILED)
          return false;
      return true;
    }

    private void gotCommit(NodeHandle site)
    {
      for (int i = 0; i < cohorts.length; i++)
        if (cohorts[i][0].equals(site))
          committed[i] = true;
    }

    private boolean hasAllCommits()
    {
      for (boolean commit : committed)
        if (!commit)
          return false;
      return true;
    }

    private void gotLockResult(int fragmentNr, boolean granted)
    {
      for (int i = 0; i < fragmentNumbers.length; i++) {
        if (fragmentNumbers[i] == fragmentNr) {
          hasLockResult[i] = true;
          hasLock[i] = granted;
          return;
        }
      }
      System.err.println("CommitManager$Transaction.gotLockResult: Unable to find fragment " + fragmentNr);
    }

    private NodeHandle getMaster(int fragmentNr)
    {
      for (int i = 0; i < fragmentNumbers.length; i++) {
        if (fragmentNumbers[i] == fragmentNr)
          return cohorts[i][masterIndex[i]];
      }
      System.err.println("CommitManager$Transaction.getMaster: Unable to find fragment " + fragmentNr);
      return null;
    }

    private boolean hasAllLockResults()
    {
      for (boolean lockResult : hasLockResult)
        if (!lockResult)
          return false;
      return true;
    }

    boolean hasAllLocks()
    {
      for (boolean lock : hasLock)
        if (!lock)
          return false;
      return true;
    }
  }
}
