package dascosa.site.update;

import dascosa.algebra.AlgebraConverter;
import dascosa.message.request.*;
import dascosa.parser.Node;
import dascosa.parser.ParseException;
import dascosa.site.*;
import dascosa.site.localdb.LocalFragmentHandler;
import dascosa.site.localdb.TableFragment;
import rice.p2p.commonapi.NodeHandle;

import javax.sql.DataSource;
import java.sql.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 21.nov.2007
 * Time: 13:56:23
 * To change this template use File | Settings | File Templates.
 */
public final class InsertHandler implements MessageClient, InvalidateReplicatesClient
{
  private final DataSource dataSource;
  private final InvalidateReplicatesHandler invalidateReplicatesHandler;
  private final LocalFragmentHandler localFragmentHandler;
  private final NetworkLayer networkLayer;
  private final RequestResponseManager requestResponseManager;
  private final SiteManager siteManager;

  public InsertHandler(SiteManager siteManager, NetworkLayer networkLayer,
                       RequestResponseManager requestResponseManager,
                       InvalidateReplicatesHandler invalidateReplicatesHandler,
                       DataSource dataSource,
                       LocalFragmentHandler localFragmentHandler)
  {
    this.siteManager = siteManager;
    this.networkLayer = networkLayer;
    this.requestResponseManager = requestResponseManager;
    this.invalidateReplicatesHandler = invalidateReplicatesHandler;
    this.dataSource = dataSource;
    this.localFragmentHandler = localFragmentHandler;
  }

// --------------------- Begin Interface InvalidateReplicatesClient ---------------------

  public void replicatesInvalidated(String tableName, int fragmentNr, Object parameters[], DataSource dataSource)
  {
    InsertRequestMsg request = (InsertRequestMsg)parameters[0];
    executeStatement(request, this.dataSource);
  }

// --------------------- End Interface InvalidateReplicatesClient ---------------------


// --------------------- Begin Interface MessageClient ---------------------

  public void responseReceived(AbstractResponseMsg msg, Object parameters[], DataSource dataSource)
  {
    if (msg instanceof LookupResponseMsg) { // Initial lookup
      LookupResponseMsg message = (LookupResponseMsg)msg;
      String sql = (String)parameters[0];
      Node topNode = (Node)parameters[1];
      try {
        int pkValue = AlgebraConverter.findPrimaryKeyValue(topNode, message.getPrimaryKey());
        AbstractRequestMsg request = new LookupPKRequestMsg(message.getTableName(), pkValue);
        requestResponseManager.sendRequest(request, message.getTableName(), this, new Object[] {sql, pkValue});
      } catch (ParseException e) {
        siteManager.fireDebugErrorEvent(e.getMessage());
      }
    }
    else if (msg instanceof LookupPKResponseMsg) { // Primary key lookup
      LookupPKResponseMsg message = (LookupPKResponseMsg)msg;
      String sql = (String)parameters[0];
      int pkValue = (Integer)parameters[1];
      NodeHandle sites[] = message.getSites();
      NodeHandle masterSite = sites[message.getMasterIndex()];
      AbstractRequestMsg request = new LockRequestMsg(message.getTableName(), message.getFragmentNr(), true);
      requestResponseManager.sendRequest(request, masterSite, this, new Object[] {sql, sites, masterSite, pkValue});
    }
    else if (msg instanceof LockResponseMsg) {
      LockResponseMsg message = (LockResponseMsg)msg;
      if (message.isGranted()) {
        String sql = (String)parameters[0];
        if (sql.endsWith(";"))
          sql = sql.substring(0, sql.length() - 1);
        NodeHandle sites[] = (NodeHandle[])parameters[1];
        NodeHandle masterSite = (NodeHandle)parameters[2];
        int pkValue = (Integer)parameters[3];
        CommitStatement commitStatement = new UpdateStatement(sql);
        AbstractRequestMsg request = new InsertRequestMsg(commitStatement, message.getTableName(),
                                                          message.getFragmentNr(), sites, pkValue);
        requestResponseManager.sendRequest(request, masterSite, this, new Object[] {message.getTableName(),
                                                                                    message.getFragmentNr()});
      }
      else
        siteManager.fireDebugErrorEvent("Error: Unable to get fragment lock.\n");
    }
    else if (msg instanceof InsertResponseMsg) {
      InsertResponseMsg message = (InsertResponseMsg)msg;
      if (message.getResult() == CommitStatement.STATEMENT_OK_CHANGED)
        siteManager.fireDebugMessageEvent("Statement committed.\n");
      else
        siteManager.fireDebugErrorEvent("Error: Statement failed.\n");
      String tableName = (String)parameters[0];
      int fragmentNr = (Integer)parameters[1];
      AbstractRequestMsg unlockMsg = new LockRequestMsg(tableName, fragmentNr, false);
      unlockMsg.setReturnToSite(siteManager.getLocalSite());
      networkLayer.sendMessage(unlockMsg, message.getMasterSite());
    }
  }

// --------------------- End Interface MessageClient ---------------------

  // Entry point

  public void executeInsert(String sql, Node sqlNode)
  {
    String tableName = AlgebraConverter.findTableName(sqlNode);
    AbstractRequestMsg message = new LookupRequestMsg(tableName);
    requestResponseManager.sendRequest(message, tableName, this,
                                       new Object[] {sql, sqlNode});
  }

  // Executed at master site
  public void insertRequested(InsertRequestMsg message, DataSource dataSource)
  {
    if (message.getReplicateSites().length > 1) { // Must invalidate
      invalidateReplicatesHandler.invalidateReplicates(message.getTableName(), message.getFragmentNr(),
                                                       message.getReplicateSites(),
                                                       this, new Object[] {message});
    }
    else
      executeStatement(message, dataSource);
  }

  // Executed at master site
  private void executeStatement(InsertRequestMsg message, DataSource dataSource)
  {
    try {
      CommitStatement commitStatement = message.getCommitStatement();
      Connection connection = dataSource.getConnection();
      Statement statement = connection.createStatement();
      int result = commitStatement.executeStatement(message.getTableName(),
                                                    message.getFragmentNr(), statement);
      statement.close();
      connection.close();
      if (result == CommitStatement.STATEMENT_OK_CHANGED) {
        TableFragment fragment = localFragmentHandler.getTableFragment(message.getTableName(), message.getFragmentNr());
        fragment.tupleWritten(message.getReturnToSite(), message.getPkValue());
      }
      networkLayer.sendMessage(new InsertResponseMsg(message, siteManager.getLocalSite(), result),
                               message.getReturnToSite());
    } catch (SQLException e) {
      e.printStackTrace();
      networkLayer.sendMessage(new InsertResponseMsg(message, siteManager.getLocalSite(),
                                                     CommitStatement.STATEMENT_FAILED),
                               message.getReturnToSite());
    }
  }
}
