package dascosa.site.update;

import dascosa.message.request.LockRequestMsg;
import dascosa.message.request.LockResponseMsg;
import dascosa.site.NetworkLayer;
import dascosa.site.localdb.*;
import rice.p2p.commonapi.NodeHandle;

import java.util.HashMap;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 07.jul.2008
 * Time: 15:11:13
 * To change this template use File | Settings | File Templates.
 */
public final class LockManager implements DatabaseListener
{
  private final LocalFragmentHandler localFragmentHandler;
  private final Map<TableFragment, LockInfo> locks = new HashMap<TableFragment, LockInfo>();
  private final NetworkLayer networkLayer;

  public LockManager(LocalFragmentHandler localFragmentHandler, NetworkLayer networkLayer)
  {
    this.localFragmentHandler = localFragmentHandler;
    this.networkLayer = networkLayer;
    localFragmentHandler.addDatabaseListener(this);
  }

// --------------------- Begin Interface DatabaseListener ---------------------

  public void localReplicateAdded(TableFragment replicate)
  {
  }

  public synchronized void localReplicateRemoved(TableFragment replicate)
  {
    if (locks.containsKey(replicate))
      locks.remove(replicate);
  }

// --------------------- End Interface DatabaseListener ---------------------


  public synchronized void forceLocalLock(String tableName, int fragmentNr, NodeHandle localSite)
  {
    TableFragment tableFragment = localFragmentHandler.getTableFragment(tableName, fragmentNr);
    if (tableFragment == null) {
      System.err.println("LockManager: Unknown fragment " + tableFragment);
      return;
    }
    LockInfo lockInfo = locks.get(tableFragment);
    if (lockInfo == null) {
      lockInfo = new LockInfo(localSite);
      locks.put(tableFragment, lockInfo);
    }
    else if (!lockInfo.locked) {
      lockInfo.lockOwner = localSite;
      lockInfo.locked = true;
    }
  }

  public synchronized boolean isLocked(String tableName, int fragmentNr)
  {
    TableFragment tableFragment = localFragmentHandler.getTableFragment(tableName, fragmentNr);
    if (tableFragment == null) {
      System.err.println("LockManager: Unknown fragment " + tableFragment);
      return false;
    }
    LockInfo lockInfo = locks.get(tableFragment);
    return lockInfo != null && lockInfo.locked;
  }

  public synchronized boolean localLockRequested(String tableName, int fragmentNr, boolean isLockRequest,
                                                 NodeHandle localSite)
  {
    TableFragment tableFragment = localFragmentHandler.getTableFragment(tableName, fragmentNr);
    if (tableFragment == null) {
      System.err.println("LockManager: Unknown fragment " + tableFragment);
      return false;
    }
    else if (!tableFragment.isMaster()) {
      System.err.println("LockManager: Can't lock nonmaster " + tableFragment);
      return false;
    }
    else if (isLockRequest) {
      LockInfo lockInfo = locks.get(tableFragment);
      if (lockInfo == null) {
        lockInfo = new LockInfo(localSite);
        locks.put(tableFragment, lockInfo);
        return true;
      }
      else if (!lockInfo.locked) {
        lockInfo.lockOwner = localSite;
        lockInfo.locked = true;
        return true;
      }
      else
        return false;
    }
    else {
      LockInfo lockInfo = locks.get(tableFragment);
      if (lockInfo.locked) {
        if (lockInfo.lockOwner.equals(localSite))
          lockInfo.locked = false;
        else {
          System.err.println("LockManager: Can't unlock - not owner " + tableFragment);
          System.err.println("LockManager: Owner " + lockInfo.lockOwner);
          System.err.println("LockManager: Client Local");
        }
      }
      return true;
    }
  }

  public synchronized void lockRequested(LockRequestMsg request)
  {
    TableFragment tableFragment = localFragmentHandler.getTableFragment(request.getTableName(),
                                                                        request.getFragmentNr());
    if (tableFragment == null)
      System.err.println("LockManager: Unknown fragment " + request);
    else if (!tableFragment.isMaster())
      System.err.println("LockManager: Can't lock nonmaster " + request);
    else if (request.isLockRequest()) {
      LockInfo lockInfo = locks.get(tableFragment);
      if (lockInfo == null) {
        lockInfo = new LockInfo(request.getReturnToSite());
        locks.put(tableFragment, lockInfo);
        networkLayer.sendMessage(new LockResponseMsg(request, true), request.getReturnToSite());
      }
      else if (!lockInfo.locked) {
        lockInfo.lockOwner = request.getReturnToSite();
        lockInfo.locked = true;
        networkLayer.sendMessage(new LockResponseMsg(request, true), request.getReturnToSite());
      }
      else
        networkLayer.sendMessage(new LockResponseMsg(request, false), request.getReturnToSite());
    }
    else {
      LockInfo lockInfo = locks.get(tableFragment);
      if (lockInfo.locked) {
        if (lockInfo.lockOwner.equals(request.getReturnToSite()))
          lockInfo.locked = false;
        else {
          System.err.println("LockManager: Can't unlock - not owner " + request);
          System.err.println("LockManager: Owner " + lockInfo.lockOwner);
          System.err.println("LockManager: Client " + request.getReturnToSite());
        }
      }
    }
  }

// -------------------------- INNER CLASSES --------------------------

  private static final class LockInfo
  {
    private NodeHandle lockOwner;
    private boolean locked;

    private LockInfo(NodeHandle lockOwner)
    {
      this.lockOwner = lockOwner;
      locked = true;
    }
  }
}
