package dascosa.util;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 19.feb.2008
 * Time: 13:17:39
 * To change this template use File | Settings | File Templates.
 */
public final class ArrayHashMap<V>
{
  private static final float DEFAULT_LOAD_FACTOR = 0.75f;
  private static final int DEFAULT_INITIAL_CAPACITY = 16;
  private static final int MAXIMUM_CAPACITY = 1 << 30;
  private final float loadFactor;
  private final int indexes[];
  private transient volatile Collection<V> values;
  private transient Entry<V>[] table;
  private transient volatile Set<Object[]> keySet;
  private transient int size;
  private int threshold;
  private transient volatile int modCount;

  /**
   * Applies a supplemental hash function to a given hashCode, which
   * defends against poor quality hash functions.  This is critical
   * because HashMap uses power-of-two length hash tables, that
   * otherwise encounter collisions for hashCodes that do not differ
   * in lower bits. Note: Null keys always map to hash 0, thus index 0.
   */
  private static int hash(int h)
  {
    // This function ensures that hashCodes that differ only by
    // constant multiples at each bit position have a bounded
    // number of collisions (approximately 8 at default load factor).
    h ^= h >>> 20 ^ h >>> 12;
    return h ^ h >>> 7 ^ h >>> 4;
  }

  private static int hashIndex(Object key[], int idx[])
  {
    int hash = 0;
    if (idx == null)
      return 0;
    if (key[idx[0]] != null)
      hash = key[idx[0]].hashCode();
    for (int i = 1; i < idx.length; i++)
      if (key[idx[i]] != null)
        hash ^= key[idx[i]].hashCode();
    return hash;
  }

  /**
   * Returns index for hash code h.
   */
  private static int indexFor(int h, int length)
  {
    return h & length - 1;
  }

  /**
   * Constructs an empty <tt>HashMap</tt> with the default initial capacity
   * (16) and the default load factor (0.75).
   */
  public ArrayHashMap(int indexes[])
  {
    this.indexes = indexes;
    loadFactor = DEFAULT_LOAD_FACTOR;
    threshold = (int)(DEFAULT_INITIAL_CAPACITY * DEFAULT_LOAD_FACTOR);
    table = new Entry[DEFAULT_INITIAL_CAPACITY];
  }

  public boolean containsKey(Object[] key)
  {
    if (key == null)
      return false;
    return get(key) != null;
  }

  /**
   * Returns the value to which the specified key is mapped,
   * or {@code null} if this map contains no mapping for the key.
   * <p/>
   * <p>More formally, if this map contains a mapping from a key
   * {@code k} to a value {@code v} such that {@code (key==null ? k==null :
   * key.equals(k))}, then this method returns {@code v}; otherwise
   * it returns {@code null}.  (There can be at most one such mapping.)
   * <p/>
   * <p>A return value of {@code null} does not <i>necessarily</i>
   * indicate that the map contains no mapping for the key; it's also
   * possible that the map explicitly maps the key to {@code null}.
   * The {@link #containsKey containsKey} operation may be used to
   * distinguish these two cases.
   */
  public V get(Object key[])
  {
    if (key == null)
      return getForNullKey();
    int hash = ArrayHashMap.hash(ArrayHashMap.hashIndex(key, indexes));
    for (Entry<V> e = table[ArrayHashMap.indexFor(hash, table.length)];
         e != null;
         e = e.next) {
      Object k[];
      if (e.hash == hash && ((k = e.key) == key || keyEquals(key, k, indexes)))
        return e.value;
    }
    return null;
  }

  public V get(Object[] key, int[] idx)
  {
    if (key == null)
      return getForNullKey();
    int hash = ArrayHashMap.hash(ArrayHashMap.hashIndex(key, idx));
    for (Entry<V> e = table[ArrayHashMap.indexFor(hash, table.length)];
         e != null;
         e = e.next) {
      Object k[];
      if (e.hash == hash && ((k = e.key) == key || keyEquals(key, k, idx)))
        return e.value;
    }
    return null;
  }

  public Set<Object[]> keySet()
  {
    Set<Object[]> ks = keySet;
    return ks != null ? ks : (keySet = new KeySet());
  }

  /**
   * Associates the specified value with the specified key in this map.
   * If the map previously contained a mapping for the key, the old
   * value is replaced.
   *
   * @param key   key with which the specified value is to be associated
   * @param value value to be associated with the specified key
   * @return the previous value associated with <tt>key</tt>, or
   *         <tt>null</tt> if there was no mapping for <tt>key</tt>.
   *         (A <tt>null</tt> return can also indicate that the map
   *         previously associated <tt>null</tt> with <tt>key</tt>.)
   */
  public V put(Object[] key, V value)
  {
    if (key == null)
      return putForNullKey(value);
    int hash = ArrayHashMap.hash(ArrayHashMap.hashIndex(key, indexes));
    int i = ArrayHashMap.indexFor(hash, table.length);
    for (Entry<V> e = table[i]; e != null; e = e.next) {
      Object k[];
      if (e.hash == hash && ((k = e.key) == key || keyEquals(key, k, indexes))) {
        V oldValue = e.value;
        e.value = value;
        return oldValue;
      }
    }

    modCount++;
    addEntry(hash, key, value, i);
    return null;
  }

  /**
   * Returns a {@link Collection} view of the values contained in this map.
   * The collection is backed by the map, so changes to the map are
   * reflected in the collection, and vice-versa.  If the map is
   * modified while an iteration over the collection is in progress
   * (except through the iterator's own <tt>remove</tt> operation),
   * the results of the iteration are undefined.  The collection
   * supports element removal, which removes the corresponding
   * mapping from the map, via the <tt>Iterator.remove</tt>,
   * <tt>Collection.remove</tt>, <tt>removeAll</tt>,
   * <tt>retainAll</tt> and <tt>clear</tt> operations.  It does not
   * support the <tt>add</tt> or <tt>addAll</tt> operations.
   */
  public Collection<V> values()
  {
    Collection<V> vs = values;
    return vs != null ? vs : (values = new Values());
  }

  /**
   * Adds a new entry with the specified key, value and hash code to
   * the specified bucket.  It is the responsibility of this
   * method to resize the table if appropriate.
   * <p/>
   * Subclass overrides this to alter the behavior of put method.
   */
  void addEntry(int hash, Object key[], V value, int bucketIndex)
  {
    Entry<V> e = table[bucketIndex];
    table[bucketIndex] = new Entry<V>(hash, key, value, e);
    if (size++ >= threshold)
      resize(2 * table.length);
  }

  /**
   * Removes all of the mappings from this map.
   * The map will be empty after this call returns.
   */
  void clear()
  {
    modCount++;
    Entry<V>[] tab = table;
    for (int i = 0; i < tab.length; i++)
      tab[i] = null;
    size = 0;
  }

  /**
   * Special-case code for containsValue with null argument
   */
  private boolean containsNullValue()
  {
    Entry<V>[] tab = table;
    for (Entry<V> aTab : tab)
      for (Entry<V> e = aTab; e != null; e = e.next)
        if (e.value == null)
          return true;
    return false;
  }

  /**
   * Returns <tt>true</tt> if this map maps one or more keys to the
   * specified value.
   *
   * @param value value whose presence in this map is to be tested
   * @return <tt>true</tt> if this map maps one or more keys to the
   *         specified value
   */
  boolean containsValue(Object value)
  {
    if (value == null)
      return containsNullValue();

    Entry<V>[] tab = table;
    for (Entry<V> aTab : tab)
      for (Entry<V> e = aTab; e != null; e = e.next)
        if (value.equals(e.value))
          return true;
    return false;
  }

  /**
   * Offloaded version of get() to look up null keys.  Null keys map
   * to index 0.  This null case is split out into separate methods
   * for the sake of performance in the two most commonly used
   * operations (get and put), but incorporated with conditionals in
   * others.
   */
  private V getForNullKey()
  {
    for (Entry<V> e = table[0]; e != null; e = e.next) {
      if (e.key == null)
        return e.value;
    }
    return null;
  }

  private boolean keyEquals(Object key1[], Object key2[], int idx1[])
  {
    boolean equal = true;
    if (idx1 == null)
      return true;
    for (int i = 0; i < idx1.length; i++) {
      if (key1[idx1[i]] == null && key2[indexes[i]] != null)
        return false;
      else if (key1[idx1[i]] != null)
        equal = equal && key1[idx1[i]].equals(key2[indexes[i]]);
    }
    return equal;
  }

  Iterator<Object[]> newKeyIterator()
  {
    return new KeyIterator();
  }

  Iterator<V> newValueIterator()
  {
    return new ValueIterator();
  }

  /**
   * Offloaded version of put for null keys
   */
  private V putForNullKey(V value)
  {
    for (Entry<V> e = table[0]; e != null; e = e.next) {
      if (e.key == null) {
        V oldValue = e.value;
        e.value = value;
        return oldValue;
      }
    }
    modCount++;
    addEntry(0, null, value, 0);
    return null;
  }

  /**
   * Removes and returns the entry associated with the specified key
   * in the HashMap.  Returns null if the HashMap contains no mapping
   * for this key.
   */
  Entry<V> removeEntryForKey(Object key[])
  {
    int hash = key == null ? 0 : ArrayHashMap.hash(ArrayHashMap.hashIndex(key, indexes));
    int i = ArrayHashMap.indexFor(hash, table.length);
    Entry<V> prev = table[i];
    Entry<V> e = prev;

    while (e != null) {
      Entry<V> next = e.next;
      Object k[];
      if (e.hash == hash &&
          ((k = e.key) == key || key != null && keyEquals(key, k, indexes))) {
        modCount++;
        size--;
        if (prev == e)
          table[i] = next;
        else
          prev.next = next;
        return e;
      }
      prev = e;
      e = next;
    }

    return e;
  }

  /**
   * Rehashes the contents of this map into a new array with a
   * larger capacity.  This method is called automatically when the
   * number of keys in this map reaches its threshold.
   * <p/>
   * If current capacity is MAXIMUM_CAPACITY, this method does not
   * resize the map, but sets threshold to Integer.MAX_VALUE.
   * This has the effect of preventing future calls.
   *
   * @param newCapacity the new capacity, MUST be a power of two;
   *                    must be greater than current capacity unless current
   *                    capacity is MAXIMUM_CAPACITY (in which case value
   *                    is irrelevant).
   */
  void resize(int newCapacity)
  {
    Entry<V>[] oldTable = table;
    int oldCapacity = oldTable.length;
    if (oldCapacity == MAXIMUM_CAPACITY) {
      threshold = Integer.MAX_VALUE;
      return;
    }

    Entry<V>[] newTable = new Entry[newCapacity];
    transfer(newTable);
    table = newTable;
    threshold = (int)(newCapacity * loadFactor);
  }

  /**
   * Transfers all entries from current table to newTable.
   */
  void transfer(Entry<V>[] newTable)
  {
    Entry<V>[] src = table;
    int newCapacity = newTable.length;
    for (int j = 0; j < src.length; j++) {
      Entry<V> e = src[j];
      if (e != null) {
        src[j] = null;
        do {
          Entry<V> next = e.next;
          int i = ArrayHashMap.indexFor(e.hash, newCapacity);
          e.next = newTable[i];
          newTable[i] = e;
          e = next;
        } while (e != null);
      }
    }
  }

// -------------------------- INNER CLASSES --------------------------

  private final class Entry<V> implements Map.Entry<Object[], V>
  {
    final Object[] key;
    V value;
    Entry<V> next;
    final int hash;

    /**
     * Creates new entry.
     */
    private Entry(int h, Object k[], V v, Entry<V> n)
    {
      value = v;
      next = n;
      key = k;
      hash = h;
    }

    public Object[] getKey()
    {
      return key;
    }

    public V getValue()
    {
      return value;
    }

    public V setValue(V newValue)
    {
      V oldValue = value;
      value = newValue;
      return oldValue;
    }

    public boolean equals(Object o)
    {
      if (!(o instanceof Map.Entry))
        return false;
      Map.Entry<Object[], V> e = (Map.Entry<Object[], V>)o;
      Object k1[] = getKey();
      Object k2[] = e.getKey();
      if (k1 == k2 || k1 != null && keyEquals(k1, k2, indexes)) {
        Object v1 = getValue();
        Object v2 = e.getValue();
        if (v1 == v2 || v1 != null && v1.equals(v2))
          return true;
      }
      return false;
    }

    public int hashCode()
    {
      return (key == null ? 0 : ArrayHashMap.hashIndex(key, indexes)) ^
             (value == null ? 0 : value.hashCode());
    }

    public String toString()
    {
      return Arrays.toString(getKey()) + '=' + getValue();
    }
  }

  private abstract class HashIterator<E> implements Iterator<E>
  {
    Entry<V> next;  // next entry to return
    int expectedModCount;  // For fast-fail
    int index;    // current slot
    Entry<V> current;  // current entry

    HashIterator()
    {
      expectedModCount = modCount;
      if (size > 0) { // advance to first entry
        Entry<V>[] t = table;
        while (index < t.length && (next = t[index++]) == null)
          ;
      }
    }

    public boolean hasNext()
    {
      return next != null;
    }

    Entry<V> nextEntry()
    {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException();
      Entry<V> e = next;
      if (e == null)
        throw new NoSuchElementException();

      if ((next = e.next) == null) {
        Entry<V>[] t = table;
        while (index < t.length && (next = t[index++]) == null)
          ;
      }
      current = e;
      return e;
    }

    public void remove()
    {
      if (current == null)
        throw new IllegalStateException();
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException();
      Object k[] = current.key;
      current = null;
      removeEntryForKey(k);
      expectedModCount = modCount;
    }
  }

  private final class KeyIterator extends HashIterator<Object[]>
  {
    public Object[] next()
    {
      return nextEntry().getKey();
    }
  }

  private final class KeySet extends AbstractSet<Object[]>
  {
    public Iterator<Object[]> iterator()
    {
      return newKeyIterator();
    }

    public int size()
    {
      return size;
    }

    public boolean contains(Object o)
    {
      return containsKey((Object[])o);
    }

    public boolean remove(Object o)
    {
      return removeEntryForKey((Object[])o) != null;
    }

    public void clear()
    {
      ArrayHashMap.this.clear();
    }
  }

  private final class ValueIterator extends HashIterator<V>
  {
    public V next()
    {
      return nextEntry().value;
    }
  }

  private final class Values extends AbstractCollection<V>
  {
    public Iterator<V> iterator()
    {
      return newValueIterator();
    }

    public int size()
    {
      return size;
    }

    public boolean contains(Object o)
    {
      return containsValue(o);
    }

    public void clear()
    {
      ArrayHashMap.this.clear();
    }
  }
}
