package dascosa.util.fragsim;

import dascosa.util.fragsim.workload.Site;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 08.okt.2008
 * Time: 12:57:43
 * To change this template use File | Settings | File Templates.
 */
final class AccessLogger
{
  public static final int START_WIDTH = AdaptiveAlgorithm.MINIMUM_FRAGMENT;
  private final Map<Site, BucketSet> currentBuckets = new HashMap<Site, BucketSet>();
  private final Map<Site, BucketSet> oldBuckets = new HashMap<Site, BucketSet>();
  private final Replicate replicate;
  private int bucketWidth;

  AccessLogger(Replicate replicate, int bucketWidth)
  {
    this.replicate = replicate;
    this.bucketWidth = bucketWidth;
  }

  private void checkWidthDecrease()
  {
    int maxBuckets = -1;
    for (BucketSet bucketSet : currentBuckets.values())
      maxBuckets = Math.max(maxBuckets, bucketSet.getBucketsUsed());
    for (BucketSet bucketSet : oldBuckets.values())
      maxBuckets = Math.max(maxBuckets, bucketSet.getBucketsUsed());

    if (maxBuckets * BucketSet.WIDTH_INCREASE_FACTOR <=
        BucketSet.MAX_BUCKETS && bucketWidth >= BucketSet.WIDTH_INCREASE_FACTOR) {
      bucketWidth /= BucketSet.WIDTH_INCREASE_FACTOR;
      for (BucketSet bucketSet : currentBuckets.values())
        bucketSet.makeSmallerBuckets(BucketSet.WIDTH_INCREASE_FACTOR);
      for (BucketSet bucketSet : oldBuckets.values())
        bucketSet.makeSmallerBuckets(BucketSet.WIDTH_INCREASE_FACTOR);
    }
  }

  void clearBuckets()
  {
    oldBuckets.clear();
    currentBuckets.clear();
  }

  void combineLogs(AccessLogger logger1, AccessLogger logger2)
  {
    Collection<Site> sites = new HashSet<Site>(logger1.currentBuckets.keySet());
    sites.addAll(logger2.currentBuckets.keySet());

    int maxBuckets = -1;
    BucketSet oldBucketSets[] = new BucketSet[2];
    for (Site site : sites) {
      oldBucketSets[0] = logger1.currentBuckets.get(site);
      oldBucketSets[1] = logger2.currentBuckets.get(site);
      BucketSet bucketSet = new BucketSet(this, oldBucketSets);
      currentBuckets.put(site, bucketSet);
      maxBuckets = Math.max(maxBuckets, bucketSet.getBucketsUsed());
    }

    if (maxBuckets > BucketSet.MAX_BUCKETS) {
      bucketWidth *= BucketSet.WIDTH_INCREASE_FACTOR;
      for (BucketSet bucketSet : currentBuckets.values())
        bucketSet.makeBiggerBuckets(BucketSet.WIDTH_INCREASE_FACTOR);
    }
  }

  int computeBucketNumber(int id)
  {
    return computeBucketNumber(id, getBucketWidth());
  }

  int computeBucketNumber(int id, int width)
  {
    return (int)((id - (long)replicate.getFragment().getMinAssigned()) / width);
  }

  /**
   * ID equal to the lower bound of the bucket
   */
  int computeID(int bucketNumber)
  {
    return (int)replicate.getFragment().getMinAssigned() + bucketNumber * getBucketWidth();
  }

  Collection<Site> getActiveSites()
  {
    return currentBuckets.keySet();
  }

  List<Integer> getBucketNumbers()
  {
    Set<Integer> bucketNumbers = new HashSet<Integer>();
    for (BucketSet bucketSet : currentBuckets.values())
      bucketNumbers.addAll(bucketSet.getBucketNumbers());
    for (BucketSet bucketSet : oldBuckets.values())
      bucketNumbers.addAll(bucketSet.getBucketNumbers());
    List<Integer> l = new ArrayList<Integer>(bucketNumbers);
    Collections.sort(l);
    return l;
  }

  int getBucketWidth()
  {
    return bucketWidth;
  }

  int getHighestBucketNumber()
  {
    int num = Integer.MIN_VALUE;
    for (BucketSet bucketSet : currentBuckets.values())
      num = Math.max(num, bucketSet.getHighestBucketNumber());
    for (BucketSet bucketSet : oldBuckets.values())
      num = Math.max(num, bucketSet.getHighestBucketNumber());
    return num;
  }

  int getLowestBucketNumber()
  {
    int num = Integer.MAX_VALUE;
    for (BucketSet bucketSet : currentBuckets.values())
      num = Math.min(num, bucketSet.getLowestBucketNumber());
    for (BucketSet bucketSet : oldBuckets.values())
      num = Math.min(num, bucketSet.getLowestBucketNumber());
    return num;
  }

  long getMaxAssigned()
  {
    return replicate.getFragment().getMaxAssigned();
  }

  int getReads(List<Integer> bucketNumbers, int startIndex, int endIndex)
  {
    int operationCount = 0;
    for (BucketSet currentSet : currentBuckets.values())
      operationCount += currentSet.getReads(bucketNumbers, startIndex, endIndex);
    for (BucketSet oldSet : oldBuckets.values())
      operationCount += oldSet.getReads(bucketNumbers, startIndex, endIndex);
    return operationCount;
  }

  int getReadsFrom(Site site, List<Integer> bucketNumbers, int startIndex, int endIndex)
  {
    int operationCount = 0;
    BucketSet currentSet = currentBuckets.get(site);
    if (currentSet != null)
      operationCount += currentSet.getReads(bucketNumbers, startIndex, endIndex);
    BucketSet oldSet = oldBuckets.get(site);
    if (oldSet != null)
      operationCount += oldSet.getReads(bucketNumbers, startIndex, endIndex);
    return operationCount;
  }

  int getWrites(List<Integer> bucketNumbers, int startIndex, int endIndex)
  {
    int operationCount = 0;
    for (BucketSet currentSet : currentBuckets.values())
      operationCount += currentSet.getWrites(bucketNumbers, startIndex, endIndex);
    for (BucketSet oldSet : oldBuckets.values())
      operationCount += oldSet.getWrites(bucketNumbers, startIndex, endIndex);
    return operationCount;
  }

  int getWritesFrom(Site site, List<Integer> bucketNumbers, int startIndex, int endIndex)
  {
    int operationCount = 0;
    BucketSet currentSet = currentBuckets.get(site);
    if (currentSet != null)
      operationCount += currentSet.getWrites(bucketNumbers, startIndex, endIndex);
    BucketSet oldSet = oldBuckets.get(site);
    if (oldSet != null)
      operationCount += oldSet.getWrites(bucketNumbers, startIndex, endIndex);
    return operationCount;
  }

  void makeOtherBucketsBigger(BucketSet originatingBucketSet, int factor)
  {
    bucketWidth *= factor;
    for (BucketSet bucketSet : currentBuckets.values())
      if (bucketSet != originatingBucketSet)
        bucketSet.makeBiggerBuckets(factor);
    for (BucketSet bucketSet : oldBuckets.values())
      bucketSet.makeBiggerBuckets(factor);
  }

  void recordAccess(Site site, int id, boolean isWrite)
  {
    BucketSet bucketSet = currentBuckets.get(site);
    if (bucketSet == null) {
      bucketSet = new BucketSet(this);
      currentBuckets.put(site, bucketSet);
    }
    bucketSet.recordAccess(id, isWrite);
  }

  void resetBuckets()
  {
    oldBuckets.clear();
    oldBuckets.putAll(currentBuckets);
    currentBuckets.clear();
    checkWidthDecrease();
  }
}
