package dascosa.util.fragsim;

import dascosa.util.fragsim.workload.Site;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 11.jan.2008
 * Time: 14:55:34
 * To change this template use File | Settings | File Templates.
 */
final class AdaptiveAlgorithm
{
  static final int MINIMUM_FRAGMENT = 10;
  static final int NUMBER_OF_COSTS = 5;
  static final int REMOTE_WRITE = 0;
  static final int REMOTE_READ = 1;
  static final int REMOTE_UPDATE = 2;
  static final int MIGRATE_FRAGMENT = 3;
  static final int MIGRATE_TUPLES = 4;
  private static final long RUN_TIME_INTERVAL = 30000; // Every 30 sec
  private final List<RefragmentSuggestion> suggestions = new ArrayList<RefragmentSuggestion>();
  private final List<RefragmentSuggestion> refragmentations = new ArrayList<RefragmentSuggestion>();
  private final Map<Replicate, Long> lastRun = new HashMap<Replicate, Long>(); 
  private final Simulator index;
  private final double wFS;
  private final double wLOp;
  private final int[] messages = new int[NUMBER_OF_COSTS];

  AdaptiveAlgorithm(Simulator index, double wLOp, double wFS)
  {
    this.index = index;
    this.wLOp = wLOp;
    this.wFS = wFS;
  }

  int[] computeCost(Replicate replicate, Site site, boolean isWrite)
  {
    Arrays.fill(messages, 0);
    if (replicate.getCurrentLocation() != site) {
      if (replicate.getCurrentLocation() == null)
        replicate.setCurrentLocation(site);
      else {
        if (isWrite)
          messages[REMOTE_WRITE] = 1;
        else
          messages[REMOTE_READ] = 1;
      }
    }
    if (isWrite)
      messages[REMOTE_UPDATE] = replicate.getFragment().getNumberOfReadReplicas();

    if (lastRun.get(replicate) == null)
      lastRun.put(replicate, System.currentTimeMillis());
    else if (lastRun.get(replicate) + RUN_TIME_INTERVAL < System.currentTimeMillis()) {
      if (Simulator.DEBUG)
        System.out.println("##### RUN ALGORITHM on " + replicate + " #####");
      if (replicate.isMaster()) {
        boolean split = runMasterAlgorithm(replicate);
        if (split)
          index.joinLocalMasterReplicas(messages);
      }
      else
        runNonMasterAlgorithm(replicate);
      lastRun.put(replicate, System.currentTimeMillis());
    }

    return messages;
  }

  int[] computeCost(Replicate replicate, Site site, boolean isWrite, long clock)
  {
    Arrays.fill(messages, 0);
    if (replicate.getCurrentLocation() != site) {
      if (replicate.getCurrentLocation() == null)
        replicate.setCurrentLocation(site);
      else {
        if (isWrite)
          messages[REMOTE_WRITE] = 1;
        else
          messages[REMOTE_READ] = 1;
      }
    }
    if (isWrite)
      messages[REMOTE_UPDATE] = replicate.getFragment().getNumberOfReadReplicas();

    if (lastRun.get(replicate) == null)
      lastRun.put(replicate, clock);
    else if (lastRun.get(replicate) + RUN_TIME_INTERVAL < clock) {
      if (Simulator.DEBUG)
        System.out.println("##### RUN ALGORITHM on " + replicate + " #####");
      if (replicate.isMaster())
        runMasterAlgorithm(replicate);
      else
        runNonMasterAlgorithm(replicate);
      lastRun.put(replicate, clock);
    }

    return messages;
  }

  private boolean runMasterAlgorithm(Replicate replicate)
  {
    AccessLogger logger = replicate.getLogger();
    suggestions.clear();
    List<Integer> bucketNumbers = logger.getBucketNumbers();
    for (int startIndex = 0; startIndex < bucketNumbers.size(); startIndex++) {
      for (int endIndex = startIndex; endIndex < bucketNumbers.size(); endIndex++) {
        int localWrites = logger.getWritesFrom(replicate.getCurrentLocation(), bucketNumbers, startIndex, endIndex);
        int fragmentSize = replicate.getFragment().getSize(startIndex, endIndex);
        for (Site site : logger.getActiveSites()) {
          if (site != replicate.getCurrentLocation()) {
            int remoteWrites = logger.getWritesFrom(site, bucketNumbers, startIndex, endIndex);
            double utility = wLOp * remoteWrites - localWrites - wFS * fragmentSize;
            if (utility > 0 && (startIndex < endIndex || logger.computeID(bucketNumbers.get(endIndex) + 1) - 1 <= replicate.getFragment().getMaxAssigned()) &&
                (bucketNumbers.get(endIndex) - bucketNumbers.get(startIndex) + 1) * logger.getBucketWidth() >= MINIMUM_FRAGMENT) {
              suggestions.add(new RefragmentSuggestion(utility, bucketNumbers.get(startIndex),
                                                       bucketNumbers.get(endIndex), site));
            }
          }
        }
      }
    }
    if (suggestions.isEmpty()) {
      int writes = logger.getWrites(bucketNumbers, 0, bucketNumbers.size() - 1);
      int fragmentSize = replicate.getFragment().getSize(0, bucketNumbers.size() - 1);
      for (Site site : logger.getActiveSites()) {
        if (site != replicate.getCurrentLocation() && !replicate.getFragment().hasReplicateOn(site)) {
          int remoteReads = logger.getReadsFrom(site, bucketNumbers, 0, bucketNumbers.size() - 1);
          double createUtility = wLOp * remoteReads - writes - wFS * fragmentSize;
          if (createUtility > 0) {
            if (Simulator.DEBUG)
              System.out.println(replicate.getFragment() + " bør repliseres på " + site);
            replicate.getFragment().addReplicate(site);
            messages[MIGRATE_FRAGMENT]++;
            messages[MIGRATE_TUPLES] += replicate.getFragment().getSize();
          }
        }
      }

      logger.resetBuckets();
      return false;
    }
    else {
      Collections.sort(suggestions);
      refragmentations.clear();
      refragmentations.add(suggestions.remove(0));
      while (!suggestions.isEmpty()) {
        RefragmentSuggestion suggestion = suggestions.remove(0);
        boolean compatible = true;
        for (RefragmentSuggestion refragmentation : refragmentations)
          if (!refragmentation.isCompatible(suggestion))
            compatible = false;
        if (compatible)
          refragmentations.add(suggestion);
      }

      if (refragmentations.get(0).startBucket == logger.getLowestBucketNumber() &&
          refragmentations.get(0).endBucket == logger.getHighestBucketNumber()) { // Migrate the whole fragment
        replicate.setCurrentLocation(refragmentations.get(0).site); // Migrate the whole fragment
        if (Simulator.DEBUG)
          System.out.println("Migrering av hele: " + replicate.getFragment());
        logger.clearBuckets();
        messages[MIGRATE_FRAGMENT]++;
        messages[MIGRATE_TUPLES] += replicate.getFragment().getSize();
      }

      else {
        if (Simulator.DEBUG)
          System.out.println("Skal splitt+migrere " + replicate.getFragment());

        Fragment newFragments[] = replicate.getFragment().split(refragmentations);
        for (Fragment newFragment : newFragments) {
          if (newFragment.getMasterReplica().getCurrentLocation() != replicate.getCurrentLocation()) {
            messages[MIGRATE_FRAGMENT]++;
            messages[MIGRATE_TUPLES] += newFragment.getSize();
          }
        }
        index.splitFragment(replicate.getFragment(), newFragments, messages);
        lastRun.remove(replicate);
      }
      return true;
    }
  }

  private void runNonMasterAlgorithm(Replicate replicate)
  {
    AccessLogger logger = replicate.getLogger();
    List<Integer> bucketNumbers = logger.getBucketNumbers();

    // 1: Check for replicate deletion
    int writes = logger.getWrites(bucketNumbers, 0, bucketNumbers.size() - 1);
    int localReads = logger.getReadsFrom(replicate.getCurrentLocation(), bucketNumbers, 0, bucketNumbers.size() - 1);
    double deleteUtility = wLOp * writes - localReads;
    if (deleteUtility > 0) {
      if (Simulator.DEBUG)
        System.out.println(replicate + " slettet");
      replicate.getFragment().removeReplicate(replicate);
      lastRun.remove(replicate);
    }

    // 2: Check for new replicates
    int fragmentSize = replicate.getFragment().getSize(0, bucketNumbers.size() - 1);
    for (Site site : logger.getActiveSites()) {
      if (site != replicate.getCurrentLocation() && !replicate.getFragment().hasReplicateOn(site)) {
        int remoteReads = logger.getReadsFrom(site, bucketNumbers, 0, bucketNumbers.size() - 1);
        double createUtility = wLOp * remoteReads - writes - wFS * fragmentSize;
        if (createUtility > 0) {
          if (Simulator.DEBUG)
            System.out.println(replicate.getFragment() + " bør repliseres på " + site);
          replicate.getFragment().addReplicate(site);
          messages[MIGRATE_FRAGMENT]++;
          messages[MIGRATE_TUPLES] += replicate.getFragment().getSize();
        }
      }
    }
    logger.resetBuckets();
  }
}
