package dascosa.util.fragsim;

import dascosa.util.fragsim.workload.Site;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: joh
 * Date: 03.des.2007
 * Time: 15:05:04
 * To change this template use File | Settings | File Templates.
 */
final class Fragment implements Comparable<Fragment>
{
  private final List<Replicate> readReplicas = new ArrayList<Replicate>();
  private final Map<Site, Replicate> replicateCache = new HashMap<Site, Replicate>();
  private final Replicate masterReplica;
  private final long minAssigned;
  private final long maxAssigned;
  private long minUsed = Long.MAX_VALUE;
  private long maxUsed = Long.MIN_VALUE;
  private long writeCounter;

  Fragment(long minAssigned, long maxAssigned)
  {
    this(minAssigned, maxAssigned, AccessLogger.START_WIDTH);
  }

  /**
   * Joins two fragments
   */
  Fragment(Fragment fragment1, Fragment fragment2, int messages[])
  {
    this(Math.min(fragment1.minAssigned, fragment2.minAssigned),
         Math.max(fragment1.maxAssigned, fragment2.maxAssigned),
         Math.max(fragment1.masterReplica.getLogger().getBucketWidth(),
                  fragment2.masterReplica.getLogger().getBucketWidth()));
    masterReplica.setCurrentLocation(fragment1.masterReplica.getCurrentLocation());
    minUsed = Math.min(fragment1.minUsed, fragment2.minUsed);
    maxUsed = Math.max(fragment1.maxUsed, fragment2.maxUsed);
    writeCounter = fragment1.writeCounter + fragment2.writeCounter;

    masterReplica.getLogger().combineLogs(fragment1.masterReplica.getLogger(),
                                          fragment2.masterReplica.getLogger());

    // ReadReplicas
    // 1: Combine replicas on sites which have both
    // 2: Migrate+combine replicas that requires least migration effort (sum of transferred tuples)
    // 3: Invalidate other replicas (implicit)
    Collection<Replicate> aloneReplicas1 = new HashSet<Replicate>(fragment1.readReplicas);
    Collection<Replicate> aloneReplicas2 = new HashSet<Replicate>(fragment2.readReplicas);
    for (Replicate replicate1 : fragment1.readReplicas) {
      for (Replicate replicate2 : fragment2.readReplicas) {
        if (replicate1.getCurrentLocation() == replicate2.getCurrentLocation()) {
          readReplicas.add(new Replicate(this, replicate1, replicate2));
          if (Simulator.DEBUG)
            System.out.println("Combining replicas of " + this + " at " + replicate1.getCurrentLocation());
          aloneReplicas1.remove(replicate1);
          aloneReplicas2.remove(replicate2);
        }
      }
    }
    long migrateTuples1 = aloneReplicas2.size() * fragment1.getSize();
    long migrateTuples2 = aloneReplicas1.size() * fragment2.getSize();
    if (migrateTuples1 < migrateTuples2) { // Migrate fragment1, invalidate remaining replicas 2
      for (Replicate replicate2 : aloneReplicas2) {
        Replicate replicate1 = new Replicate(this, AccessLogger.START_WIDTH, false);
        replicate1.setCurrentLocation(replicate2.getCurrentLocation());
        readReplicas.add(new Replicate(this, replicate1, replicate2));
        if (Simulator.DEBUG)
          System.out.println("Migrate+Combining replicas of " + this + " at " + replicate1.getCurrentLocation());
      }
      messages[AdaptiveAlgorithm.MIGRATE_FRAGMENT] += aloneReplicas2.size();
      messages[AdaptiveAlgorithm.MIGRATE_TUPLES] += migrateTuples1;
    }
    else {
      for (Replicate replicate1 : aloneReplicas1) {
        Replicate replicate2 = new Replicate(this, AccessLogger.START_WIDTH, false);
        replicate2.setCurrentLocation(replicate1.getCurrentLocation());
        readReplicas.add(new Replicate(this, replicate1, replicate2));
        if (Simulator.DEBUG)
          System.out.println("Migrate+Combining replicas of " + this + " at " + replicate1.getCurrentLocation());
      }
      messages[AdaptiveAlgorithm.MIGRATE_FRAGMENT] += aloneReplicas1.size();
      messages[AdaptiveAlgorithm.MIGRATE_TUPLES] += migrateTuples2;
    }
  }

  private Fragment(long minAssigned, long maxAssigned, int bucketWidth)
  {
    this.minAssigned = minAssigned;
    this.maxAssigned = maxAssigned;
    if (maxAssigned < minAssigned) {
      throw new RuntimeException("ERROR: Negative fragment size");
    }
    masterReplica = new Replicate(this, bucketWidth, true);
  }

  private Fragment(long minAssigned, long maxAssigned, Fragment oldFragment)
  {
    this(minAssigned, maxAssigned, AccessLogger.START_WIDTH);
    masterReplica.setCurrentLocation(oldFragment.masterReplica.getCurrentLocation());
    for (Replicate oldReplicate : oldFragment.readReplicas) {
      Replicate newReplicate = new Replicate(this, AccessLogger.START_WIDTH, false);
      newReplicate.setCurrentLocation(oldReplicate.getCurrentLocation());
      readReplicas.add(newReplicate);
      for (Site site : oldFragment.replicateCache.keySet()) {
        if (oldFragment.replicateCache.get(site) == oldReplicate)
          replicateCache.put(site, newReplicate);
      }
    }
  }

// --------------------- Begin Interface Comparable ---------------------

  public int compareTo(Fragment o)
  {
    if (minAssigned < o.minAssigned)
      return -1;
    if (minAssigned == o.minAssigned)
      return 0;
    return 1;
  }

// --------------------- End Interface Comparable ---------------------

  public String toString()
  {
//    System.out.println(currentBuckets.values());
    return "Fragment [" + minAssigned + ',' + maxAssigned + "] @ " + masterReplica + ' ' + readReplicas;
  }

  void addReplicate(Site site)
  {
    Replicate replicate = new Replicate(this, AccessLogger.START_WIDTH, false);
    replicate.setCurrentLocation(site);
    readReplicas.add(replicate);
  }

  private double coverage(Fragment newFragment)
  {
    return (newFragment.maxUsed - newFragment.minUsed + 1) / (double)(maxUsed - minUsed + 1);
  }

  Replicate getMasterReplica()
  {
    return masterReplica;
  }

  long getMaxAssigned()
  {
    return maxAssigned;
  }

  long getMinAssigned()
  {
    return minAssigned;
  }

  int getNumberOfReadReplicas()
  {
    return readReplicas.size();
  }

  /**
   * Number of unique values
   */
  int getSize()
  {
    if (writeCounter == 0)
      return 0;
    double m = maxUsed - minUsed + 1;
    double n = writeCounter;
    return (int)(m * (1 - Math.pow((m - 1) / m, n)));
  }

  /**
   * Number of unique values
   */
  int getSize(int startIndex, int endIndex)
  {
    if (writeCounter == 0)
      return 0;
    double m = Math.min(maxUsed - minUsed + 1, (endIndex - startIndex + 1) * masterReplica.getLogger().getBucketWidth());
    double n = Math.min(writeCounter, writeCounter * m / (maxUsed - minUsed + 1));
    return (int)(m * (1 - Math.pow((m - 1) / m, n)));
  }

  boolean hasReplicateOn(Site site)
  {
    if (masterReplica.getCurrentLocation() == site)
      return true;
    for (Replicate replicate : readReplicas)
      if (replicate.getCurrentLocation() == site)
        return true;
    return false;
  }

  boolean isResponsible(long id)
  {
    return id >= minAssigned && id <= maxAssigned;
  }

  Replicate recordAccess(Site site, int id, boolean isWrite)
  {
    if (!isResponsible(id))
      throw new RuntimeException("ERROR: Access outside responsibility");
    if (isWrite) {
      minUsed = Math.min(id, minUsed);
      maxUsed = Math.max(id, maxUsed);
      if (masterReplica.getCurrentLocation() == null)
        masterReplica.setCurrentLocation(site);
      masterReplica.getLogger().recordAccess(site, id, isWrite);
      for (Replicate readReplicate : readReplicas)
        readReplicate.getLogger().recordAccess(site, id, isWrite); // Update replicas
      writeCounter++;
      return masterReplica;
    }
    else {
      if (readReplicas.isEmpty() || masterReplica.getCurrentLocation() == site) { // 1: No replicas => Master
        masterReplica.getLogger().recordAccess(site, id, isWrite);
        replicateCache.put(site, masterReplica);
        return masterReplica;
      }
      for (Replicate replicate : readReplicas) // 2: Local replicate
        if (replicate.getCurrentLocation() == site) {
          replicate.getLogger().recordAccess(site, id, isWrite);
          replicateCache.put(site, replicate);
          return replicate;
        }
      if (replicateCache.containsKey(site)) { // 3: Already read from one?
        Replicate replicate = replicateCache.get(site);
        replicate.getLogger().recordAccess(site, id, isWrite);
        return replicate;
      }
      int counts[] = new int[readReplicas.size() + 1]; // 4: Read from least used replicate
      for (Replicate replicate : replicateCache.values()) {
        if (replicate.isMaster())
          counts[counts.length - 1]++;
        else
          counts[readReplicas.indexOf(replicate)]++;
      }
      int minValue = Integer.MAX_VALUE;
      int minIndex = -1;
      for (int i = 0; i < counts.length; i++) {
        if (counts[i] < minValue) {
          minIndex = i;
          minValue = counts[i];
        }
      }
      Replicate replicate;
      if (minIndex == counts.length - 1)
        replicate = masterReplica;
      else
        replicate = readReplicas.get(minIndex);
      replicate.getLogger().recordAccess(site, id, isWrite);
      replicateCache.put(site, replicate);
      return replicate;
    }
  }

  void removeReplicate(Replicate replicate)
  {
    readReplicas.remove(replicate);
    for (Iterator<Site> iterator = replicateCache.keySet().iterator(); iterator.hasNext();) {
      Site site = iterator.next();
      if (replicateCache.get(site) == replicate)
        iterator.remove();
    }
  }

  Fragment[] split(List<RefragmentSuggestion> refragmentations)
  {
    Collections.sort(refragmentations, new Comparator<RefragmentSuggestion>()
    {
      public int compare(RefragmentSuggestion o1, RefragmentSuggestion o2)
      {
        return o1.startBucket - o2.startBucket;
      }
    });

    List<Fragment> newFragments = new ArrayList<Fragment>();
    int lastBucketUsed = -1;
    for (RefragmentSuggestion refragmentation : refragmentations) {
      long min = minAssigned + refragmentation.startBucket * masterReplica.getLogger().getBucketWidth();
      long max = minAssigned + (refragmentation.endBucket + 1) * masterReplica.getLogger().getBucketWidth() - 1;
      max = Math.min(maxAssigned, max);

      // 1: Create local fragment (if any)
      if (lastBucketUsed + 1 < refragmentation.startBucket) {
        long lMin = minAssigned + (lastBucketUsed + 1) * masterReplica.getLogger().getBucketWidth();
        Fragment fragment = new Fragment(lMin, min - 1, this);
        fragment.writeCounter = (long)(writeCounter * coverage(fragment));
        fragment.minUsed = Math.max(minUsed, fragment.minAssigned);
        fragment.maxUsed = fragment.maxAssigned;
        newFragments.add(fragment);
      }

      // 2: Create remote fragment
      Fragment rFragment = new Fragment(min, max, this);
      rFragment.getMasterReplica().setCurrentLocation(refragmentation.site);
      rFragment.writeCounter = (long)(writeCounter * coverage(rFragment));
      rFragment.minUsed = Math.max(minUsed, rFragment.minAssigned);
      rFragment.maxUsed = Math.min(maxUsed, rFragment.maxAssigned);
      newFragments.add(rFragment);

      lastBucketUsed = refragmentation.endBucket;
    }
    // 3: Create final local fragment (if any)
    if (newFragments.get(newFragments.size() - 1).maxAssigned < maxAssigned) {
      Fragment fragment = new Fragment(newFragments.get(newFragments.size() - 1).maxAssigned + 1, maxAssigned, this);
      fragment.writeCounter = (long)(writeCounter * coverage(fragment));
      fragment.minUsed = fragment.minAssigned;
      fragment.maxUsed = maxUsed;
      newFragments.add(fragment);
    }

//    Collections.sort(newFragments);
//    if (newFragments.size() > 1) {
//      for (int i = 0; i < newFragments.size() - 1; i++) {
//        Fragment fragment1 = newFragments.get(i);
//        Fragment fragment2 = newFragments.get(i + 1);
//        if (fragment2.getMinAssigned() != fragment1.getMaxAssigned() + 1) {
//          System.err.println("Split failed: Overlap mismatch" + this);
//          System.err.println(fragment1 + " <=> " + fragment2);
//          System.err.println(newFragments);
//          System.err.println(refragmentations);
//          throw new RuntimeException("FRAGMENT ERROR");
//        }
//      }
//    }
//    if (getMinAssigned() != newFragments.get(0).getMinAssigned()) {
//      System.err.println("Split failed: Start mismatch");
//      System.err.println(this + " => " + newFragments.get(0));
//      System.err.println(newFragments);
//      System.err.println(refragmentations);
//      throw new RuntimeException("FRAGMENT ERROR");
//    }
//    if (getMaxAssigned() != newFragments.get(newFragments.size() - 1).getMaxAssigned()) {
//      System.err.println("Split failed: End mismatch");
//      System.err.println(this + " => " + newFragments.get(newFragments.size() - 1));
//      System.err.println(newFragments);
//      System.err.println(refragmentations);
//      System.err.println(getMasterReplica().getLogger().getHighestBucketNumber());
//      throw new RuntimeException("FRAGMENT ERROR");
//    }


    Fragment newF[] = new Fragment[newFragments.size()];
    for (int i = 0; i < newFragments.size(); i++)
      newF[i] = newFragments.get(i);

    if (Simulator.DEBUG) {
      System.out.println("Split of: " + this);
      System.out.println("Result: " + Arrays.toString(newF));
    }

    return newF;
  }
}
